/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \defgroup choquet_interaction Intgrale de choquet avec intraction

Dcoupage de l'oprateur en 3 fonctions :

	- lecture du nom du fichier contenant les paramtres :
	param *choquet_interaction_lect(choquet_interaction_t *desc, param *ptp, char *debq);

	- initialisation des valeurs des paramtres :
	int choquet_interaction_init(choquet_interaction_t *choquet_interaction_desc);

	- calcul de la valeur de l'intgrale de Choquet pour une srie de valeurs donne :
	int choquet_interaction_calc(choquet_interaction_t *choquet_interaction_desc,float *donnee,float *resultat);

	- libration de l'espace mmoire :
	int choquet_interaction_ferm(choquet_interaction_t *desc);





Fonctionnement :

Le fichier de dpart doit contenir sur chaque ligne i, le coefficient de pondration associ au critre i suivi des
coefficients d'intractions associs :


nb_entre <br>
v1  I12  I13  I14  ... <br>
v2  I23  I24  ...    <br>  
v3  I34  ... <br>
...     <br>



Ces diffrents coefficients doivent vrifier 2 conditions (le programme les vrifie) :
	- concernant les moyennes pondres :  $\sum_{i=1}^{n} \nu_{i} = 1$
	- concernant les coefficients d'intraction :  $\forall{i} \sum_{j=0}^{n} |I_{ij}| \leq 2.\nu_{i}$



Les constantes suivantes aident au debogage :

- AUCUNE_ERREUR est une constante qui vaut 0 et qui est affecte  l'erreur du
descripteur lorsqu'aucun problme est apparu a l'initialisation. Le calcul est alors possible.

- ERREUR_CHOQUET_FILE est une constante qui vaut 1 et qui est affecte  l'erreur du
descripteur lorsqu'un problme  l'ouverture du fichier s'est produite.

- ERREUR_CHOQUET_LIGNE_MATRICE est une constante qui vaut 2 et qui est affecte  l'erreur du
descripteur lorsqu'une incohrence entre nbcoeff et le nombre rel de paramtres ne
permettait pas de crer la matrice. Le problme se situe au niveau du nombre de colonnes dans le fichier
qui ne correspond pas au nombre de coefficients.

- ERREUR_CHOQUET_COLONNE_MATRICE est une constante qui vaut 3 et qui est affecte  l'erreur du
descripteur lorsqu'une incohrence entre nbcoeff et le nombre rel de paramtres ne
permettait pas de crer la matrice. Le probleme se situe au niveau du nombre de lignes dans le fichier
qui ne correspond pas au nombre de coefficients.

- ERREUR_CHOQUET_NU est une constante qui vaut 4 et qui est affecte  l'erreur du
descripteur lorsque la somme des coefficients de pondration est diffrente de 1.

- ERREUR_CHOQUET_INTERACTION est une constante qui vaut 5 et qui est affecte  l'erreur du
descripteur lorsque la relation entre les coefficients d'intractions et la moyenne pondre n'est pas remplie, 
ne permettant pas le passage des coefficients d'intractions vers la mesure floue additive.








@name lchoquet_interaction
@author Sylvie JULLIEN ET Emmanuel RAMASSO ET Lionel VALET 
@version 1.0 (5/06/03); Include : Prototypes.h

@see ff_symb
*/





#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "choquet.h"
#include "image.h"    /* types des donnees manipulees dans ima2D */
#include "proto2D.h"  /* descripteurs et prototypes des fonctions de ima2D */
#include "math.h"

/* *************************  LECTURE  *******************************/

/** \ingroup choquet_interaction
    lecture des parametres de l'operateur choquet_interaction. \\
    dimX, dimY : exemple d'entiers : nombre de colonnes, de lignes d'un masque\\
    gain : exemple de reel\\
    forme : exemple de chaine de charactere : rectangle/triangle\\

    @param desc1 : pointeur sur descripteur\\
    @param ptp pointeur sur parametre courant deja alloue (a lire),\\
    @param debq pointeur de chaine de caracteres placee en debut de question\\
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite)\\

    @author JULLIEN Sylvie ET RAMASSO Emmanuel
    @version 1.0 (5/06/03); Include : Prototypes.h

*/


/* void choquet_interaction_lect (choquet_interaction_desc *desc1)
	{

	printf("\nQuel est le chemin du fichier de parametres ? : ");
	scanf("%s",desc1->fichier);
	}*/


param *choquet_interaction_lect(choquet_interaction_t *desc1, param *ptp, char *debq){
	char question[500];

	sprintf(question, "%s choquet_interaction: nom complet du fichier", debq);
	lec_param(question, ptp);
	strcpy(desc1->fichier, ptp->rep);
	ptp = ptp->next;

	return(ptp);
	}



/* *************************  INITIALISATION  ***************************/
/** \ingroup choquet_interaction
    initialisation de l'operateur choquet_interaction.\\
    initialisation de l'erreur,
    ouverture du fichier,
    allocation de la matrice,
    verification de la coherence entre nombre de coefficients et coefficients effectivement rentres,
    si incorrect erreur = 1
    verification de la satisfaction des conditions sur les coefficients :
    - moyenne ponderee = 1 (sinon erreur = 2)
    - coefficients d'interaction (sinon erreur = 3).\\
\\

    @param desc2 : pointeur sur descripteur
    @return 0 si OK et 1 si probleme

    @author JULLIEN Sylvie ET RAMASSO Emmanuel
    @version 1.0 (5/06/03); Include : Prototypes.h

*/

int choquet_interaction_init(choquet_interaction_t *desc2)
	{

	/* declaration des variables locales */
	int l,c;
	FILE *fp;
	float somme,intercoeff;


	/*initialisation des parametres*/
	(desc2->erreur)= AUCUNE_ERREUR;

	/* allocation de la matrice des parametres */
	printf("-> Ouverture fichier en cours...");
	fp = fopen(desc2->fichier,"rt");
	/* test ouverture */
	if(fp == NULL)
		{
		printf("\t\tECHEC.\n");
		desc2->erreur = ERREUR_CHOQUET_FILE;
		return 1;
		}
		
	else
	{
	printf("\t\tOK.\n");

	/* initialisation de nbcoeff */
	printf("-> Initialisation de nbcoeff en cours...");
	fscanf(fp,"%d",&(desc2->nbcoeff));
	printf("\tNbcoeff = %d.\n",desc2->nbcoeff);


	/* allocation des pointeurs sur les lignes de la matrice de parametres */
	printf("-> Allocation memoire en cours...");
	(desc2->ptparam) = (parametre *)malloc((desc2->nbcoeff)*sizeof(parametre));
	printf("\t\tOK.\n");

	/* allocation des cases de la matrice*/
	printf("-> Allocation cases matrice en cours...");
	for(l=0;l<(desc2->nbcoeff);l++)
		(desc2->ptparam[l])=(float *)malloc( (desc2->nbcoeff)*sizeof(float));
	printf("\t\tOK.\n");



	/* remplir matrice avec coefficients */

	intercoeff = 0.0;

	for(l=0;l<(desc2->nbcoeff);l++)
		{

		/* on verifie que les parametres soient bien entres */
		if(intercoeff != EOF)
			{
			for(c=l;c<(desc2->nbcoeff);c++)
				{
				/* on lit un nouveau coefficient */
				fscanf(fp,"%f",&intercoeff);

				/* remplissage d une ligne */
				if(intercoeff != EOF)
					{
					(desc2->ptparam[l][c]) = intercoeff;

					/* remplissage de la partie inferieure de la matrice */
					(desc2->ptparam[c][l]) = (desc2->ptparam[l][c]);

					}
				else
					{
					/* si l erreur est detectee on sort de la fonction */
					(desc2->erreur) = ERREUR_CHOQUET_LIGNE_MATRICE;
					return 1;
					}


				}

			}
		else
			{
			(desc2->erreur) = ERREUR_CHOQUET_COLONNE_MATRICE;
			return 1;
			}

		}

	fclose(fp);

	/* on verifie les conditions prealables au calcul de l'integrale */


	/* somme des "nu i" != 1 => erreur 3  */

	somme = 0.0;
	for(l=0;l<(desc2->nbcoeff);l++)
			somme += desc2->ptparam[l][l];

	printf("-> La somme des nu_i est \t\t\t%1.4f\n",somme);

	if(somme != 1)
		{
		desc2->erreur = ERREUR_CHOQUET_NU;

		return 1;
		}


	/* verification pour chaque critere sinon => erreur 4*/

	for(l=0;l<(desc2->nbcoeff);l++)
		{

		/* initialisation de la somme */
		somme = 0.0;

		/* calcul de la somme */
		for(c=0;c<(desc2->nbcoeff);c++)
			{
			somme = somme + fabs(desc2->ptparam[l][c]);

			}

		/* verification de la somme */
		if(somme > (3*(desc2->ptparam[l][l])))
			{
			(desc2->erreur) = ERREUR_CHOQUET_INTERACTION;

			return 1;
			}
		}

	return 0;
	}
	}



/* *************************  CALCUL  ***************************/
/** \ingroup choquet_interaction
    fonction calcul de l'operateur choquet_interaction.\\
    Calcul du resultat de l'operateur choquet_interaction

    @param desc3 : pointeur sur descripteur
    @param donnee : pointeur sur flottants permettant la reutilisation du calcul
    @param resultat :  pointeur sur le resultat
    @return 0 si OK et 1 si probleme

    @author JULLIEN Sylvie ET RAMASSO Emmanuel
    @version 1.0 (5/06/03); Include : Prototypes.h
*/

int choquet_interaction_calc(choquet_interaction_t *desc3,float *donnee,float *resultat)
	{
	/* declaration des variables locales */
	float somme_prop,somme_inter;
	int l,c;

	if(desc3->erreur == AUCUNE_ERREUR)
		{
		/* initialisation de la somme de la partie proportionnelle */
		somme_prop = 0.0;

		/* calcul de la somme de la partie proportionnelle */
		for(l=0;l<(desc3->nbcoeff);l++)
			somme_prop = somme_prop + donnee[l]*(desc3->ptparam[l][l]);

		//printf("Moyenne ponderee : \t\t\t\t%2.4f\n",somme_prop);

		/* initialisation de la somme de la partie d'interaction */
		somme_inter = 0.0;

		/* calcul de la somme de la partie d'interaction */
		for(l=0;l<(desc3->nbcoeff);l++)
			for(c=l+1;c<(desc3->nbcoeff);c++)
				somme_inter = somme_inter + (desc3->ptparam[l][c])*fabs(donnee[c]-donnee[l]);

			
		//printf("Somme avec interactions : \t\t\t%2.4f\n",somme_inter);


		/* calcul final de l integrale de Choquet */
		*resultat = somme_prop - 0.5*somme_inter;
	
		return 0;
		}
	else
		{
		printf("\nCalcul impossible car probleme a l'initialisation.\n");

		return 1;
		}


	}



/*************************  FERMETURE  ***************************/
/** \ingroup choquet_interaction
    fonction fermeture de l'operateur choquet_interaction.
    Elle permet de liberer l'espace reservee durant l'execution.

    @param desc4 : pointeur sur descripteur

    @author JULLIEN Sylvie ET RAMASSO Emmanuel
    @version 1.0 (5/06/03); Include : Prototypes.h
*/


int choquet_interaction_ferm(choquet_interaction_t *desc4)
	{

	int ligne;

	/* liberation des matrices : desallocation des pointeurs sur les lignes de la matrice de parametres */

	for (ligne=0;ligne<(desc4->nbcoeff);ligne++)
 		free(desc4->ptparam[ligne]);

	free(desc4->ptparam);

	return 0;
	}


 

 
 
 

 
 
 
 
 
 
 
 
 
 
 
