/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file
   Circular Self-Organizing Map (CSOM) (Kohonen et al) Neural Network library for the segmentation of streams
   @author Petre Rodan 
   @version 1.0 (07/03/02)
   @see www.cis.hut.fi/teuvo
*/ 

/*----------------------------------------------------------------*/
/*								  */
/* Kohonen Circular Self Organizing Map Neural Network            */
/*								  */
/* author: Petre Rodan <rodan@subdimension.com>                   */
/*								  */
/* -------------------------------------------------------------- */

/* CSOM Neural Network library
    Description : Circular Self-Organizing Map (CSOM) (Kohonen et al) Neural Network library for the segmentation of streams
    It takes multiple (float *) streams of the same dimension and delivers a segmented output.
    @author Petre Rodan 
    @version 1.0 (07/03/02)
    @see www.cis.hut.fi/teuvo
*/


/** @addtogroup fkoh Kohonen Fusion
   @{
*/

#include "image.h"
#include "proto2D.h"
#include "lcsom.h"
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>

//---------------------------------------------------------------------------

/* *************************  LECTURE  *******************************/
/** lecture des parametres de l'operateur csom-> \n
    - input_neurons : unsigned int : number of input neurons\n
    - output_neurons : unsigned int : number of output neurons\n
    - elements : unsigned int : dimension of one attribute\n
    - max_neigh : unsigned int : how many neurons share the winning (on each side of the winner)\n
    - max_epoch : unsigned int : how many itterations to perform  \n
    - max_learn_rate : float : initial value of the learning rate \n
    - random_init : bool : type of initialization of the coefficients \n
    - rescale : bool : to rescale or not the input attributes \n

    @param des pointeur descripteur     
    @param ptp pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author Petre Rodan
    @version 1.0 (07/02/02); Include : csom.h
*/
param *csom_lect( csom_t *csom, param *ptp, char *debq)
{
  char question[500];
  //  unsigned int i;

  sprintf(question, "%s input_neurons (unsigned int)", debq); 
  lec_param(question, ptp);
  csom->input_neurons = (int)atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s output_neurons (unsigned int)", debq); 
  lec_param(question, ptp);
  csom->output_neurons = (int)atoi(ptp->rep);
  ptp = ptp->next;

//  sprintf(question, "%s size of input space (unsigned int)", debq); 
//  lec_param(question, ptp);
//  csom->element = (int)atoi(ptp->rep);
//  ptp = ptp->next;

  sprintf(question, "%s max_neigh (unsigned int)", debq); 
  lec_param(question, ptp);
  csom->max_neigh = (int)atoi(ptp->rep);
  while ((float)csom->max_neigh>(float)(csom->output_neurons/2.00))
    {printf("WARNING: max_neigh must be < %f   (output_neurons/2)",(float)(csom->output_neurons/2.00));
    sprintf(question, "%s max_neigh (unsigned int)", debq); 
    lec_param(question, ptp);
    csom->max_neigh = (int)atoi(ptp->rep);
    }
  ptp = ptp->next;

  sprintf(question, "%s max_learn_rate (float)", debq); 
  lec_param(question, ptp);
  csom->max_learn_rate = (float)atof(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s max_epoch (unsigned int)", debq); 
  lec_param(question, ptp);
  csom->max_epoch = (int)atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s random init or not? (1/0)", debq); 
  lec_param(question, ptp);
  csom->random_init = (int)atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s rescale the input space or not? (1/0)", debq);
  lec_param(question, ptp);
  csom->rescale_sw = (int)atoi(ptp->rep);
  ptp = ptp->next;
  
  return(ptp);
}



/* *************************  INITIALIZATION  *******************************/

/** mem allocations for input and coefficient space, \n
    init of neighborhood and learning functions.
    @param csom pointer to the structure of the neural network
    @author Petre Rodan
    @version 1.0 (08/03/02); Include : csom.h
*/
int csom_init(csom_t *csom)
{

  unsigned int i,j;

  /* malloc for the attribute space */
  if( (csom->x=(float **)malloc((csom->input_neurons)*sizeof(float*))) == NULL )
    {printf ("\n fatal error: a.13 allocation  impossible\n");  exit(1);}
  if( (csom->x[0]=(float *)malloc((csom->input_neurons)*(csom->element)*sizeof(float))) == NULL )
    {printf ("\n fatal error: a.14 allocation  impossible\n");  exit(1);}
  for (i=1;i<csom->input_neurons;i++)
    {csom->x[i]=csom->x[0]+((csom->element)*i);}

  /* mem allocation for the coefficients */
  if( (csom->w=(float **)malloc((csom->input_neurons)*sizeof(float*))) == NULL )
    {printf ("\n fatal error: a.01 allocation  impossible\n");  exit(1);}
  if( (csom->w[0]=(float *)malloc((csom->input_neurons)*(csom->output_neurons)*sizeof(float))) == NULL )
    {printf ("\n fatal error: a.02 allocation  impossible\n");  exit(1);}
  for (i=1;i<csom->input_neurons;i++)
    {csom->w[i]=csom->w[0]+((csom->output_neurons)*i);}

  /* mem allocation for new coefficients */
  if( (csom->w_new=(float **)malloc((csom->input_neurons)*sizeof(float*))) == NULL )
    {printf ("\n fatal error: a.03 allocation  impossible\n");  exit(1);}
  if( (csom->w_new[0]=(float *)malloc((csom->input_neurons)*(csom->output_neurons)*sizeof(float))) == NULL )
    {printf ("\n fatal error: a.04 allocation  impossible\n");  exit(1);}
  for (i=1;i<csom->input_neurons;i++)
    {csom->w_new[i]=csom->w_new[0]+((csom->output_neurons)*i);}

  /* mem allocation for distances */
  if( (csom->d=(float *)malloc((csom->output_neurons)*sizeof(float))) == NULL)
    {printf ("\n fatal error: a.05 allocation  impossible\n");  exit(1);}

  /* other mem hungry allocations */

#ifdef DEBUG
  if( (csom->activations=(int *)malloc((csom->output_neurons)*sizeof(int))) == NULL)
    {printf ("\n fatal error: a.06 allocation  impossible\n");  exit(1);}
#endif

  if( (csom->neigh_rate=(float *)malloc((csom->max_epoch)*sizeof(float))) == NULL)
    {printf ("\n fatal error: a.07 allocation  impossible\n");  exit(1);}
  if( (csom->learn_rate=(float *)malloc((csom->max_epoch)*sizeof(float))) == NULL)
    {printf ("\n fatal error: a.08 allocation  impossible\n");  exit(1);}
  if( (csom->csom_out=(int *)malloc((csom->element)*sizeof(int))) == NULL)
    {printf ("\n fatal error: a.09 allocation  impossible\n");  exit(1);}

  /* the neighborhood rate function */
  for (i=0;i<csom->max_epoch;i++) 
    {csom->neigh_rate[i]=ceil((float)-(csom->max_neigh*1.00/csom->max_epoch*1.00)*i*1.00 + csom->max_neigh);}
  /* DIV by 0 protection */
  csom->neigh_rate[csom->max_epoch-1]=1;

  /* the learning rate function */
  for (i=0;i<csom->max_epoch;i++) { csom->learn_rate[i]=csom->max_learn_rate*(pow((0.01/csom->max_learn_rate),((float)i/csom->max_epoch)));}

  /* the coefficients space init */
  //srand(0);
  srand((unsigned int) time(NULL));

  for (i=0;i<csom->output_neurons;i++)
    { 
      if (csom->random_init)
	{for (j=0;j<csom->input_neurons;j++)
	  {csom->w[j][i]=(int)(255 *(double)rand() / RAND_MAX);}
	}
      else
	{for (j=0;j<csom->input_neurons;j++)
	  {csom->w[j][i]=(float)i*254.00/csom->output_neurons;}
	}
    }
      
  return 0;
}

/* *************************  CALCUL  *******************************/
/** calculus using the Circular Self-Organizing Map nn, \n
    the output is found in csom->csom_out,
    @param csom pointer to the structure of the neural network
    @author Petre Rodan
    @version 1.0 (08/03/02); Include : csom.h
*/
int csom_calc(csom_t *csom)
{
  /** contors */
  int i,j;	  
  /** euclidian distances */
  float d_min; 
  /** the winner*/
  int winner=0;
  /** the distance from winner */
  int winner_df; 
  /** STOP condition flag */
  char keep_training; 
  /** a vector is a particular realization of the attribute space */
  int vector_no;
  /** distance left and right from the winner */
  int st,dr;
  /** rescaling variables */
  float min_input,max_input,rescale_factor;

  /* START */
  csom->epoch=0;
  keep_training=1;

  /* rescale the input space if needed*/
  if (csom->rescale_sw) 
    {
      for (j=0;j<csom->input_neurons;j++)
	{max_input=csom->x[j][0];
	min_input=csom->x[j][0];
	for (i=1;i<csom->element;i++) 
	  {if (csom->x[j][i]>max_input) {max_input=csom->x[j][i];}
	  if (csom->x[j][i]<min_input) {min_input=csom->x[j][i];}
	  }
	rescale_factor=255.00/(max_input-min_input);
	for (i=0;i<csom->element;i++) 
	  {csom->x[j][i]=csom->x[j][i]*rescale_factor;}
	}	
    }

  while (keep_training) /* STOP condition*/
    {
      //step 1. initializations
      //step 2. the current vector is applied to the input ( x(current_epoch) )
      //step 3. the distances between x(current_epoch) and every w(i,j) is computed 
      //step 4. the winning neuron is selected ( by establishing the smallest distance )
      //step 5. new w's are computed for the winning neuron and for the neighbours
      //step 6. current_epoch+=1, goto 2 ( until csom->max_epoch is reached )

      //printf("epoch %d\n",csom->epoch);
      csom->evolution=0;
#ifdef DEBUG   
      for (j=0;j<csom->output_neurons;j++) {csom->activations[j]=0;}
#endif
      for (vector_no=0; vector_no<csom->element; vector_no++)
        {d_min=99999999.99;
	csom->element_current=vector_no;
	//euclidian distances from input vector to the coefficients
	for (j=0;j<csom->output_neurons;j++)
	  {
	    csom->d[j]=0;
	    {for (i=0;i<csom->input_neurons;i++)
	      {csom->d[j]+=(csom->x[i][vector_no] - csom->w[i][j])*(csom->x[i][vector_no] - csom->w[i][j]);}
	    }
	    //who is the winner for this vector?
	    if (csom->d[j]<d_min) {d_min=csom->d[j]; winner=j;}
	  }
	//declare the winner
	csom->csom_out[csom->element_current]=winner;
#ifdef DEBUG
	csom->activations[winner]++;
#endif
	//share the winning
	 
	winner_df=0;
	st=winner;
	dr=winner;
	
	
	while (winner_df<csom->neigh_rate[csom->epoch])
	  {
	    for (i=0;i<csom->input_neurons;i++)
	      {
		/* refine the coefficients on the right side of the winner */
		csom->w_new[i][dr]=csom->w[i][dr]+csom->learn_rate[csom->epoch]*(1-(winner_df*1.000000001/csom->neigh_rate[csom->epoch]))*(csom->x[i][vector_no]-csom->w[i][dr]);
#ifdef DEBUG
		csom->evolution+=pow(csom->w_new[i][dr]-csom->w[i][dr],2);
#endif
		csom->w[i][dr]=csom->w_new[i][dr];

		/* refine the coefficients on the left side of the winner */
		csom->w_new[i][st]=csom->w[i][st]+csom->learn_rate[csom->epoch]*(1-(winner_df*1.000000001/csom->neigh_rate[csom->epoch]))*(csom->x[i][vector_no]-csom->w[i][st]);
#ifdef DEBUG
		csom->evolution+=pow(csom->w_new[i][st]-csom->w[i][st],2);
#endif
		csom->w[i][st]=csom->w_new[i][st];
	      }

	    if (st==0) {st=csom->output_neurons;}
	    if (dr==csom->output_neurons-1) {dr=-1;}

	    st--; dr++; winner_df++;
	  } /* done with the sharing */
        } /* all vector space computed */
      /* a new epoch */
      csom->epoch++;
#ifdef DEBUG
      printf("epoch %d\n",csom->epoch);
#endif
      /* the STOP condition */
      if (csom->epoch>=csom->max_epoch) {keep_training=0;}
    } //epoch

  /* karma += 1 */
  return 0;
}

//---------------------------------------------------------------------------
/* *************************  DESTRUCT  *******************************/
/** mem dealocations
    @param csom pointer to the structure of the neural network
    @author Petre Rodan
    @version 1.0 (08/03/02); Include : csom.h
*/
int csom_destruct(csom_t *csom)
{
  //  int i;

  //  for (i=0;i<csom->input_neurons;i++) {free(csom->x[i]);}
  free (csom->x[0]);
  free (csom->x);

  //free a.01 & a.02
  //for (i=0;i<csom->input_neurons;i++) {free(csom->w[i]);csom->w[i]=NULL;}
  free (csom->w[0]);
  free (csom->w);
  csom->w=NULL;

  //free a.03 & a.04
  //for (i=0;i<csom->input_neurons;i++) {free(csom->w_new[i]);csom->w_new[i]=NULL;}
  free (csom->w_new[0]);
  free (csom->w_new); csom->w_new=NULL;

  //free a.05
  free(csom->d); csom->d=NULL;

#ifdef DEBUG
  //free a.06
  free(csom->activations); csom->activations=NULL;
#endif

  //free a.07
  free(csom->neigh_rate); csom->neigh_rate=NULL;

  //free a.08
  free(csom->learn_rate); csom->learn_rate=NULL;

  //free a.09
  free(csom->csom_out); csom->csom_out=NULL;

  /*
  csom->input_neurons=-1;
  csom->output_neurons=-1;
  csom->element=-1;
  csom->element_current=-1;
  csom->epoch=-1;
  csom->max_epoch=-1;
  csom->evolution=-1;
  csom->max_neigh=-1;
  csom->rescale_sw=-1;
  */

  return 0;
}

//---------------------------------------------------------------------------
/* *************************  imau1tocsomvect  *******************************/
/** transforms a given number of imau1 images and places the data into a (float **)\\
    @param data - double pointer to a already allocated float
    @param nb_attr - number of imau1 images that are read
    @param ... - the imau1 images
    @author Petre Rodan
    @version 1.0 (08/03/02); Include : csom.h
*/
int imau1tocsomvect(float **data, int nb_attr, ...)
{

  int i,j,k;
  va_list ap;
  imau1 imag;

  va_start(ap, nb_attr);

  for (i=0; i<nb_attr; i++)
    {
      imag=va_arg(ap, imau1);
      for(j=0; j<imag.nr; j++)
	for(k=0; k<imag.nc; k++)
	  {data[i][k+j*imag.nc]=(float)imag.p[j][k]*1.00;}
    }

  va_end(ap);
  return 0;
}

//---------------------------------------------------------------------------
/* *************************  csomvecttoimau1   *******************************/
/** transforms a computed csom structure into an imau1 image
    @param csom - pointer to a csom structure that has already been through csom_calc(..)
    @param dimX - the number of columns of the resulting image
    @param dimY - the number of rows of the resulting image
    @param imres - pointer to the resulting image
    @author Petre Rodan
    @version 1.0 (08/03/02); Include : csom.h
*/
int csomvecttoimau1(csom_t *csom, int dimX, int dimY, imau1 *imres)
{
  unsigned char  *colormap; /* gray level for output pixels */
  int eticheta;
  int i,j;

  if( (colormap=(unsigned char *)malloc((csom->output_neurons)*sizeof(unsigned char))) == NULL)
    {printf ("\n fatal error: a.05 allocation  impossible\n");  exit(1);}

  /* if the output vector will be a 8bit picture, this is his colormap */
  //for (i=0;i<csom->output_neurons;i++) {csom->colormap[i]=(int)show_layer[i]*floor(i*254/(csom->output_neurons-1));}
  for (i=0;i<csom->output_neurons;i++) {colormap[i]=floor(i*254/(csom->output_neurons-1));}

  imres->nc = dimX;
  imres->nr = dimY;
  sprintf(imres->nom, "csom_out");

  alloc_imau1(imres);

  for(j=0; j<imres->nr; j++)
    for(i=0; i<imres->nc; i++)
      {eticheta=csom->csom_out[i+j*imres->nc]; 
      imres->p[j][i] = colormap[eticheta]; 
      }
	
  free(colormap);
  colormap=NULL;
  return 0;
}

//---------------------------------------------------------------------------
/** @} */  
