/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ************************************************************* */
/*              ldefuzzyfication_voisinage_3x3.c                 */
/* ************************************************************* */
/*                                                               */
/*      descripteurs et prototypes associs  l'oprateur        */
/*                  defuzzyfication_voisinage_3x3                */
/*                                                               */
/*  Dernire modif : Ramasso E. & Jullien S.  08/12/03           */
/*                                                               */
/* ***************************************************************/

/**
\ingroup classif
\defgroup defuzz defuzzification par voisinage 3x3
*/

#include "classifier.h"
#include "defuzzyfication_voisinage_3x3.h"
#include "image.h"
#include "proto2D.h"
#include <stdlib.h>
#include <stdio.h>



/** 
 *   \ingroup defuzz
 *   @name Oprateur
 * \n
 * C'est oprateur dfuzzifie de la manire suivante : pour chaque point d'un vecteur 
 * de donnes de type data_output faire :
 * - si toutes les valeurs sont diffrentes sur chaque vecteur alors c'est le maximum 
 * qui est pris.
 * - sinon il existe des points incertains et une fentre 3x3 est utilise pour 
 * dfuzzifier en fonction du voisinage et la moyenne dans la fentre est affecte au 
 * point incertain considr. 
 * 
 */
//@{


/* *************************  LECTURE  *******************************/
/*! \brief Fonction lisant les paramtres de l'oprateur defuzzyfication_voisinage_3x3
 *
 *   Comme nous n'avons besoin d'aucun paramtre en dehors de ceux contenus dans les structures de sortie  dfuzzifier, nous ne ferons rien dans l'opration de lecture.
 *
 *  \param desc pointeur descripteur     
 *  \param ptp pointeur sur paramtre courant dj allou (a lire),     
 *  \param debq pointeur de chane de caractres place en debut de question
 *  \return renvoie nouveau pointeur de paramtres dj allou (pour la suite) 
 *  \author Premire version: F. Florea
 *  \author Seconde version: E. Ramasso & S. Jullien
 *  \version 01/12/03 
*/
param *defuzzyfication_voisinage_3x3_lect(defuzzyfication_voisinage_3x3_t *desc, param *ptp, char *debq)
{

  /* Rien besoin de faire pour la lecture car par supposition, cet oprateur est reli  un classifieur */  

  return ptp;
}





/* *************************  INITIALISATION  ***************************/
/*! \brief Fonction initialisant l'oprateur defuzzyfication_voisinage_3x3
 *
 *    
 *    Initialisation dans la structure de sortie:
 *     - du nombre de classes
 *     - du nombre de points
 *     - du type de donnes de sortie
 *     - du nombre de lignes
 *     - du nombre de colonnes
 *
 *    Allocation des matrices qui contiendront la somme des coefficients d'appartenance et la position des pixels incertains.
 *
 *    Nous mettons  zro la matrice des positions des pixels incertains et les champs du descripteur suivant
 *    - nb_data_incertaines_finales
 *    - nb_data_incertaines_initiales
 *
 *    
 *    \param desc pointeur descripteur
 *    \param dataOF structure contenant les donnes de sortie floue
 *    \param dataOC pointeur sur la structure contenant les donnes de sortie nettes
 *    \return renvoie un entier qui indique si l'opration s'est bien droule 
 *    \author Premire version: F. Florea
 *    \author Seconde version: E. Ramasso & S. Jullien
 *    \version 01/12/03
*/
int defuzzyfication_voisinage_3x3_init(defuzzyfication_voisinage_3x3_t *desc, data_output dataOF, data_output *dataOC )
{
  int i;

  printf("\nDefuzzyfication init...");

  /* on remplie les champs de la structure de sortie */
  desc->nb_classes = dataOF.nb_class;
  dataOC->nb_class = desc->nb_classes;
  dataOC->nb_pts = dataOF.nb_pts;
  dataOC->type = CRISP_CHOICE;
  dataOC->equ_nb_cols = dataOF.equ_nb_cols;
  dataOC->equ_nb_rows = dataOF.equ_nb_rows;
  
  /*on alloue une structure de sortie */
  alloc_classifier_data_output(dataOC); 
  
  /* allocation des variables */
  if((desc->somme_coeff_appartenance=(float*)malloc(desc->nb_classes*sizeof(float)))==NULL ||
     (desc->position_data_incertaines=(int*)malloc((dataOF.nb_pts)*sizeof(int)))==NULL)
    {
      printf("Erreur defuzz init memoire insuffisante\n");
      exit(1);
    }

  
  /* remplir desc->position_data_incertaines avec 0 car le malloc ne le fait pas */
  for(i=0;i<dataOF.nb_pts;i++)
      desc->position_data_incertaines[i] = 0;

  
  /* initialization nb. de pixels incertains */
  desc->nb_data_incertaines_finales = 0; 
  desc->nb_data_incertaines_initiales = 0; 
  
  printf("OK\n");

}



/* *************************  CALCUL  ***************************/
/*!
    \brief Fonction calculant le resultat de l'oprateur defuzzyfication_voisinage_3x3

    Rgle de dfuzzyfication :
 
    Pour chaque pixel inspect(chaque coordonne),il faut:
    - trouver la valeur maximale de chaque pixel en fonction des classes 
    - le maximum donnera la classe  laquelle appartient le pixel
    - si il y a plus d'un maximum, une inspection dans un voisinage 3x3 est effectue. Elle consiste en la somme-pondere de tous ces pixels
 
    Nous allons considrer que le pixel courant appartient  la classe ayant la somme maximale.

    \param desc  pointeur descripteur
    \param dataOF structure contenant les donnes de sortie floue
    \param dataOC pointeur sur la structure contenant les donnes de sortie nettes
    \author Premire version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
    \version 01/12/03 
*/
int defuzzyfication_voisinage_3x3_calc(defuzzyfication_voisinage_3x3_t *desc, data_output dataOF, data_output *dataOC )
{

  int i, j, x, y, indice, compteur, max_multiple;
  
  printf("\nDefuzzyfication calcul...");

  for(i=0;i<dataOF.nb_pts;i++)
    {
      desc->max_coeff_appartenance = 0;
      indice = 0;
      for(j=0;j<desc->nb_classes;j++) 
	if (dataOF.classes.fuzzy[j][i]>desc->max_coeff_appartenance) 
	  {
	    desc->max_coeff_appartenance=dataOF.classes.fuzzy[j][i];  
	    /* trouve le coeff. d'appartenance desc->max_coeff_appartenance.  */
	    indice=j;
	  }
      
      max_multiple=0;
      for(j=0;j<desc->nb_classes;j++)
	if(dataOF.classes.fuzzy[j][i]==desc->max_coeff_appartenance) 
	  (max_multiple)++; /* calculer le nb de desc->max_coeff_appartenance. */
      
      /* Les pixels du bord de l'image ne sont pas pris en compte dans la vrification dans la fentre 3x3 */
      if((i<dataOF.equ_nb_cols)||
	 (i%(dataOF.equ_nb_cols) == 0)|| 
	 ((i+1)%(dataOF.equ_nb_cols) == 0) || 
	 (i>(dataOF.nb_pts - dataOF.equ_nb_cols))) 
	dataOC->classes.crisp[i] = indice;                
      
      else if (max_multiple == 1) /* vrification si il y a seulement un max */
	dataOC->classes.crisp[i] = indice; /* si oui -> defuzzyfication  */
      
      else 
	{
	  /* si non il y en a plusieurs, sauvegarder la position de ce pixel incertain*/
	  desc->position_data_incertaines[i] = 1; 

	  /* incrementer nb. de pixels incertains */
	  desc->nb_data_incertaines_initiales++; 
	}
    }
  
  printf("Nb. de donnes incertaines initiales = %d.",desc->nb_data_incertaines_initiales);
 
  if(desc->nb_data_incertaines_initiales == 0) 
    {
      printf("Donc pas besoin de fentre de vrification.\n");
      return 0;
    }

  else if(desc->nb_data_incertaines_initiales > 0)
    {
      printf(" Incorrect. \nNcessit d'un post traitement des donnes incertaines avec fentre 3x3...\n");
      
      /* traitement des pixels incertains  */ 
      desc->nb_data_incertaines_finales = 0;
      
      for(i=0;i<dataOF.nb_pts;i++)
	if (desc->position_data_incertaines[i] == 1)
	  {
	    for(j=0;j<desc->nb_classes;j++)
	      {
		desc->somme_coeff_appartenance[j]=0;
		for(x=-1;x<2;x++)
		  for(y=-1;y<2;y++)
		    {
		      /* calculer les coefficients d'appartenence  */
		      desc->somme_coeff_appartenance[j] = desc->somme_coeff_appartenance[j]
			                                + dataOF.classes.fuzzy[j][i+x*(dataOF.equ_nb_cols)+y]; 
		      /* moyenne, dans un voisinage 3x3 */
		    }
	      }
	    
	    desc->max_coeff_appartenance = 0;
	    indice = 0;
	    
	    for(j=0;j<desc->nb_classes;j++)
	      {
		if (desc->somme_coeff_appartenance[j] > desc->max_coeff_appartenance) 
		  {
		    desc->max_coeff_appartenance = desc->somme_coeff_appartenance[j];
		    indice=j;
		  }
	      }
	    
	    compteur=0;
	    
	    for(j=0;j<desc->nb_classes;j++)
	      if (desc->max_coeff_appartenance == desc->somme_coeff_appartenance[j]) compteur++;
	    
	    if (compteur > 1) desc->nb_data_incertaines_finales++;
	    
	    dataOC->classes.crisp[i] = indice;
	  }
      
      if(desc->nb_data_incertaines_finales > 0) 
	printf("\nAprs la vrification avec la fentre 3x3 il y a encore %d incertitudes.\n",desc->nb_data_incertaines_finales);
      
      else printf("\nLa fentre 3x3 a supprim les incertitudes\n");
      
      return 0;
    }

  else 
    {
      printf("Erreur defuzz calc : nb de data < 0\n");
      exit(1);
    }

  printf("OK.\n");

}
//@}


