/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ****************************************************************** */
/*                             lfuzzy_c_means.c                                 */ 
/* ****************************************************************** */
/*                                                                    */
/*            Classifictation du type Fuzzy C-mean                    */
/*                                                                    */
/* ****************************************************************** */
/*  Dernire modification: Ramasso E. & Jullien S. 08/12/03           */
/*                                                                    */
/* ****************************************************************** */


/**
\ingroup classif
\defgroup fcm fuzzy C mean
*/


#include "classifier.h"
#include "image.h"
#include "proto2D.h"
#include "fuzzy_c_means.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>



#define ABS(X)		((X)<0?-(X):(X))

/**
 *  \ingroup fcm
 *   @name Oprateur
 * \n
 * Cet oprateur de classification (algorithme propos par Bezdek en 1984), 
 * permet de classifier des donnes de faon automatique, sans apprentissage. 
 * C'est une mthode qui utilise le partitionnement itratif d'ensemble de 
 * donnes. \n
 * Il prend pour cela des donnes formates en input_data en entre et ressort,  
 * en fonction des paramtres rentrs par l'utilisateur lors de la lecture de
 * l'oprateur, des donnes de type output_data.
 */
//@{

/* *************************  LECTURE  *******************************/
/*! \brief Fonction lisant les paramtres de l'oprateur fcm
 *
 *    Lecture des paramtres de l'oprateur fcm:
 *   - nb_class : nombre de classes de sortie
 *   - fuzzy_coeff : l'indice de flou
 *   - threshold : seuil d'arret (reel: [0, 1])
 *   - nb_it : le nombre maximum d'itrations
 *   - extension : extension noms fichiers des images d'appartenance 
 * (.ima|.ras|.imf)
 * 
 *   
 *  \param fcm pointeur descripteur     
 *  \param ptp pointeur sur paramtre courant dj allou (a lire)     
 *  \param debq pointeur de chane de caractres place en debut de question
 *  \return renvoie nouveau pointeur de paramtres dj allou (pour la suite) 
 *  \author Premiere version: F. Florea
 *  \author Seconde version: E. Ramasso & S. Jullien
 *  \version 01/12/03 
*/
param *fuzzy_c_means_lect( fuzzy_c_means_t *fcm, param *ptp, char *debq)
{
  char question[500];

  sprintf(question, "%s nombre de classes [nb_class] (entier)", debq); 
  lec_param(question, ptp);
  fcm->nb_class = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s indice de flou [fuzzy_coeff] (>1) (reel)", debq); 
  lec_param(question, ptp);
  fcm->fuzzy_coeff = (float)atof(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s seuil d'arret [threshold] [0, 1] (ex:0.001) (reel)", debq); 
  lec_param(question, ptp);
  fcm->threshold = (float)atof(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s nombre maximum d'iterations [nb_it] (entier)", debq); 
  lec_param(question, ptp);
  fcm->nb_it = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s extension noms fichiers images d'appartenance (.ima|.ras|.imf), non sauvees si ''", debq); 
  lec_param(question, ptp);
  strcpy(fcm->extension, ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s Quelle initialisation? ( 0 | 1 | nom_fichier )", debq); 
  lec_param(question, ptp);
  strcpy(fcm->init_type, ptp->rep);
  ptp = ptp->next;

  return(ptp);
}





/* *************************  INITIALISATION  ***************************/
/*! \brief Fonction initialisant l'oprateur fcm
 *   
 *    Initialisation dans la structure de sortie:
 *     - du nombre de classes
 *     - du nombre de points
 *     - du type de donnes de sortie
 *     - du nombre de lignes
 *     - du nombre de colonnes
 *
 *    Allocation de la matrice contenant les centodes
 *
 *    
 *    \param fcm  pointeur descripteur
 *    \param dataI structure contenant les donnes d'entre
 *    \param dataO pointeur sur la structure contenant les donnes de sortie
 *    \return renvoie un entier qui indique si l'operation s'est bien droule 
 *    \author Premiere version: F. Florea
 *    \author Seconde version: E. Ramasso & S. Jullien
 *    \version 01/12/03
*/

int fuzzy_c_means_init(fuzzy_c_means_t *fcm, data_input dataI , data_output *dataO)
{
  int i, k, j;

  printf("\n fuzzy_c_mean init...");

  dataO->nb_class = fcm->nb_class;
  dataO->nb_pts = dataI.nb_pts;
  dataO->type = FUZZY_CHOICE;
  dataO->equ_nb_cols = dataI.equ_nb_cols;
  dataO->equ_nb_rows = dataI.equ_nb_rows;

   fcm->centre=(unsigned short **)Calloc(fcm->nb_class,sizeof(*fcm->centre));
   for (k=0;k<fcm->nb_class;k++)
       fcm->centre[k]=(unsigned short *)Calloc(dataI.nb_attr,sizeof(unsigned short));
 
   alloc_classifier_data_output(dataO);

   /* appel init */
   init_coeff_appartenance(fcm, dataI, dataO);

   printf("\ninit fcm OK.\n");
}





/* *************************  CALCUL  ***************************/
/*!
    \brief Fonction calculant le resultat de l'oprateur fcm
   
    Calcul le degr d'appartenance  chaque classe en fonction des centroides. \n
    Il recommence l'opration tant que le rsultat ne se trouve pas en dessous d'un 
    seuil fix par l'utilisateur.

       
    \param fcm  pointeur descripteur
    \param dataI structure contenant les donnes d'entre
    \param dataO pointeur sur la structure contenant les donnes de sortie
    \return renvoie un entier qui indique si l'opration s'est bien droule 
    \author Premiere version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
    \version 01/12/03
*/
int fuzzy_c_means_calc(fuzzy_c_means_t *fcm, data_input dataI , data_output *dataO)
{ 
  int i,j,k,t;
  float media,tmp;    
  
  printf("\nfcm calc...\n\n"); 

  t=0;
  printf("Nb. de iterations[nb_it]=%d\n",fcm->nb_it);
  printf("\n");
  while(t<fcm->nb_it)
      {
	  printf("\n");
	  printf(" Itration numro %d\n",t);
	  printf("\n");

	  calcul_centroides(fcm,dataI, dataO);
	  calcul_coeff_appartenance(fcm, dataI, dataO);

	  media=fcm->variation/(fcm->nb_class*(dataI.nb_pts));
	  printf("===> seuil = %1.5f",fcm->threshold);
	  printf("   --->>    la variation actuelle = %1.5f",media);
          if (media<fcm->threshold) break;

	  if (t<fcm->nb_it-1) printf("  ...  donc, continue\n");
	  
	  t++;	  
      }

 
  printf("\n");
  printf("\n");

  printf("Calcul effectu!\n");
  if (t!=fcm->nb_it) printf("L'algorithme a termin le calcul aprs  %d cycles.\n",t+1);
  else printf("\nNb. maximum d'itrations atteint !!! \n");

  printf("\n");
}

//@}


/* Ces fonctions servent  effectuer le calcul.
Elles sont utilises dans la fonction fcm_calc. */

/* **************************  FUNCTIONS  ****************************** */
/* *************************  ALLOCATION  ************************** */
/*
    \brief Fonction d'allocation

        
    \param nelem correspond au nombre d'lments que l'on veut allouer
    \param elsize correspond a la taille des lments  allouer
    \return renvoie un caractre qui indique si l'opration s'est bien droule
    \author Premiere version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
    \version 01/12/03
*/
char *Calloc(unsigned nelem,unsigned elsize)
{
char *ptr;
if((ptr=calloc(nelem,elsize))==NULL)
     {
     perror("MEMORY ALLOCATION ERROR");
     exit(3);
     }
return ptr;
}


/* *************************  APPARTENANCE  ***************************/
/*
    \brief Fonction d'appartenance

    Cette fonction recalcule les coefficients d'appartenance

    \param fcm le descripteur du classifieur fuzzy C mean
    \param dataI structure contenant les donnes d'entre
    \param dataO pointeur sur la structure contenant les donnes de sortie
    \author Premiere version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
   \version 11.0 (01/12/03); Include : proto2D.h, fcm.h
*/
void calcul_coeff_appartenance(fuzzy_c_means_t *fcm, data_input dataI , data_output *dataO)
{
  int i,j,k,l,x,y,z,a,b;
  float S,S1,S2,S3;

  fcm->variation=0; 
  for(a=0;a<dataI.nb_pts;a++)
    for(i=0;i<fcm->nb_class;i++)
      {
	S3=S1=S2=S=0;
	
	for(x=0;x<dataI.nb_attr;x++)
	  S1=S1+pow((dataI.attributes[x][a]-fcm->centre[i][x]),2);
	
	S3=1/pow(S1+0.001,1/(fcm->fuzzy_coeff-1));
	
	for(j=0;j<fcm->nb_class;j++)
	  {
	    for(y=0;y<dataI.nb_attr;y++)
	      S2=S2+pow((dataI.attributes[y][a]-fcm->centre[j][y]),2);
	    
	    S2=1/pow(S2,1/(fcm->fuzzy_coeff-1));
	    S=S+S2;
	  }
	
	fcm->variation=fcm->variation+ABS(dataO->classes.fuzzy[i][a]-(S3/S))*10;
	dataO->classes.fuzzy[i][a]=S3/S;
      } 

}



/* *************************  CENTROIDES  ***************************/
/*
    \brief Fonction d'appartenance

    Cette fonction recalcule les centrodes
  
    \param fcm le descripteur du classifieur fuzzy C mean
    \param dataI structure contenant les donnes d'entre
    \param dataO pointeur sur la structure contenant les donnes de sortie
    \author Premiere version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
    \version 01/12/03
*/
void calcul_centroides(fuzzy_c_means_t *fcm, data_input dataI , data_output *dataO)
{
  int i,j,k,l;
  float S1,S2;
  
  for(j=0;j<dataI.nb_attr;j++)
    {
     for(i=0;i<fcm->nb_class;i++)
	 {
	 S1=S2=0;
	 for(k=0;k<dataI.nb_pts;k++)
	   {
	     S1=S1+pow(dataO->classes.fuzzy[i][k],fcm->fuzzy_coeff)*dataI.attributes[j][k];
	     S2=S2+pow(dataO->classes.fuzzy[i][k],fcm->fuzzy_coeff);
	   }
	 fcm->centre[i][j]=(int)(S1/S2);
	 printf("centre[%d][%d] = %d\n",i,j,fcm->centre[i][j]);
	 }
    printf("\n");
    }
}



/* *************************  INITIALISATION CENTRODES  ***************/
/*
    \brief Fonction d'initialisation des centroides

    Cette fonction initialise la valeur des centrodes.

    
    \param fcm le descripteur du classifieur fuzzy C mean
    \param dataI structure contenant les donnes d'entre
    \param dataO pointeur sur la structure contenant les donnes de sortie
    \author Premiere version: F. Florea
    \author Seconde version: E. Ramasso & S. Jullien
    \version 01/12/03
*/

void init_coeff_appartenance(fuzzy_c_means_t *fcm, data_input dataI , data_output *dataO)
{
  int tmp;
  int i,j,k,x;
  float *a;
  int max,rnd;
  FILE *toto;
  
  if ((strlen(fcm->init_type)==1)&&( (strstr(fcm->init_type,"0")!=NULL) || (strstr(fcm->init_type,"1")!=NULL) ))
    {
      printf("...Choix d'initialisation = %d\n",atoi(fcm->init_type));
      
      /* si init_type=0 => initialisation tableaux avec les coef. d'appartenance directement random */
      if (atoi(fcm->init_type) == 0)
	{
	  srand((unsigned int)time(NULL));
	  for(i=0;i<dataI.nb_pts;i++)
	    {
	      max=0;
	      for(k=0;k<fcm->nb_class;k++)
		{
		  dataO->classes.fuzzy[k][i]=(rand()%(100));
		  max=max+dataO->classes.fuzzy[k][i];
		}
	      
	      for(k=0;k<fcm->nb_class;k++)
		if (max!=0) dataO->classes.fuzzy[k][i]=(dataO->classes.fuzzy[k][i])/max;
	    }		 
	}
      
      /* si init_type=1 => initialisation centre des classes aleatoires, et ensuite calcul des 
	 tableaux avec coeff. d'appartenance directement */
      else if (atoi(fcm->init_type) == 1)
	{
	  a=(float *)Calloc(fcm->nb_class,sizeof(float));
	  
	  /* calcul max dans l'image d'entre */
	  max=0;
	  for(k=0;k<dataI.nb_attr;k++)
	    for(i=0;i<dataI.nb_pts;i++)
	      if (dataI.attributes[k][i]>max) max=dataI.attributes[k][i];
	  
	  
	  /* initialisation des centrodes aleatoires */
	  for(x=0;x<dataI.nb_attr;x++)
	    for(i=0;i<fcm->nb_class;i++)
	      fcm->centre[i][x]=rand()%(max);
	  
	  
	  /* initialisation des coefficients d'appartenance  partir des centrodes */
	  for(x=0;x<dataI.nb_attr;x++)
	    for(i=0;i<dataI.nb_pts;i++)
	      {
		max=0;
		for(k=0;k<fcm->nb_class;k++)
		  {
		    a[k]=ABS(fcm->centre[k][x]-dataI.attributes[x][i]);
		    max=max+a[k];
		  }
		for(k=0;k<fcm->nb_class;k++)
		  dataO->classes.fuzzy[k][i]=a[k]/max;
	      }
	  
	}
    }
  
  /* nom_fichier = initialisation centre des classes avec l'information lue dans un fichier ".txt" */
  else
    {
      printf("...initialisation par fichier = lecture dans '%s' les coeff suivants :\n\n",fcm->init_type);
      /* allocation */
      a=(float *)Calloc(fcm->nb_class,sizeof(float));

      /* ouvrir fichier */
      toto=fopen(fcm->init_type,"r");
      /* recuperation */
      for(i=0;i<fcm->nb_class;i++)
	{
	  for(j=0;j<dataI.nb_attr;j++)
	    {
	    fscanf(toto,"%d ", &tmp);
	    fcm->centre[i][j] = (unsigned short)tmp;
	    printf("centre[%d][%d] = %d \t",i,j,fcm->centre[i][j]);
	  }   
	printf("\n");
	}

      /* fermeture fichier */
      fclose(toto);
      
      /* initialisation des coefficients d'appartenance  partir des centrodes */
      for(x=0;x<dataI.nb_attr;x++)
	for(i=0;i<dataI.nb_pts;i++)
	  {
	    max=0;
	    for(k=0;k<fcm->nb_class;k++)
	      {
		a[k]=ABS(fcm->centre[k][x]-dataI.attributes[x][i]);
		max=max+a[k];
	      }

	    for(k=0;k<fcm->nb_class;k++)
	      dataO->classes.fuzzy[k][i]=a[k]/max;
	  }
    }
  
  /* initialisation complte alatoire */
}





 




