/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ************************************************************* */
/*                    lconfusion_matrix.c                        */
/* ************************************************************* */
/*                                                               */
/*          Calcul de la matrice de confusion                    */
/*                                                               */
/*  Dernière modif : Jullien S.  21/03/07                        */
/* ************************************************************* */

/** \defgroup confusion_matrix

Découpage de l'opérateur en 3 fonctions :

	- lecture des paramètres du calcul :
	param *confusion_matrix_lect(confusion_matrix_t *desc, param *ptp, char *debq);

	- initialisation des valeurs des paramètres et de l'emplacement mémoire:
	void confusion_matrix_init(confusion_matrix_t *desc);

	- calcul de la matrice de confusion stocké dans matrice:
	void confusion_matrix_calc(confusion_matrix_t *desc,  int **matrice);



Fonctionnement :

	La matrice de confusion évalue la performance d'un résultat de classification 
	en fonction d'un ensemble de points d'évaluation. Ce programe traite uniquement 
	des résultats sous la forme d'images en 2D ou en 3D. La matrice se présante sous 
	la forme suivante:

	matrice[réalité][prédiction]
		|A prédit |B prédit
	A réel  |   15    |   2
	B réel  |   4     |   16

	15 points ont été correctement classifiés dans la classe A, 16 points ont correctement 
	été classifiés dans la classe B. 4 points ont été classifiés dans la classe A alors
	qu'ils appartenaient réellement à la classe B. 2 points ont été classifiés dans 
	la classe B alors qu'ils appartenaient à la classe A.

@name lconfusion_matrix
@author Sylvie JULLIEN  
@version 1.0 (21/03/07); Include : confusion_matrix.h

*/

#include <stdio.h>
#include "image.h"
#include "proto2D.h"
#include "classifier.h"
#include "confusion_matrix.h"



/** \ingroup confusion_matrix
    lecture des parametres du calcul de la matrice de confusion. \\

    @param desc1 : pointeur sur descripteur\\
    @param ptp pointeur sur parametre courant deja alloue (a lire),\\
    @param debq pointeur de chaine de caracteres placee en debut de question\\
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite)\\

    @author JULLIEN Sylvie 
    @version 1.0 (21/03/07); Include : confusion_matrix.h

*/
param *confusion_matrix_lect(confusion_matrix_t *desc, param *ptp, char *debq)
{

  char question[500];
  int i;
  char fich[500];


  //Nombre de régions
  sprintf(question, " Nombre de region a etudier: ");
  lec_param(question, ptp);
  desc->nb_region_etudiee= atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, " Nom du fichier contenant l'image à évaluer: ");
  lec_param(question, ptp);
  strcpy(fich,ptp->rep);
  ptp = ptp->next;

  lect_ima3Du1(fich,&(desc->imadec),ALL);
  
  //Type de seuillage
  sprintf(question, " Image de test 2D (0) ou 3D (1): ");
  lec_param(question, ptp);
  desc->type_eval= atoi(ptp->rep);
  ptp = ptp->next;

  switch(desc->type_eval)
    {
    case 0:


      printf("\n Image contenant l'ensemble de test \n");
      ptp = read_ima_lect (&(desc->read_ima), ptp, "");
      read_imau1_init (&(desc->read_ima),&(desc->imaeval_2D));

      sprintf(question, " Numero de la coupe: ");
      lec_param(question, ptp);
      desc->num_coupe= atoi(ptp->rep);
      ptp = ptp->next;


      break;

    case 1:

      sprintf(question, " Nom du fichier contenant l'ensemble de test: ");
      lec_param(question, ptp);
      strcpy(fich,ptp->rep);
      ptp = ptp->next;
      
      lect_ima3Du1(fich,&(desc->imaeval_3D),ALL);
  
      break;

    }


  return(ptp);

}

/** \ingroup confusion_matrix
    initialisation de l'operateur confusion_matrix.\\
    Rééchelonnement des valeurs contenues dans la prédiction sur une échelle 1,2,3...
    Inutile si l'utilisateur a déjà effectué la remise à l'échelle de la prédiction\\
\\

    @param desc : pointeur sur descripteur
    @return 

    @author JULLIEN Sylvie 
    @version 1.0 (21/03/07); Include : confusion_matrix.h

*/


//Fonction d'initialisation de l'opérateur threshold

void confusion_matrix_init(confusion_matrix_t *desc)
{
  int i;
  int x,y,z;

      for(z=0;z<desc->imadec.dimz;z++)
	for(y=0;y<desc->imadec.dimy;y++)
	  for(x=0;x<desc->imadec.dimx;x++)
	    {
	      desc->imadec.data[z][y][x]=(int)(desc->imadec.data[z][y][x]*desc->nb_region_etudiee/255);
	    }
}

/** \ingroup confusion_matrix
    fonction calcul de l'operateur confusion_matrix.\\
    Calcul de la matrice de confusion
    la place mémoire pour stocker la matrice de confusion doit \^etre allouer dans le programme principal

    @param desc : pointeur sur descripteur
    @param matrice : pointeur sur la matrice de confusion
    @return 

    @author JULLIEN Sylvie 
    @version 1.0 (21/03/07); Include : confusion_matrix.h
*/

//Fonction de calcul

void confusion_matrix_calc(confusion_matrix_t *desc,  int **matrice)
{

  float max;
  int num,x,y,z,r,i,nb;
  float somme;


  //En fonction du type d'image de test
  switch(desc->type_eval)
    {
      
    case 0:
      //l'image de test est en 2D
      
      z=desc->num_coupe;
      for(y=0;y<desc->imadec.dimy;y++)
	for(x=0;x<desc->imadec.dimx;x++)
	  {
	    if(desc->imaeval_2D.p[x][y]<=desc->nb_region_etudiee && desc->imadec.data[z][y][x]<=desc->nb_region_etudiee)
	      matrice[desc->imaeval_2D.p[x][y]][desc->imadec.data[z][y][x]]++;
	  }
      
      break;
      
      
    case 1:
      
      //l'image de test est en 3D
      
      for(z=0;z<desc->imadec.dimz;z++)
	for(y=0;y<desc->imadec.dimy;y++)
	  for(x=0;x<desc->imadec.dimx;x++)
	    {
	      if(desc->imaeval_3D.data[z][y][x]<=desc->nb_region_etudiee && desc->imadec.data[z][y][x]<=desc->nb_region_etudiee)
		matrice[desc->imaeval_3D.data[z][y][x]][desc->imadec.data[z][y][x]]++;
	    }
      break;           
      
    }
}



/** \ingroup confusion_matrix
    fonction fermeture de l'operateur confusion_matrix.\\
    Désallocation de la mémoire pour l'image de prédiction et l'image test

    @param desc : pointeur sur descripteur
    @return 

    @author JULLIEN Sylvie 
    @version 1.0 (21/03/07); Include : confusion_matrix.h
*/
//Fonction de fermeture

void confusion_matrix_ferm(confusion_matrix_t *desc)
{
  int i;

  liberer_ima3Du1(&(desc->imadec));


  switch(desc->type_eval)
    {
    case 0:

      free_imau1 (&(desc->imaeval_2D));

      break;
       
    case 1:
      
      liberer_ima3Du1(&(desc->imaeval_3D));
      
      break;
    }


}





