/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* **************************** lcontourRSO.c ****************************/
/*                                                                    */
/* ********************************************************************/

/** \ingroup detection
    \defgroup contourRSO detection de contour
    operateur contourRSO : detection de contours dans les images radar.\\
Detecteur de contour type Touzi : min ratio entre deux zones (moyenne des intensites). \\
2 images resultats : le rapport [0, 1] et l'orientation [0, 2pi[, sens des aiguilles d'une montre, 
(=trigo + si axe des Y vers le bas), codee sur [1, 256[, 0 pour le masque. \\
Possibilte de selection des contours max locaux dans la direction normale.\\
Les points ecartes (ou non-calcules au bord) ont un rapport=1 et une orientation=0 \\
\\
Descripteur : \\
    char dim : taille fenetre (impair) \\
    char qsauver ; sauvegarde de masques de directions (1/0) \\
    char maxloc ; selection des contours max locaux dans la direction normale \\
    imau1 msqv, msqh, msqdp, msqds : les masques de direction \\
    imafl rapv, raph, rapdp, rapds : buffer 3 lignes des rapports \\
\\
Type entree : imau2 \\
Type sortie : imau1/imafl pour le ratio code 0-255/0-1\\
Type sortie : imau1 pour la direction code 1-255 pour 0-2pi \\

@name lcontourRSO
@author F. Bujor, E. Trouve
@version 1.0 (15/06/00); Include : image.h, proto2D.h, contour.h
@see lcontourRSO
*/

#include "image.h"
#include "proto2D.h"
#include "contourRSO.h"
/* #include "limits.h" */
#include <stdlib.h>
#include <string.h>

#define MININV(a) ( (a) <= 1. ? (a) : 1./(a) ) 
#define FLT_MAX         3.402823466E+38F  /* max decimal value of a "float" */

/* **************** fonctions locales ********************* */

/* application d'un masque, retourne rapport moy1/moy2 ******** */
float appl_masque12(imau2 im0, imau1 msq, int i, int j, float *tab){
  int k, l, m, n, dim;
  float tmp1, tmp2;
  float rap;

  tmp1 = 0;
  tmp2 = 0;
  dim = msq.nr;
  for(m=-(dim/2);m<(dim/2)+1;m++)
      for(n=-(dim/2);n<(dim/2)+1;n++){
	      if(msq.p[m+(dim/2)][n+(dim/2)] == 1){
		  tmp1 += tab[im0.p[i+m][j+n]];
	      }
	      if(msq.p[m+(dim/2)][n+(dim/2)] == 2){
		  tmp2 += tab[im0.p[i+m][j+n]];
	      }
	  }
  if(tmp2 == 0.){
      if(tmp1 == 0.)
	  rap = 1.;
      else
	  rap = FLT_MAX;
  }
  else
      rap = tmp1/tmp2;
  return(rap);
}

/* selection de la direction [0|1|32|64 ...  224] et la valeur [0, 1] du ratio */ 
void valdirmax(float v, float h, float dp, float ds, float *ptval, pixu1 *ptdir){
    float tmp;

    if(v < 1.){
	*ptval = v;
	*ptdir = 128;
    }
    else if (v > 1.){
	*ptval = 1./v;
	*ptdir = 1;	
    }
    else{
	*ptval = 1.;
	*ptdir = 0;
    }

    tmp = MININV(h);
    if(tmp < *ptval){
	*ptval = tmp;
	if(h<1.)
	    *ptdir = 192;
	else
	    *ptdir = 64;
    }

    tmp = MININV(dp);
    if(tmp < *ptval){
	*ptval = tmp;
	if(dp<1.)
	    *ptdir = 224;
	else
	    *ptdir = 96;
    }

    tmp = MININV(ds);
    if(tmp < *ptval){
	*ptval = tmp;
	if(ds<1.)
	    *ptdir = 160;
	else
	    *ptdir = 32;
    }
}

/* test si contour est max local dans la direction du gradient, retourne 1 si oui, 0 sinon */
int maxlocgrad(float val, float dir, int j, contourRSO_t *des){
    float v1, v2;

    if(dir == 0)
	return(0);
    else if( (dir == 1) || (dir == 128) ){ /* coutour vertical */
	v1 = MININV(des->rapv.p[1][j-1]);
        v2 = MININV(des->rapv.p[1][j+1]);
    }
    else if( (dir == 64) || (dir == 192) ){ /* coutour horizontal */
	v1 = MININV(des->raph.p[0][j]);
        v2 = MININV(des->raph.p[2][j]);
    }
    else if( (dir == 32) || (dir == 160) ){ /* coutour diag princip */
	v1 = MININV(des->rapdp.p[0][j+1]);
        v2 = MININV(des->rapdp.p[2][j-1]);
    }
    else if( (dir == 96) || (dir == 224) ){ /* coutour diag second */
	v1 = MININV(des->rapds.p[0][j-1]);
        v2 = MININV(des->rapds.p[2][j+1]);
    }
    else{
	printf("\n\n>> ERREUR maxlocgrad : direction %d inconnue\n\n", (int)dir);
	exit(1);
    }
    
    if( (val <= v1) && (val <= v2) )
	return(1);
    else
	return(0);
}

/* ****************  *********  LECTURE  *******************************/
/** \ingroup contourRSO
    lecture des parametres de l'operateur contourRSO. \\
    @name contourRSO_lect
    @param des pointeur descripteur     
    @param ptp pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author F. Bujor
    @version 1.0 (17/04/00); Include : image.h, proto2D.h, contour.h
*/
param *contourRSO_lect(contourRSO_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s dimension fenetre d'analyse (impair)", debq); 
  lec_param(question, ptp);
  des->dim = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s sauver imagette masque (oui=1/non=0)", debq);
  lec_param(question, ptp);
  des->qsauver = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s selection des contours max locaux (oui=1/non=0)", debq);
  lec_param(question, ptp);
  des->maxloc = atoi(ptp->rep);
  ptp = ptp->next;

  return(ptp);
}

/* *************************  INITIALISATION  ***************************/
/** \ingroup contourRSO
    initialisation de l'operateur contourRSO.\\
    allocation image resultat; \\
    allocation et calcule des masques, sauvegarde facultative \\
    @name contourRSOu2_init 
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @param imval  pointeur image resultat ratio entre 0-1
    @param imdir  pointeur image resultat direction 
    @author F. Bujor
    @version 1.0 (17/04/00); Include : image.h, proto2D.h, contour.h
*/
int contourRSOu2_init(contourRSO_t *des, imau2 im0, imafl *imval, imau1 *imdir){
  int  i, j, k, n;
  
  /* images resultat */ 
  imval->nc = im0.nc;
  imval->nr = im0.nr;
  imdir->nc = im0.nc; 
  imdir->nr = im0.nr;
  sprintf(imval->nom, "%s_ctr_val%dx%d", im0.nom, des->dim, des->dim); 
  sprintf(imdir->nom, "%s_ctr_dir%dx%d", im0.nom, des->dim, des->dim); 
  alloc_imafl(imval);
  alloc_imau1(imdir); 

  /* variables de l'operateur */

  /* *********************** intializari masti**********************/
  /* initialization masque vertical  */
  des->msqv.nc = des->dim; 
  des->msqv.nr = des->dim;
  alloc_imau1(&(des->msqv));
  for(i=0; i<des->dim; i++)
      for(j=0; j<des->dim; j++){des->msqv.p[i][j]=0;}

  /* initialization masque horizontal  */
  des->msqh.nc = des->dim; 
  des->msqh.nr = des->dim;
  alloc_imau1(&(des->msqh));
  for(i=0; i<des->dim-1; i++)
      for(j=0; j<des->dim-1; j++){des->msqh.p[i][j]=0;}

  /* initialization masque diagonal principal */
  des->msqdp.nc = des->dim; 
  des->msqdp.nr = des->dim;
  alloc_imau1(&(des->msqdp));
  for(i=0; i<des->dim-1; i++)
      for(j=0; j<des->dim-1; j++){des->msqdp.p[i][j]=0;}

  /* initialization masque diagonal secondaire */
  des->msqds.nc = des->dim; 
  des->msqds.nr = des->dim;
  alloc_imau1(&(des->msqds));
  for(i=0; i<des->dim-1; i++)
      for(j=0; j<des->dim-1; j++){des->msqds.p[i][j]=0;}
  
  /* buuffers pour selection du max local */
  if(des->maxloc == 1){
      des->rapv.nc = im0.nc; 
      des->rapv.nr = 3;
      alloc_imafl(&(des->rapv));
      for(i=0; i<3; i++)
	  for(j=0; j<im0.nc; j++){des->rapv.p[i][j]=1.;}

      des->raph.nc = im0.nc; 
      des->raph.nr = 3;
      alloc_imafl(&(des->raph));
      for(i=0; i<3; i++)
	  for(j=0; j<im0.nc; j++){des->raph.p[i][j]=1.;}

      des->rapdp.nc = im0.nc; 
      des->rapdp.nr = 3;
      alloc_imafl(&(des->rapdp));
      for(i=0; i<3; i++)
	  for(j=0; j<im0.nc; j++){des->rapdp.p[i][j]=1.;}

      des->rapds.nc = im0.nc; 
      des->rapds.nr = 3;
      alloc_imafl(&(des->rapds));
      for(i=0; i<3; i++)
	  for(j=0; j<im0.nc; j++){des->rapds.p[i][j]=1.;}
  }

  /* ******************* fin intialization ***********************/

  /* ************* calcul du masque ************* */

  /* calcul  masti verticale */
  for(i=0; i<des->dim; i++)
      for(j=0; j<des->dim/2; j++){
	  des->msqv.p[i][j] = 1;
      }
  for(i=0; i<des->dim; i++)
      for(j = des->dim/2+1; j<des->dim ; j++){
	  des->msqv.p[i][j] = 2;
      }

  /* calcul masti orizontale */
  for(i=0; i<(des->dim/2); i++)
      for(j=0; j<des->dim; j++){
	  des->msqh.p[i][j] = 1;
      }
  for(i=(des->dim/2 + 1); i<des->dim; i++)
      for(j=0;j<des->dim; j++){
	  des->msqh.p[i][j] = 2;
      }

   /* calcul masti diagonala principala * */
   n=1; 
   for(i=0;i<des->dim-1; i++){ 
       n++; 
       for(j=n-1; j<des->dim; j++){
	   des->msqdp.p[i][j]=1;
       } 
   } 
   n=1; 
   for(i=1;i<des->dim; i++){ 
       n++; 
       for(j=0; j<n-1; j++){
	   des->msqdp.p[i][j]=2;
       } 
   }   

   /* calcul masti diagonala secundara * */
   n=1; 
   for(i=0; i<des->dim-1; i++){ 
       n++; 
       for(j=0; j<des->dim-n+1; j++){
	   des->msqds.p[i][j]=1;
       } 
   } 
   n=1;  
   for(i=1; i<des->dim; i++){  
       n++;  
       for(j=des->dim-1; j>des->dim-n; j--){
	   des->msqds.p[i][j]=2;
       }  
   }  
/* ************* fin masques ****************/
   
   if(des->qsauver == 1){
       write_ima_t wriv;
       write_ima_t wrih;
       write_ima_t wridp;
       write_ima_t wrids;
       strcpy(wriv.nom, "msqv");
       strcpy(wriv.ext, "ima"); 
       strcpy(wrih.nom, "msqh");
       strcpy(wrih.ext, "ima"); 
       strcpy(wridp.nom, "msqdp");
       strcpy(wridp.ext, "ima"); 
       strcpy(wrids.nom, "msqds");
       strcpy(wrids.ext, "ima"); 
       write_ima_init(&wriv);
       write_ima_init(&wrih);
       write_ima_init(&wridp);
       write_ima_init(&wrids);
       write_imau1_ferm(&wriv, des->msqv);
       write_imau1_ferm(&wrih, des->msqh);
       write_imau1_ferm(&wridp, des->msqdp);
       write_imau1_ferm(&wrids, des->msqds);
   }
}
/* *************************  CALCUL  ***************************/
/** \ingroup contourRSO
    calcul de l'operateur contourRSO.\\
    Calcul du ratio dans les 4 directions, conservation du min et sa direction. \\
    @name contourRSOu2_calc  
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @param imval  pointeur image resultat ratio entre 0-1
    @param imdir  pointeur image resultat direction 
    @author F. Bujor
    @version 1.0 (17/04/00); Include : image.h, proto2D.h, contour.h  
*/
int contourRSOu2_calc(contourRSO_t *des, imau2 im0, imafl *imval, imau1 *imdir){
  int i, j, dims2, k;
  float val, tab_car[65536], v, h, dp, ds;
  float *tmp_ptfl;
  pixu1 dir;

  dims2 = des->dim/2;
  for(i=0; i< 65536; i++)
      tab_car[i] = (float)i*(float)i;
  
  for(i=0; i<im0.nr; i++)
      for(j=0; j<im0.nc; j++){
	  imval->p[i][j] = 1.;
	  imdir->p[i][j] = 0;
      }

  /*  CALCUL AVEC BUFFERISATION POUR SELECTION DES MAX LOCAUX */
  if( des->maxloc == 1){
      /* premiere ligne calculable */
      i = dims2;
      for(j=dims2; j<im0.nc - dims2; j++){
	  des->rapv.p[1][j] = appl_masque12(im0, des->msqv, i, j, tab_car);
	  des->raph.p[1][j] = appl_masque12(im0, des->msqh, i, j, tab_car);
	  des->rapdp.p[1][j] = appl_masque12(im0, des->msqdp, i, j, tab_car);
	  des->rapds.p[1][j] = appl_masque12(im0, des->msqds, i, j, tab_car);
      }

  /* rapport sur ligne i, resultat maxloc sur ligne k=i-1 */
      for(i=dims2+1; i<im0.nr - dims2; i++){
	  k = i-1;
	  for(j=dims2; j<im0.nc - dims2; j++){
	      des->rapv.p[2][j] = appl_masque12(im0, des->msqv, i, j, tab_car);
	      des->raph.p[2][j] = appl_masque12(im0, des->msqh, i, j, tab_car);
	      des->rapdp.p[2][j] = appl_masque12(im0, des->msqdp, i, j, tab_car);
	      des->rapds.p[2][j] = appl_masque12(im0, des->msqds, i, j, tab_car);
	  }
	  for(j=dims2; j<im0.nc - dims2; j++){
	      valdirmax(des->rapv.p[1][j], des->raph.p[1][j], des->rapdp.p[1][j],
			des->rapds.p[1][j], &val, &dir);
	      if( maxlocgrad(val, dir, j, des) ){
		      imval->p[k][j] = val;
		      imdir->p[k][j] = dir;	      
	      }
	  }
	  /* permutation des lignes des 4 buffers */
	  tmp_ptfl = des->rapv.p[0];
	  des->rapv.p[0] = des->rapv.p[1];
	  des->rapv.p[1] = des->rapv.p[2];
	  des->rapv.p[2] = tmp_ptfl;
	  tmp_ptfl = des->raph.p[0];
	  des->raph.p[0] = des->raph.p[1];
	  des->raph.p[1] = des->raph.p[2];
	  des->raph.p[2] = tmp_ptfl;
	  tmp_ptfl = des->rapdp.p[0];
	  des->rapdp.p[0] = des->rapdp.p[1];
	  des->rapdp.p[1] = des->rapdp.p[2];
	  des->rapdp.p[2] = tmp_ptfl;
	  tmp_ptfl = des->rapds.p[0];
	  des->rapds.p[0] = des->rapds.p[1];
	  des->rapds.p[1] = des->rapds.p[2];
	  des->rapds.p[2] = tmp_ptfl;
      }
      /* resultat maxloc sur derniere ligne */
      k = i-1;
	  for(j=dims2; j<im0.nc - dims2; j++){
	      des->rapv.p[2][j] = 1.;
	      des->raph.p[2][j] = 1.;
	      des->rapdp.p[2][j] = 1.;
	      des->rapds.p[2][j] = 1.;
	  }
	  for(j=dims2; j<im0.nc - dims2; j++){
	      valdirmax(des->rapv.p[1][j], des->raph.p[1][j], des->rapdp.p[1][j],
			des->rapds.p[1][j], &val, &dir);
	      if( maxlocgrad(val, dir, j, des) ){
		      imval->p[k][j] = val;
		      imdir->p[k][j] = dir;	      
	      }
	  }
  }

  /*  CALCUL DIRECT  SANS SELECTION DES MAX LOCAUX */
  else{ 
      for(i=dims2; i<im0.nr - dims2; i++)
	  for(j=dims2; j<im0.nc - dims2; j++){
	      v = appl_masque12(im0, des->msqv, i, j, tab_car);
	      h = appl_masque12(im0, des->msqh, i, j, tab_car);
	      dp = appl_masque12(im0, des->msqdp, i, j, tab_car);
	      ds = appl_masque12(im0, des->msqds, i, j, tab_car);
	      valdirmax(v, h, dp, ds, imval->p[i]+j, imdir->p[i]+j);
	  }
  }
}

