/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup manip2d 
    \defgroup conv16to8 conversion image 16bits - 8bits
    operateur de conversion d'une image 16 bits (imw) en 8 bits (ima|ras).\\

\\
Descripteur : \\
   int methode :  methode de conversion : 
                    1=[min,moy+K*sigma], 2=[n(K%), n(100-k%)], 3=[0, K]\\
   float K :  parametrage selon la methode       \\
   int his[65536] : histogramme \\
   pixu1 lut[65536] : table de conversion \\
   char legende[500] : interval de conversion \\
\\
Type entree : imau2 \\
Type sortie : imau1 \\
@name lconv16to8
@author E. Trouve
@version 1.0 (01/02/2001); Include : image.h, proto2D.h, radar.h 
@see mconv16to8_u2
*/

#include "image.h"
#include "proto2D.h"
#include "radar.h" 
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <limits.h> /* pour  USHRT_MAX ... */


/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup conv16to8
    lecture des parametres de l'operateur conv16to8. \\
    methode :  methode de conversion : 
                    1=[min,moy+K*sigma], 2=[n(K%), n(100-k%)], 3=[0, K]\\   
    K :  parametrage selon la methode     \\   
    @name conv16to8_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (01/02/2001); Include : image.h, proto2D.h, radar.h
*/
param * conv16to8_lect(conv16to8_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s methode de conversion : 1=[min,moy+K*sigma], 2=[n(K%%), n(100-k%%)], 3=[0, K]", 
	  debq); 
  lec_param(question, ptp);
  des->methode = atoi(ptp->rep);
  ptp = ptp->next; 
  if(  des->methode < 1 || des->methode > 3 ){
      printf("\n>> ERREUR conv16to8_lect : methode %d inconnue \n", des->methode);
      exit(1);
  }

  sprintf(question, "%s valeur du parametre K", debq); 
  lec_param(question, ptp);
  des->K = (float)atof(ptp->rep);
  ptp = ptp->next;

  return(ptp);
}


/* **********************  INITIALISATION  *****************************/
/** \ingroup conv16to8
    initialisation de l'operateur conv16to8.\\
    allocation image resultat \\
    @name conv16to8_init
    @param des : pointeur descripteur
    @param im0 : image initiale (taille connue)
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (01/02/2001); Include : image.h, proto2D.h, radar.h
*/

/* initialisation (16 bit)*/
int conv16to8_init(conv16to8_t *des, imau2 im0, imau1 *imres){
  int n;

  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(conv16to8m%d)", im0.nom, des->methode);
  alloc_imau1(imres);

  if(des->methode == 1 || des->methode == 2){
      for(n=0; n<65536; n++)
	  des->his[n] = 0;
  }

  return(0);  
}

/* *************************  CALCUL  ***************************/
/** \ingroup conv16to8
    calcul de l'operateur conv16to8.\\
    @name conv16to8_calc
    @param des : pointeur descripteur
    @param im0 : image initiale 
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (01/02/2001); Include : image.h, proto2D.h, radar.h
*/

/* calcul (16 bit) */
int conv16to8_calc(conv16to8_t *des, imau2 im0, imau1 *imres){
  int i, j, n, nbp, min,  val, minlut, maxlut;
  float som, som2, moy, sigma, sigma2, a, b, tmpfl;

  if(des->methode == 1 || des->methode == 2){  
      min = im0.p[0][0];
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++){
	      val = (int)im0.p[j][i];
	      des->his[val]++;
	      if(val < min)
		  min = val;
	  }
  }
  nbp = im0.nc * im0.nr;
  switch(des->methode){
  case 1 :
      som = 0;
      som2 =0;
      for(n=1; n<65536; n++){
	  som += (float)n*(float)des->his[n];
	  som2 += (float)n*(float)n*(float)des->his[n];
      }
      moy = som/nbp;
      sigma2 = (som2 - nbp*moy*moy)/(nbp-1);
      sigma = (float)sqrt((double)sigma2);
      minlut = min;
      tmpfl = moy + des->K * sigma + 0.5;
      maxlut = (int)tmpfl;
      sprintf(des->legende, "conversion lineaire sur [min:%d, moy(%.2f) + %.2f * sigma(%.2f) = %d]\n", 
             minlut, moy, des->K, sigma, maxlut);
      break;
  case 2 :
      tmpfl = des->K * nbp / 100. + 0.5;
      n=0; 
      som = 0;
      do{
	  som += des->his[n];
	  n++;
      }while(som < tmpfl);
      minlut = n;
      n=65535; 
      som = 0;
      do{
	  som += des->his[n];
	  n--;
      }while(som < tmpfl);
      maxlut = n;
      sprintf(des->legende, "conversion lineaire sur [n(%.2f%%):%d, n(100-%.2f%%):%d]\n", 
             des->K, minlut, des->K, maxlut);
      break;
  case 3 :
      minlut = 0;
      maxlut = (int)des->K;
      sprintf(des->legende, "conversion lineaire sur [%d, %d]\n",  minlut, maxlut);
      break;
  }
  printf("\n>>conv16to8 : %s\n",  des->legende);

  /* lut */    
  a = 255.9999/(double)(maxlut-minlut);
  b = -a*(double)minlut;
  for(n=0; n<65536; n++)
          if( n <= minlut )
              des->lut[n] = 0;
          else if( n >= maxlut )
              des->lut[n] = 255;
          else{
              tmpfl = a*(float)n + b + 0.5;
              des->lut[n] = (pixu1)tmpfl;
          }
  /* conversion */    
  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++)
	  imres->p[j][i] = des->lut[im0.p[j][i]];


  return(0);
}



