/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \ingroup divers
    \defgroup distance_imau1 Image de distance
    Operateur distance_imau1 : construit une image de DISTANCE  partir d'une image binaire (au format imau1 donc sur 8bits)
 et ceci avec un maillage rectangulaire (h*l avec h=1). Les distances sont calcules avec un masque de chanfrein de taille 5*5. Les coefficients du masque sont les suivants (X reprsente le pixel central du masque):
\begin{center}
\begin{tabular}{ccc}
  & e & \\
d & b & c\\
X & a & \\
\end{tabular}
\end{center}
Les valeurs optimales (relles) de ces coefficients sont calcules connaissant la valeur de {\bf l}. Ils permettent d'obtenir une distance discrete (obtenue par propagation d'information locale de distance) trs proche de la distance euclidienne.Les coefficients du masque de chanfrein devant tre entiers, les coefficients rels sont multiplis par un facteur {\bf n} permettant d'avoir le moins d'erreur possible par rapport  la distance euclidienne. Une tude prconise de prendre n=73 pour un maillage carr (l=1) et des images de taille suprieure  128 (sinon prendre n=5).
\\
Dans l'image binaire d'entre, les pixels reprsentant le fond ont un niveau de gris gal  255 (blanc) tandis que les objets sont au niveau 0 (noir). Pour chaque pixel de l'image, la distance au plus proche objet est dtermine  l'aide du masque de chanfrein. Les valeurs de distances sont converties au format u1 (8 bits non sign) pour la visualisation.\\
L'image de sortie est alloue dans la fonction distance_imau1_init, donc pas besion de le faire avant.\\
\\
Descripteur :
\begin{itemize}
    \item float l      :  longueur du pixel
    \item int n        :  coefficient d'arrondi
\end{itemize}

@name ldistance_imau1
@author Lionel VALET & Didier COQUIN
@version 1.0 (02/12/99); Include : distance_ima.h
@see mdistance_imau1
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "image.h"
#include "proto2D.h"
#include "distance_ima.h"


#define   max(x,y)  (((x)>(y))?(x):(y))
#define   min(x,y)  (((x)<(y))?(x):(y))
#define   abs(x) ((x>0)?x:-x)

typedef  unsigned char Byte;



/* ****************************************************************************/
/*                             FONCTION MINIMUN                               */
/*   Cette fonction retourne le minimum d'un tableau de 25 elements .         */
/*   ainsi que sa position x0 et y0					      */
/* ****************************************************************************/


int  minimum( unsigned int   masque[5][5],
	      unsigned int   *x0, unsigned int *y0)
{
     register int i,j;
     int mini, mini1;

     mini=masque[0][0];
     *x0 = 0;
     *y0 = 0;
     for ( i=0 ; i<5 ; i++ )
         for ( j=0 ; j<5 ; j++ )
         {            
             mini1=min(masque[i][j],mini);
             if (mini1 < mini)
             {
             	mini = mini1;
             	*x0 = i;
             	*y0 = j;
             }  
         }

         return   mini;               
     }


/* ***************************************************************************/
/*                       FIN  FONCTION MINIMUN                               */
/* ***************************************************************************/


/* ***************************************************************************/
/*                             FONCTION DISTANCE                             */
/*   Cette fonction calcule la distance de chaque pixel d'une image au       */
/*                    fond ( Background ).                                   */
/* ***************************************************************************/

void  distance(imau1 *im1, imau1 *im2 , float l, int n)
{

       register int   x,y;
       int            a1,b1,c1,d1,e1;
       int            maximum,fin,niveau,recop,recop1;
       int            w,compt;
       unsigned int   x1,y1;
       double         val;
       float          valeur,inter;
       float          tt,vv,s;
       float          n1,n2,n3,num,denom;
       float          delta1,delta2,delta3;
       float          A,B,C,D,E;
       int            A1,B1,C1,D1,E1;   
       int            ydebut;
       int            *ima_dist;
       int            *ima_dist1;
       int            *ima_sir;
       unsigned int   fen[5][5];
       int            largeur, hauteur;
       Byte  * ima1, *ima2;

       	       /*-----------------------------------------------------*/
               /*      CALCUL DES COEFFICIENTS OPTIMAUX A,B,C,D,E.    */
               /*-----------------------------------------------------*/

       largeur = im1->nc;
       hauteur = im1->nr;
       ima1 = im1->p[0];
       ima2 = im2->p[0];

       ima_dist=(int *)calloc(largeur*hauteur,sizeof(int));
       ima_dist1=(int *)calloc(largeur*hauteur,sizeof(int));
       ima_sir=(int *)calloc(largeur*hauteur,sizeof(int)); 
         
       
       tt=sqrt(4.0*l*l+1.0);
       vv=sqrt(1.0*l*l+4.0);
       s=sqrt(1.0*l*l+1.0); 
       delta1=4.0*l*l*(1.0+(tt-2.0*l)*(tt-2.0*l)); 
       delta3=l*l+(vv-2.0)*(vv-2.0); 
       A=(-2.0*l+sqrt(delta1))/(tt-2.0*l)/(tt-2.0*l);
       a1=floor((A*n)+0.5);
       C=A*tt/l;
       c1=floor((C*n)+0.5);
       n1=C*(1.0+2.0*l*l)-2.0*l*l/s;
       denom=1.0+4.0*l*l-l*l/s/s;  
       n2=C*C*(1.0+l*l)-4.0*l*l;
       delta2=n1*n1-n2*denom;
       num=n1+sqrt(delta2);
       B=num/denom;
       b1=floor((B*n)+0.5);
       num=2.0*l*sqrt(delta3)-2.0*l*l; 
       denom=(vv-2.0)*(vv-2.0);
       D=num/denom;
       d1=floor((D*n)+0.5);
       E=D*vv;
       e1=floor((E*n)+0.5);

       printf("Aopt=%d  Bopt=%d  Copt=%d  Dopt=%d   Eopt=%d\n",a1,b1,c1,d1,e1); 
       printf("pour n=%d  et L=%f\n",n,l);


            /*----------------------------------------------*/
            /*          MISE EN FORME DE L'IMAGE SOURCE     */
            /*----------------------------------------------*/
/* On traite une image binaire: Objet = 0,  Fond = 255   *****/
/* les points du fond initialement a 255 sont mis a l'infini ***/
/* les autres, donc l'objet reste a 0		        ***/
   
       fin=(largeur)*(hauteur);

	   printf("largeur= %d  hauteur= %d\n",largeur,hauteur);       
      
       for ( x=0 ; x<fin ; x++ )
           {
           if ( ima1[x]!=0)
           ima_sir[x]=20000;
           else  ima_sir[x]=0; 
           }

  
/* ima_sir = image de reference utiliser pour calculer l'image de distance **/
/* ima_dist1 = image de distance apres le passage du masque_avant     *******/
/* ima_dist  = image de distance finale   *****/      

           /*----------------------------------------------*/
           /*               BOUCLE ALLER                   */
           /*  Balayage de gche a drte et de haut en bas   */
           /*----------------------------------------------*/
	
       for ( y=2 ; y<hauteur-2 ; y++ )
           for ( x=2 ; x<largeur-2 ; x++ )
               {
               w=x+y*largeur;
               if (ima_sir[w]!=0)
               {
  				/* comparaison avec le masque avant  ****/		                                 
                  fen[0][0]=(2*b1)+(ima_sir[w-2*largeur-2]);
                  fen[0][1]=(e1)+(ima_sir[w-2*largeur-1]);  
                  fen[0][2]=(2*d1)+(ima_sir[w-2*largeur]);  
                  fen[0][3]=(e1)+(ima_sir[w-2*largeur+1]);  
                  fen[0][4]=(2*b1)+(ima_sir[w-2*largeur+2]);  
                   
                  fen[1][0]=(c1)+(ima_sir[w-largeur-2]);      
                  fen[1][1]=(b1)+(ima_sir[w-largeur-1]);    
                  fen[1][2]=(d1)+(ima_sir[w-largeur]);    
                  fen[1][3]=(b1)+(ima_sir[w-largeur+1]);   
                  fen[1][4]=(c1)+(ima_sir[w-largeur+2]);

                  fen[2][0]=(2*a1)+(ima_sir[w-2]);
                  fen[2][1]=(a1)+(ima_sir[w-1]);    
                  fen[2][2]=20000;   
                  fen[2][3]=20000;                 
                  fen[2][4]=20000; 
                     
                  fen[3][0]=20000;                
                  fen[3][1]=20000;                 
                  fen[3][3]=20000;                  
                  fen[3][4]=20000;
                  fen[3][2]=20000;
                  fen[4][0]=fen[4][1]=fen[4][2]=fen[4][3]=fen[4][4]=20000;
 
                  
                  ima_dist1[w]=minimum(fen,&x1,&y1);
                  ima_sir[w]=ima_dist1[w];
                }
                else ima_dist[w] = 0;
               }                        /*    fin du for */


      /*-------------------------------------------------------*/             
      /*                     BOUCLE RETOUR                     */
      /*  Balayage de l'image de drte a gche et de bas en haut */
      /*-------------------------------------------------------*/
	
       for ( y=hauteur-3 ; y>1 ; y-- )
           for ( x=largeur-3 ; x>1 ; x-- )
               {
               w=x+(y)*largeur;
               if (ima_sir[w]!=0)
               {
   				/* comparaison avec le masque arriere  ****/

                  fen[4][0]=(2*b1)+(ima_sir[w+2*largeur-2]);
                  fen[4][1]=(e1)+(ima_sir[w+2*largeur-1]);  
                  fen[4][2]=(2*d1)+(ima_sir[w+2*largeur]);  
                  fen[4][3]=(e1)+(ima_sir[w+2*largeur+1]);  
                  fen[4][4]=(2*b1)+(ima_sir[w+2*largeur+2]);  
                   
                  fen[3][0]=(c1)+(ima_sir[w+largeur-2]);      
                  fen[3][1]=(b1)+(ima_sir[w+largeur-1]);    
                  fen[3][2]=(d1)+(ima_sir[w+largeur]);    
                  fen[3][3]=(b1)+(ima_sir[w+largeur+1]);   
                  fen[3][4]=(c1)+(ima_sir[w+largeur+2]);

                  fen[2][4]=(2*a1)+(ima_sir[w+2]);
                  fen[2][3]=(a1)+(ima_sir[w+1]);    
                  fen[2][2]=20000;   
                  fen[2][0]=20000;                 
                  fen[2][1]=20000;                      
                  fen[0][0]=fen[0][1]=fen[0][2]=fen[0][3]=fen[0][4]=20000;

                  fen[1][0]=20000;                  
                  fen[1][1]=20000;
                  fen[1][3]=20000;                 
                  fen[1][4]=20000;
                  fen[1][2]=20000;
             
                  recop1=minimum(fen,&x1,&y1);
                  recop=ima_dist1[w];                 
                  ima_dist[w]=min(recop,recop1);
                  ima_sir[w]=ima_dist[w];
                }
                else ima_dist[w]=0;        
                             
              }              /*    fin du for */


             /*----------------------------------------------------*/   
             /*          remplissage de ima_dest                   */
             /*   Les valeurs obtenues sont ramenees au Byte (255) */
             /*----------------------------------------------------*/
      
/* recherche de la valeur maximale, pour la Normalisation  *****/
        maximum=ima_dist[0];
        for ( x=0 ; x<fin ; x++ )
             maximum=max(ima_dist[x],maximum); 


/* Pour afficher l'image de DISTANCE il faut que les valeurs  ******/
/*  soient inferieures a 256	==> normalisation des niveaux	 ******/
	for ( y=(fin-1) ; y>=0 ; y-- )
	    ima2[y] = (Byte)(ima_dist[y]*255.0/maximum);
      



        printf("\nFIN IMAGE DE DISTANCE: \n");
        printf("\n");
		
        cfree(ima_sir);
        cfree(ima_dist);
        cfree(ima_dist1);  

}                      /*    fin de distance    */


/** \ingroup distance_imau1
    lecture des parametres de l'operateur distance_imau1. \\
    l : longueur du pixel\\
    n : coefficient d'arrondi\\
    @param des  pointeur descripteur     
    @param ptp  pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author L. VALET
    @version 1.0 (02/12/99); Include : distance_ima.h
*/
param *distance_imau1_lect( image_distance_t *des, param * ptp, char * debq)
{
  char question[500];

  sprintf(question, "%s Valeur de l (reel)", debq);
  lec_param(question, ptp);
  des->l = atof(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s Valeur de n (entier)", debq);
  lec_param(question, ptp);
  des->n = atoi(ptp->rep);
  ptp = ptp->next;

  return(ptp);
  printf("sortie lect\n");
}


/**  \ingroup distance_imau1
     initialisation de l'operateur distance_imau1.\\
    allocation de l'image rsultat.
    @param des  pointeur descripteur
    @param imaO  pointeur image initiale
    @param imaRes pointeur image rsultat
    @author L. VALET
    @version 1.0 (19/10/99); Include : distance_ima.h
*/
int distance_imau1_init(image_distance_t *des, imau1 *imO, imau1 *imRes)
{
	imRes->nr = imO->nr;
	imRes->nc = imO->nc;
	alloc_imau1(imRes);
    
}

/**  \ingroup distance_imau1
     Calcul de l'image de distance .\\
    @param des  pointeur descripteur
    @param imaO  pointeur image initiale
    @param imaRes pointeur image rsultat
    @author L. VALET
    @version 1.0 (19/10/99); Include : distance_ima.h
*/
int distance_imau1_calc( image_distance_t *des, imau1 *imO, imau1 *imRes)
{
    distance( imO, imRes , des->l, des->n);
}
