/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* **************************** lhisto.c ******************************/
/*                                                                    */
/* ********************************************************************/

/** \ingroup divers 
\defgroup histo operateur histogramme.
Calcul d'histogramme sur des images 8 bits (u1), 16 bits (u2) et float (fl), 
valeur min, valeur max et pas (nombre de cases) parametrable.\\
Histogramme de toute l'image ou conditionnel 
(zone definie par un label dans un masque).\\ 
 \\
1) Images 8 (resp. 16) bits : si min=max=0 (auto), histo sur [0, 256[, 
(resp. [0, 65536[), traitement bloc par bloc OK.\\
si nombre de cases = 0 (auto), nombre de cases mis a 256 (resp 65536)\\
\\
2) Images float : si min=max=0 (auto), recherche du min et du max de l'image 
(ATT dans ce cas, traitement bloc par bloc impossible).\\
si nombre de cases = 0 (auto), nombre de cases mis a 1000\\
\\
Fichier ascii resultat en 2 colonnes :\\
- la valeur [n] (la borne inferieure de chaque intervale de quantification 
[min+n*pas, min+(n+1)*pas]\\
- le nombre de pixels dans cet intervale \\
\\
Descripteur : \\
    int bins          : nombre d'intervalles pour histogramme \\
    FILE *pfd         : pointeur fichier histogramme \\
    char nomhist[200] : nom complet du fichier \\
    unsigned int *h   : tableau histogramme \\
    float min         : valeur minimale pour histogramme \\
    float max         : valeur maximale pour histogramme \\
    float step        : valeur pas pour interval \\
    int ngm           : niveau de gris du masque [0 a 255] ou -1 => pas de masque\\
\\
Type entree : imau1 | imau2 | imafl \\
Type sortie : fichier ascii avec donnees pour representer l'histogramme
@name lhisto
@author F. Bujor, E. Trouve
@version 1.0 (09/07/00); Include : image.h, proto2D.h, histo.h
@see mhisto_XX,  XX = u1 | u2 | fl .
*/

#include "image.h"
#include "proto2D.h"
#include "histo.h"
#include <stdlib.h>
#include <limits.h>
#include <string.h>

/* *************************  LECTURE  *******************************/
/** \ingroup histo
    lecture des parametres de l'operateur histogramme. \\
    @param des pointeur descripteur     
    @param ptp pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author F. Bujor, E. Trouve
    @version 1.0 (09/07/00); Include : image.h, proto2D.h, histo.h
*/

param *histo_lect(histo_t *des, param *ptp, char *debq){
  char question[500];
  int l, j;

  sprintf(question, "%s nom complet du fichier ASCII resultat", debq);
  lec_param(question, ptp);
  strcpy(des->nomhist, ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s valeur minimale ([0, 0] : auto)", debq);
  lec_param(question, ptp);
  des->min = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s valeur maximale ([0, 0] : auto)", debq);
  lec_param(question, ptp);
  des->max = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s nombre de cases (0 : auto)", debq);
  lec_param(question, ptp);
  des->bins = atoi(ptp->rep);
  ptp = ptp->next;  

  sprintf(question, "%s niveau de gris de la region definie dans un masque (-1 => toute l'image)", debq);
  lec_param(question, ptp);
  des->ngm = atoi(ptp->rep);
  ptp = ptp->next;  

  return(ptp);
}

/* *************************  INITIALISATION  ***************************/
/** \ingroup histo
    XX = u1 | u2 | fl : initialisation de l'operateur histogramme.\\
Attention : pour les images float, si appel avec min=max=0 (borne automatiques), 
l'image analysee im0 doit deja etre presente pour que l'init puisse trouver 
le min et le max de l'image, l'image masque egalement si on utilise un masque.
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @param mas  image masque (facultatif)
    @author F. Bujor, E. Trouve
    @version 1.0 (09/07/00); Include : image.h, proto2D.h, histo.h
*/
int histofl_init(histo_t *des, imafl im0, imau1 mas){
  int i, j;
  float min, max, step;
  pixu1 ngmu1;

  if( (des->pfd=fopen(des->nomhist,"w")) == NULL ){
      printf("\n>> ERREUR histofl_init: creation fichier %s impossible\n",des->nomhist);
      exit(1);
  }

  if( des->bins == 0)
      des->bins = 1000;
  if(  (des->h = (unsigned int *)malloc(des->bins*sizeof(unsigned int))) == NULL){
      printf ("\n ERREUR histofl_init : allocation histogramme impossible\n");  
      exit(1); 
  }  
  for(j = 0; j < des->bins; j++)
      des->h[j] = 0;

  if((des->max == 0.0) && (des->min == 0.0)){
      ngmu1 = (pixu1)des->ngm;
      des->max = im0.p[0][0];
      des->min = im0.p[0][0];

      if(des->ngm < 0){
	  for(j = 0; j<im0.nr; j++)
	      for(i = 0; i<im0.nc; i++){
		  if( im0.p[j][i] >= des->max )
		      des->max = im0.p[j][i];
		  if( im0.p[j][i] <= des->min )
		      des->min = im0.p[j][i];
	      }
      }
      else{
	  for(j = 0; j<im0.nr; j++)
	      for(i = 0; i<im0.nc; i++){
		  if( mas.p[j][i] == ngmu1 ){
		      if( im0.p[j][i] >= des->max )
			  des->max = im0.p[j][i];
		      if( im0.p[j][i] <= des->min )
			  des->min = im0.p[j][i];
		  }
	      }
	  }
  }
  printf("\n histofl_init : histogramme sur [%f,%f], %d cases\n", 
	 des->min, des->max, des->bins);

  des->step = (des->max - des->min)/(des->bins);

  return(0);      
}
/* *************************  CALCUL  ***************************/
/** \ingroup histo
    XX = u1 | u2 | fl : calcul de l'operateur histogramme.\\
     
    @param des  pointeur descripteur
    @param im0  image (ou bloc) initiale 
    @param mas  image (ou bloc) masque (facultatif) 
    @author F. Bujor, E. Trouve
    @version 1.0 (09/07/00); Include : image.h, proto2D.h, histo.h
*/

int histofl_calc(histo_t *des, imafl im0, imau1 mas){ 
  int i, j;
  int tmp;
  float tmpfl;
  pixu1 ngmu1;

  if(des->ngm < 0){
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++){
	      tmpfl = (im0.p[j][i]-des->min)/des->step;
	      tmp = (int) tmpfl;
	      if ((tmp < des->bins) && (tmp >= 0))
		  des->h[tmp]++;
	      else if (tmp >= des->bins) 
		  des->h[des->bins-1]++;
	      else
		  des->h[0]++;
	  }
  }
  else{
      ngmu1 = (pixu1)des->ngm;
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++)
	      if(mas.p[j][i] == ngmu1){
		  tmpfl = (im0.p[j][i]-des->min)/des->step;
		  tmp = (int) tmpfl;
		  if ((tmp < des->bins) && (tmp >= 0))
		      des->h[tmp]++;
		  else if (tmp >= des->bins) 
		      des->h[des->bins-1]++;
		  else
		      des->h[0]++;
	  }

  }
}

/* *************************  INITIALISATION  ***************************/
/* XX = u1 : initialisation de l'operateur histogramme pour images 8 bit.\\
    @name histou1_init 
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @author F. Bujor
    @version 1.0 (26/04/00); Include : image.h, proto2D.h, histo.h
*/

int histou1_init(histo_t *des, imau1 im0, imau1 mas){
  int i, j;
  float min, max, step;

  if( (des->pfd=fopen(des->nomhist,"w")) == NULL ){
      printf("\n>> ERREUR histou1_init: creation fichier %s impossible\n",des->nomhist);
      exit(1);
  }

  if( des->bins == 0)
      des->bins = 256;
  if(  (des->h = (unsigned int *)malloc(des->bins*sizeof(unsigned int))) == NULL)  
        {printf ("\n ERREUR histou1_init : allocation histogramme impossible\n");  exit(1); }  
  for(j = 0; j < des->bins; j++)
      des->h[j] = 0;

  if((des->max == 0.0) && (des->min == 0.0)){
      des->max = 256;
      des->min = 0;
  }
  else{
      if(des->max > 256){
	  printf("\n ATTENTION histou1_init : valeur maximale corrigee a 256 \n");
	  des->max = 256;
      }
  
      if(des->min < 0){
	  printf("\n ATTENTION histou1_init : valeur minimale corrigee a 0 \n");
	  des->min =0;
      }
  }
  printf("\n histou1_init : histogramme sur [%f,%f], %d cases\n", 
	 des->min, des->max, des->bins);

  des->step = (des->max-des->min)/(des->bins);

  return(0);      
}
/* *************************  CALCUL  ***************************/
/* XX = u1 : calcul de l'operateur histogramme pour images 8 bit.\\
    @name histou1_calc  
    @param des  pointeur descripteur
    @param im0  image initiale 
    @author F. Bujor
    @version 1.0 (26/04/00); Include : image.h, proto2D.h, histo.h
*/

int histou1_calc(histo_t *des, imau1 im0, imau1 mas){ 
  int i, j;
  int tmp; 
  float tmpfl;
  pixu1 ngmu1;

  if(des->ngm < 0){
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++){
	      tmpfl = (im0.p[j][i]-des->min)/des->step;
	      tmp = (int) tmpfl;
	      if ((tmp < des->bins) && (tmp >= 0))
		  des->h[tmp]++;
	      else if (tmp >= des->bins) 
		  des->h[des->bins-1]++;
	      else
		  des->h[0]++;
	  }
  }
  else{
      ngmu1 = (pixu1)des->ngm;
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++)
	      if(mas.p[j][i] == ngmu1){
		  tmpfl = (im0.p[j][i]-des->min)/des->step;
		  tmp = (int) tmpfl;
		  if ((tmp < des->bins) && (tmp >= 0))
		      des->h[tmp]++;
		  else if (tmp >= des->bins) 
		      des->h[des->bins-1]++;
		  else
		      des->h[0]++;
	  }
  }
}

/* *************************  INITIALISATION  ***************************/
/* 
   XX = u2 : initialisation de l'operateur histogramme pour images 16 bit.\\
    @name histou2_init 
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @author F. Bujor
    @version 1.0 (26/04/00); Include : image.h, proto2D.h, histo.h
*/
int histou2_init(histo_t *des, imau2 im0, imau1 mas){
  int i, j;
  float min, max, step;

  if( (des->pfd=fopen(des->nomhist,"w")) == NULL ){
      printf("\n>> ERREUR histou2_init: creation fichier %s impossible\n",des->nomhist);
      exit(1);
  }

  if( des->bins == 0)
      des->bins = 65536;
  if( (des->h = (unsigned int *)malloc(des->bins*sizeof(unsigned int))) == NULL ){
      printf("\n ERREUR histou2_init : allocation histogramme impossible\n");
      exit(1); 
  }  
  for(j = 0; j < des->bins; j++)
      des->h[j] = 0;

  if((des->max == 0.0) && (des->min == 0.0)){
      des->max = 65536;
      des->min = 0;
  }
  else{
      if(des->max > 65536){
	  printf("\n ATTENTION histou2_init : valeur maximale corrigee a 65536 \n");
	  des->max = 65536;
      }

      if(des->min < 0){
	  printf("\n ATTENTION histou2_init : valeur minimale corrigee a 0 \n"); 
	  des->min =0;
      }
  }
  printf("\n histou2_init : histogramme sur [%f,%f], %d cases\n", 
	 des->min, des->max, des->bins);
  des->step = (des->max-des->min)/(des->bins);

  return(0);      
}
/* *************************  CALCUL  ***************************/
/* XX = u2 : calcul de l'operateur histogramme pour images 16 bit.\\
    @name histou2_calc  
    @param des  pointeur descripteur
    @param im0  image initiale 
    @author F. Bujor
    @version 1.0 (26/04/00); Include : image.h, proto2D.h, histo.h
*/

int histou2_calc(histo_t *des, imau2 im0, imau1 mas){ 
  int i, j;
  int tmp; 
  float tmpfl;
  pixu1 ngmu1;

  if(des->ngm < 0){
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++){
	      tmpfl = (im0.p[j][i]-des->min)/des->step;
	      tmp = (int) tmpfl;
	      if ((tmp < des->bins) && (tmp >= 0))
		  des->h[tmp]++;
	      else if (tmp >= des->bins) 
		  des->h[des->bins-1]++;
	      else
		  des->h[0]++;
	  }
  }
  else{
      ngmu1 = (pixu1)des->ngm;
      for(j=0; j<im0.nr; j++)
	  for(i=0; i<im0.nc; i++)
	      if(mas.p[j][i] == ngmu1){
		  tmpfl = (im0.p[j][i]-des->min)/des->step;
		  tmp = (int) tmpfl;
		  if ((tmp < des->bins) && (tmp >= 0))
		      des->h[tmp]++;
		  else if (tmp >= des->bins) 
		      des->h[des->bins-1]++;
		  else
		      des->h[0]++;
	      }
  }
}

/* *************************  FERMETURE  ***************************/
/** fermeture de l'operateur histogramme.\\
    @name histo_ferm  
    @param des  pointeur descripteur
    @author F. Bujor
    @version 1.0 (26/04/00); Include : image.h, proto2D.h, histo.h
*/
int histo_ferm(histo_t *des){
     int j;
     
     for(j = 0; j < des->bins; j++)
          if( fprintf(des->pfd, "%f \t %12d\n", des->min+j*des->step, des->h[j]) == 0){
	     printf("\n>> ERREUR histo_ferm : ecriture de fichier %s\n", des->nomhist);
	     exit(1);
	 }
     close(des->pfd); 
     return(0);
}
  
