/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup filtrage
    \defgroup idctexp operateur idctexp

operateur idctexp (DCT-1 sur axe frequentiel et exponentielle).\\
3eme et dernier operateur du filtre multidate frequentiel de Daniela Coltuc : \\
conversion de N canaux de DCT (espace transforme) en N images RSO (espace initial).
La DCT inverse est calculee en chaque point de l'image 2D. On revient 
ensuite en bruit multiplicatif en prenant l'exponentielle des valeurs et  
en divisant le resultat par la valeur theorique du biais introduit par ce filtrage.\\
\\
Descripteur : \\
   int N           :   nombre d'images de la serie =  taille de la DCT\\
   float B         :   biais attendu pour correction apres l'exponentielle (->exp()/B)\\
   float *v0       :   vecteur pour stoker les points a transformer \\
   float **matidct :   matrice de calcul de DCT-1  N points, 1 ligne par coef \\
\\
Type entree : serie images fl (canaux de DCT de 0 a N-1)\\
Type sortie : serie images u2 (images RSO multidates resultat)\\
@name lidctexp
@author E. Trouve
@version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h 
@see llogdct.c,  midctexp_u2.c, mfiltmdate_dct.c.
*/

#include "image.h"
#include "proto2D.h"
#include "radar.h"
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <limits.h> /* pour  USHRT_MAX ... */

/* #include "radar_comp.h" */



/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup idctexp
    lecture des parametres de l'operateur idctexp. \\
    N : nombre d'images de la serie = taille de la DCT\\
    B         :   biais attendu pour correction apres l'exponentielle \\
    @name idctexp_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h
*/
param * idctexp_lect(idctexp_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s nombre d'images", debq); 
  lec_param(question, ptp);
  des->N = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s biais a corriger (division apres l'exponentielle)", debq); 
  lec_param(question, ptp);
  des->B = (float)atof(ptp->rep);
  ptp = ptp->next; 

  return(ptp);
}

/* **********************  INITIALISATION  *****************************/
/** \ingroup idctexp
    XX =  u2  : initialisation de l'operateur idctexp.\\
    allocation serie N images resultat (uniquement si pt_imres[n].p == NULL),\\ 
    allocation vecteur de travail (N points),\\
    allocation et calcul de la matrice NxN de la DCT-1.\\
    @name idctexp_XX_init
    @param des : pointeur descripteur
    @param pt_im0 : pointeur serie images initiales (taille connue)
    @param pt_imres : pointeur serie images resultat
    @author E. Trouve
    @version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h
*/

/* initialisation (8 bit)*/

/* initialisation (16 bit)*/
int idctexp_u2_init(idctexp_t *des, imafl *pt_imdct, imau2 *pt_imres){
  int n, k, N;
  float c0, ck;
  double tmpdb;

  N = des->N;
  /* allocation images resultat */
  for(n=0; n<N; n++){
      if (pt_imres[n].p == NULL){
	  pt_imres[n].nc = pt_imdct[0].nc;
	  pt_imres[n].nr = pt_imdct[0].nr;
	  sprintf(pt_imres[n].nom, "%s_idctexp_%d", pt_imdct[0].nom, n);
	  alloc_imau2(pt_imres+n);
      }
  }

  /* allocation vecteur de donnees */
  if(  (des->v0 = (float *)malloc(N*sizeof(float))) == NULL ){
      printf ("\n ERREUR idctexp_u2_init : allocation vecteur impossible\n");  
      exit(1); 
  }  

  /* allocation matrice coef de calcul de DCT */ 
  if(  (des->matidct = (float **)malloc(N*sizeof(float *))) == NULL ){
      printf ("\n ERREUR idctexp_u2_init : allocation matrice impossible\n");  
      exit(1); 
  }  
  if( (des->matidct[0] = (float *)malloc(N*N*sizeof(float))) == NULL){
      printf ("\n ERREUR idctexp_u2_init : allocation image impossible\n");  
      exit(1); 
  }
  for(n=1; n<N; n++)
        des->matidct[n] =  des->matidct[0] + n*N;

  /* calcul matrice coef de calcul de IDCT */ 
  c0 = (float)sqrt(1./(double)N);  
  ck = (float)sqrt(2./(double)N);
  for(n=0; n<N; n++){
      des->matidct[n][0] = c0;
      for(k=1; k<N; k++)
	 des->matidct[n][k] = ck*(float)cos(M_PI*(double)((2*n+1)*k)/(double)(2*N));
  }

  return(0);  
}

/* initialisation (float)*/

/* initialisation (double)*/


/* *************************  CALCUL  ***************************/
/** \ingroup idctexp
    XX = u2 : calcul de l'operateur idctexp.\\
    @name idctexp_XX_calc
    @param des : pointeur descripteur
    @param pt_im0 : pointeur serie images initiales 
    @param pt_imres : pointeur serie images resultat
    @author E. Trouve
    @version 1.0 (09/09/00); Include : image.h, proto2D.h, radar.h
*/

/* calcul (8 bit) */

/* calcul (16 bit) */
int idctexp_u2_calc(idctexp_t *des, imafl *pt_imdct, imau2 *pt_imres){
  int i, j, dimX, dimY, n, k, N;
  float som, res;
  
  N = des->N;
  dimX = pt_imdct->nc; 
  dimY = pt_imdct->nr; 
  for(j=0; j<dimY; j++)
      for(i=0; i<dimX; i++){
	  for( k=0; k<N; k++)
	      des->v0[k] = pt_imdct[k].p[j][i];
	  for( n=0; n<N; n++){
	      som = des->matidct[n][0] * des->v0[0];
	      for( k=1; k<N; k++)
		  som += des->matidct[n][k] * des->v0[k];
	      res = (float)exp((double)som) / des->B + 0.5;
	      if(res >= USHRT_MAX +1)
		  pt_imres[n].p[j][i] = USHRT_MAX;
	      else if (res <=0.)
		  pt_imres[n].p[j][i] = 0;
	      else
		  pt_imres[n].p[j][i] = (pixu2)res;
	  }
      }
  return(0);
}

/* calcul (float) */

/* calcul (double) */

/* fonction : ******************************* */

