/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ************************* lindexed_img.c *************************** */
/*                                                                      */
/* ******************************************************************** */

/** \ingroup manip2d
    \defgroup indexed_img operateur lindexed_img 

operateur qui permet de creer une image
de couleurs indexees (8 bits), a partir d'une image couleur (RVB) 24 bits 
en utilisant un reseau de neurone de Kohonen (SOM).
 
@name lindexed_img
@author B. Ionescu
@version 1.0 (15/05/00); Include : proto2D.h, indexed_img.h
@see mindexed_img
*/

#include "image.h"
#include "proto2D.h"
#include "indexed_img.h"
#include <stdlib.h>



/* *************************  LECTURE  ****************************** */
/** \ingroup indexed_img
    lecture des parametres de l'operateur indexed image. \\
    nrcouches : le nombre de couleur pour chaque composante r, v, b  de la tabele de couleure associe a l'image indexdee \\
    nrepochs : le nombre de iteration pour le rafinage des poids 
    @name indexed_img_lect
    @param des pointeur descripteur     
    @param ptp pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author Benone Ionescu
    @version 1.0 (10/12/00); Include : image.h, proto2D.h, indexed_img.h
*/

param *indexed_img_lect(kohonenstruct *des, param *ptp, char *debq)
{
  char question[500];

  sprintf(question, "%s Nombre de couches - val. max 256 (entier)", debq); 
  lec_param(question, ptp);
  des->nrcouches = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s Nombre d'epoques (entier)", debq); 
  lec_param(question, ptp);
  des->nrepochs = atoi(ptp->rep);
  ptp = ptp->next;
   
  return(ptp);
}


/* *************************  INITIALISATION  ************************** */
/** \ingroup indexed_img
    initialisation de l'operateur indexed image.\\
    allocation image resultat 
    allocation structure kohonennet
    @name indexed_img_init 
    @param des  pointeur descripteur
    @param im0  image initiale (taille connue)
    @param imres  pointeur image resultat type u1 (8 bit)
    @author Benone Ionescu
    @version 1.0 (10/12/00); Include : image.h, proto2D.h, indexed_img.h
*/

int indexed_img_init(kohonenstruct *des, imarvb im0, imau1 *imres)
{   
  register long int totalElements, i;
    
  totalElements = 0;
  totalElements += im0.nr * im0.nc;  

  imres->nr = im0.nr;
  imres->nc = im0.nc;
  alloc_imau1(imres);
  alloc_lut_u12(imres);

  des->Dist.nr = des->nrcouches;
  des->Dist.nc = 1;
  alloc_imadb(&des->Dist);

  des->Weight1.nr = 3;
  des->Weight1.nc = des->nrcouches;  
  alloc_imadb(&des->Weight1);
  
  des->Weight2.nr = 3;
  des->Weight2.nc = des->nrcouches;  
  alloc_imadb(&des->Weight2);

  des->TrainingData.nr = 3;
  des->TrainingData.nc = totalElements;
  alloc_imadb(&des->TrainingData);

  des->T.nr = des->nrcouches;
  des->T.nc = 1;
  alloc_imau1(&des->T);
  

}

/* *************************  CALCUL  ************************** */
/** \ingroup indexed_img
    calcul de l'operateur indexed image.\\
    Calcul de valeurs pour l'image indexee de type u1.
   
    On construit l'ensemble de donnes pour l'entrainement en partant de l'image originale rvb (initialisation de poids)
    Apres on aplique l'algorithme de apprentisage et on rafine les poids. L'algorithme est aplique pour tout les composantes de lot d'enrainement pendent le nombre d'epoques. Dans chaque etape de rafinement le vecteur de poids obtenu est compare avec le vecteur de poids obtenu dans l'etape precedente. L'algorithm se finie apres un nombre d'iteration imposee par le  nombre d'epochs. Les poids rafinees sont chargees dans la tabel de couleur LUT de l'image indexee de 8 bits.

    @name indexed_img_calc  
    @param des  pointeur descripteur
    @param im0  image initiale 
    @param imres  pointeur image resultat (type u1 : 8 bit)
    @author B. Ionescu
    @version 1.0 (10/12/00); Include : image.h, proto2D.h, indexed_img.h
*/

int indexed_img_calc(kohonenstruct *des, imarvb im0, imau1 *imres)
{
    register int l, c, i, j, x, y, k, t, left, center, right, neighbourhood;
    double minf, auxf = 0.0, w2_w1, sigma, learningRatio;
    double sigma_max, sigma_min, a, b, q, learn0, smin, smax;
    register long int currImage, totalElements;

    pixu1 rouge, vert, bleu;
   
    FILE *fStream;
    char *filename = "WEIGHTS.TXT";    
    unsigned char tampon = 0;

    /* ***************************************** */
    /*  Construction de l'ensemble de donnes pour l'entrainement  */

  printf("\nOn cree l'ensemble des donnees pour l'entrainement ..."); 

  i = 0;
     for(y = 0; y < im0.nc; y++)
       for(x = 0; x < im0.nr; x++){
           des->TrainingData.p[0][i] = im0.r[x][y]; 
           des->TrainingData.p[1][i] = im0.v[x][y];
           des->TrainingData.p[2][i] = im0.b[x][y];
           i++;
       }

  // initialization des poids
  j = 0;
  for(i = 0; i < des->nrcouches ; i++) {
     des->Weight1.p[0][i] = des->TrainingData.p[0][j];
     des->Weight1.p[1][i] = des->TrainingData.p[1][j];
     des->Weight1.p[2][i] = des->TrainingData.p[2][j];
     j += (int)((pixdb)(des->TrainingData.nc) / (pixdb)(des->nrcouches));
  }
 
  printf("fait\n");

  /* ***************************************** */
  /* Antrainer la reseau  */
  
  k = t = 0;
  printf("\nOn entraine le reseau ...\n\n");
  smax = -(des->nrcouches)/ (2 * log(0.1));
  sigma_max = sqrt(smax);
  smin = -1 / log(0.1);
  sigma_min = sqrt(smin);
  a = (des->nrcouches / 2 - 1) / (sigma_max - sigma_min);
  b = (sigma_max - (des->nrcouches) * sigma_min / 2) / (sigma_max - sigma_min);
  q = (sigma_max - sigma_min) / des->nrepochs;
  sigma = sigma_max;
  learn0 = (pixdb)(des->TrainingData.nr) / 4 + 4;
  
  do {

    srand(0);
    for(i = 0; i < (des->TrainingData.nc); i += (int)(150 *(double)rand() / RAND_MAX)) {

       printf("%cEpoque %d / %d en cours...", 13, k + 1, des->nrepochs);
       fflush(stdout);
  
       // Calcule les distances 
       for(c = 0; c < des->nrcouches; c++) {
         des->Dist.p[0][c] = 0.0;
         des->Dist.p[0][c] += (des->TrainingData.p[0][i] - des->Weight1.p[0][c]) * (des->TrainingData.p[0][i] - des->Weight1.p[0][c]);
         des->Dist.p[0][c] += (des->TrainingData.p[1][i] - des->Weight1.p[1][c]) * (des->TrainingData.p[1][i] - des->Weight1.p[1][c]);      
         des->Dist.p[0][c] += (des->TrainingData.p[2][i] - des->Weight1.p[2][c]) * (des->TrainingData.p[2][i] - des->Weight1.p[2][c]);      
         des->Dist.p[0][c] = sqrt(des->Dist.p[0][c]);
	    }

       // finding the best matched node
       center = des->nrcouches; minf = des->Dist.p[0][0];
       for(c = 1; c < des->nrcouches; c++)
	 if(minf > des->Dist.p[0][c]){
           minf = des->Dist.p[0][c];
           center = c + des->nrcouches;
	 } 
       
       // updating the weights
       des->T.p[0][center - (des->nrcouches)]; /*  = des->T.p[0][center - (des->nrcouches)]++; */
       neighbourhood = (int)(a * sigma + b);
       learningRatio = learn0 / (learn0 + des->T.p[0][center - (des->nrcouches)]);
       left = center - neighbourhood;
       right = center + neighbourhood;
       
       for(c = left; c <= right; c++) {
          des->Weight2.p[0][c % des->nrcouches] = des->Weight1.p[0][c % des->nrcouches] + learningRatio * exp(((double)c - center) * (center - c) / (2 * sigma *sigma)) * (des->TrainingData.p[0][i] - des->Weight1.p[0][c % des->nrcouches]);
          des->Weight2.p[1][c % des->nrcouches] = des->Weight1.p[1][c % des->nrcouches] + learningRatio * exp(((double)c - center) * (center - c) / (2 * sigma *sigma)) * (des->TrainingData.p[1][i] - des->Weight1.p[1][c % des->nrcouches]);
          des->Weight2.p[2][c % des->nrcouches] = des->Weight1.p[2][c % des->nrcouches] + learningRatio * exp(((double)c - center) * (center - c) / (2 * sigma *sigma)) * (des->TrainingData.p[2][i] - des->Weight1.p[2][c % des->nrcouches]);
       }
       
       //updating the weights
	   w2_w1 = ceil(des->Weight2.p[0][0] - des->Weight1.p[0][0]);

       for(c = 0; c < des->nrcouches; c++){
           des->Weight1.p[0][c] = des->Weight2.p[0][c];
           des->Weight1.p[1][c] = des->Weight2.p[1][c];
           des->Weight1.p[2][c] = des->Weight2.p[2][c];
       } 
       t++;
       
    } 
    k++;
    if((auxf = sigma - q) < sigma_min)
        sigma = sigma_min;
    else
        sigma = auxf;
   
  } while(k < des->nrepochs);
 
  printf("fait\n"); 
  

  /* ****************************************** */
  /*   sauvegarder les poids dans fichier   */
   
   if (fStream = fopen(filename, "w")) {
      for (j = 0; j < des->nrcouches; j++)
	fprintf(fStream, "%8.4f  %8.4f  %8.4f\n", des->Weight1.p[0][j], des->Weight1.p[1][j], des->Weight1.p[2][j]);
    fclose(fStream);
   }



   /* ***************************************** */
   /* Construction de l'image indexee  */

   printf("\nOn cree l'image indexee ...");

      for(c = 0; c < des->nrcouches; c++){
	   imres->lutr[c] = (pixu1)(ceil(des->Weight2.p[0][c])); 
           imres->lutv[c] = (pixu1)(ceil(des->Weight2.p[1][c]));
	   imres->lutb[c] = (pixu1)(ceil(des->Weight2.p[2][c]));
        }


   for(x = 0; x < im0.nr; x++)
     for(y = 0; y < im0.nc; y++) {

        // calcule les distances
        for(c = 0; c < des->nrcouches; c++) {
           des->Dist.p[0][c] = 0.0;
           des->Dist.p[0][c] += (im0.r[x][y] - des->Weight2.p[0][c]) * (im0.r[x][y] - des->Weight2.p[0][c]);
           des->Dist.p[0][c] += (im0.v[x][y] - des->Weight2.p[1][c]) * (im0.v[x][y] - des->Weight2.p[1][c]);
           des->Dist.p[0][c] += (im0.b[x][y] - des->Weight2.p[2][c]) * (im0.b[x][y] - des->Weight2.p[2][c]);
           des->Dist.p[0][c] = sqrt(des->Dist.p[0][c]);
	}
        
       // find the best node
       center = 0.0;
       minf = des->Dist.p[0][0];
       for(c = 1; c < des->nrcouches; c++)
	 if(minf > des->Dist.p[0][c]) {
              minf = des->Dist.p[0][c];
              center = c;
	 }
       imres->p[x][y] = (unsigned char)center;
      }
     printf("fait \n");   
   
   free_imau1(&des->T); 
   free_imadb(&des->TrainingData);
   free_imadb(&des->Weight2);
   free_imadb(&des->Dist);
}


int alloc_lut_u12(imau1 *im)
{
   if( im->lutr!=NULL && im->lutv!=NULL && im->lutb!=NULL ){
	free(im->lutr);
	free(im->lutv);
	free(im->lutb);
    }    	


	if( (im->lutr = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL ||
	    (im->lutv = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL ||
	    (im->lutb = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL )
	  {
	  printf ("lutr, lutv, lutb: ERREUR allocation impossible\n");  
	  exit(1); 
	  }
}

int free_lut_u12(imau1 *im)
{
  free(im->lutb); im->lutb = NULL;
  free(im->lutv); im->lutv = NULL;
  free(im->lutr); im->lutr = NULL;
  
}

