/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup filtrage
    \defgroup kuan operateur filtre de Kuan

Moyenne ponderee entre la valeur du pixel et celle de la moyenne local en fonction 
de la variance locale et de celle du bruit multiplicatif.\\
ATTENTION : les hypotheses correspondent aux images ROS en intensite (module carre). 
Il faut ommencer par elever au carre les donnees d'amplitude de type PRI-ESA (.imw).
\\
Descripteur : \\
   int dimX, dimY   :  dimension de la fenetre de moyennage \\
   double nbvues    : nombre de vues (ou equivalent) \\
\\
Type entree : imau2 | fl | db\\
Type sortie : imau2 | fl | db (respectivement)
@name lkuan
@author E. Trouve
@version 1.0 (21/03/2000); Include : image.h, proto2D.h, radar.h 
@see mkuan_u2
*/


#include "image.h"
#include "proto2D.h"
#include "radar.h"
#include <stdlib.h>
#include <fcntl.h>
#include <limits.h> /* pour  USHRT_MAX ... */


/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup kuan
    lecture des parametres de l'operateur kuan. \\
    dimX : nombre de colonnes de la fenetre de moyennage (impaire) \\
    dimY : nombre de lignes de la fenetre de moyennage (impaire)   \\
    
    @name kuan_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (28/02/2000); Include : image.h, proto2D.h, radar.h
*/
param * kuan_lect(kuan_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s fenetre d'estimation, nombre de colonnes (impaire)", debq); 
  lec_param(question, ptp);
  des->dimX = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s fenetre d'estimation, nombre de lignes (impaire)", debq); 
  lec_param(question, ptp);
  des->dimY = atoi(ptp->rep);
  ptp = ptp->next;

  if( des->dimX%2 == 0 || des->dimY%2 == 0 || des->dimX<=0 || des->dimY<=0){
      printf("\n>> ERREUR kuan_lect : dimensions incorrectes\n");
      exit(1);
  }

  sprintf(question, "%s nombre de vues (equivalent) de l'image initiale", debq); 
  lec_param(question, ptp);
  des->nbvue = (double)atof(ptp->rep);
  ptp = ptp->next;

  return(ptp);
}


/* **********************  INITIALISATION  *****************************/
/** \ingroup kuan
    XX = u2 | fl | db : initialisation de l'operateur kuan.\\
    allocation image resultat (inutil si calcul sur place ie image resultat = image entree)\\
    @name kuanXX_init
    @param des : pointeur descripteur
    @param im0 : image initiale (taille connue)
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (28/02/2000); Include : image.h, proto2D.h, radar.h
*/

/* initialisation (16 bit)*/
int kuanu2_init(kuan_t *des, imau2 im0, imau2 *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(kuan%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imau2(imres);
  return(0);  
}

/* initialisation (float)*/
int kuanfl_init(kuan_t *des, imafl im0, imafl *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(kuan%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imafl(imres);
  return(0);  
}

/* initialisation (double)*/
int kuandb_init(kuan_t *des, imadb im0, imadb *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(kuan%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imadb(imres);
  return(0);  
}

/* *************************  CALCUL  ***************************/
/** \ingroup kuan
    XX = u2 | fl | db : calcul de l'operateur kuan.\\
    Le calcul sur place est possible : passer la meme image en entree et en sortie.
    @name kuanXX_calc
    @param des : pointeur descripteur
    @param im0 : image initiale (ATT : intensite)
    @param moyint : image des moyennes locale de l'intensite
    @param moycarre : image des moyenne locale du carre de l'intensite
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (21/03/2000); Include : image.h, proto2D.h, radar.h
*/

/* calcul (16 bit) */
int kuanu2_calc(kuan_t *des, imau2 im0, imau2 moyint, imafl moycarre, imau2 *imres){
  int i, j, nbneg = 0;
  double cv2, resdb;
  
  /* coef de variation (au carre) du speckle dans une image d'intensite n_vues */ 
  cv2 = 1. / des->nbvue; 

  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++){
	  resdb = filt_kuan((double)im0.p[j][i], (double)moyint.p[j][i],
			    (double)moycarre.p[j][i], cv2, &nbneg) + 0.5;
	  if( resdb < 0 ){
	      printf("\n ATT kuanu2_calc : resultat negatif!\n");
	      imres->p[j][i] = 0;
	  }
	  else if (resdb >= USHRT_MAX+1){
	      printf("\n ATT kuanu2_calc : resultat superieur a %d\n", USHRT_MAX);
	      imres->p[j][i] = USHRT_MAX;
	  }
	  else
	      imres->p[j][i] = (pixu2)resdb;
      }
  printf("\n ATT kuanu2_calc : %d points tels que CV_image < CV_speckle\n", nbneg);   
  return(0);
}

/* calcul (float) */
int kuanfl_calc(kuan_t *des, imafl im0, imafl moyint, imafl moycarre, imafl *imres){
  int i, j, nbneg = 0;
  double cv2, resdb;
  
  /* coef de variation (au carre) du speckle dans une image d'intensite n_vues */ 
  cv2 = 1. / des->nbvue; 

  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++)
	  imres->p[j][i] = (pixfl)filt_kuan((double)im0.p[j][i], (double)moyint.p[j][i],
			    (double)moycarre.p[j][i], cv2, &nbneg);
  printf("\n ATT kuanfl_calc : %d points tels que CV_image < CV_speckle\n", nbneg);   
  return(0);
}

/* calcul (double) */
int kuandb_calc(kuan_t *des, imadb im0, imadb moyint, imadb moycarre, imadb *imres){
  int i, j, nbneg = 0;
  double cv2, resdb;
  
  /* coef de variation (au carre) du speckle dans une image d'intensite n_vues */ 
  cv2 = 1. / des->nbvue; 

  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++)
	  imres->p[j][i] = (pixdb)filt_kuan((double)im0.p[j][i], (double)moyint.p[j][i],
			    (double)moycarre.p[j][i], cv2, &nbneg);
  printf("\n ATT kuandb_calc : %d points tels que CV_image < CV_speckle\n", nbneg);   
  return(0);
}

/* fonction : ******************************* */
double filt_kuan(double int0, double moyint, double moycarre, double cv2, int* nbneg){
  double  ci2, cr2;

  ci2 = moycarre/(moyint*moyint) - 1.; /* ci2 = sigma2 / mu^2, sigma2 = moycarre - mu^2 */
  cr2 = (ci2 - cv2)/(1. + cv2);
  if(cr2 < 0){
      /* printf("\n ATT kuanu2_calc : CV_image^2=%f < CV_speckle^2=%f\n", ci2, cv2); */
      *nbneg = *nbneg + 1;
      return(moyint);
  }
  else
      return(moyint + (int0 - moyint) * cr2 / ci2);
} 

