/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup filtrage
    \defgroup lee_add operateur filtre statistique de Lee pour bruit additif.

Filtrage realise a partir de la valeur du pixel (v0), de la moyenne mu et de la 
variance locale sigma_im connaissant la variance du bruit additif sigma_br selon :\\ 
v_res = mu + (v0 - mu)*(sigma_im - sigma_br)/sigma_im\\
Ce filtre est notamment utilise comme 2eme operateur du filtre multidate 
frequentiel de Daniela Coltuc pour filter les canaux issus de la DCT  ou le 
bruit est de nature additive.\\
\\
Descripteur : \\
   int dimX, dimY   :  dimension de la fenetre de moyennage \\
   double vbr       :  variance du bruit additif dans l'image initiale \\
\\
Type entree : imau1 | u2 | fl | db\\
Type sortie : imau1 | u2 | fl | db (respectivement)
@name llee_add
@author E. Trouve
@version 1.0 (11/09/2000); Include : image.h, proto2D.h, radar.h 
@see mlee_add, mfiltmdate_dct
*/

#include "image.h"
#include "proto2D.h"
#include "radar.h"
#include <stdlib.h>
#include <fcntl.h>
#include <limits.h> /* pour  USHRT_MAX ... */

/* #include "radar_comp.h" */

double filt_lee_add(double amp0, double moyamp, double moycarre, double vbr, int* nbneg);

/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup lee_add 
    lecture des parametres de l'operateur lee_add. \\
    dimX : nombre de colonnes de la fenetre de moyennage (impaire) \\
    dimY : nombre de lignes de la fenetre de moyennage (impaire)   \\
    vbr : variance du bruit additif dans l'image initiale \\
    
    @name lee_add_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (11/09/2000); Include : image.h, proto2D.h, radar.h
*/
param * lee_add_lect(lee_add_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s fenetre d'estimation, nombre de colonnes (impaire)", debq); 
  lec_param(question, ptp);
  des->dimX = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s fenetre d'estimation, nombre de lignes (impaire)", debq); 
  lec_param(question, ptp);
  des->dimY = atoi(ptp->rep);
  ptp = ptp->next;

  if( des->dimX%2 == 0 || des->dimY%2 == 0 || des->dimX<=0 || des->dimY<=0){
      printf("\n>> ERREUR lee_add_lect : dimensions incorrectes\n");
      exit(1);
  }

  sprintf(question, "%s variance du bruit dans l'image initiale", debq); 
  lec_param(question, ptp);
  des->vbr = (double)atof(ptp->rep);
  ptp = ptp->next;

  return(ptp);
}


/* **********************  INITIALISATION  *****************************/
/** \ingroup lee_add
    XX = u1 | u2 | fl | db : initialisation de l'operateur lee_add.\\
    allocation image resultat \\
    @name lee_addXX_init
    @param des : pointeur descripteur
    @param im0 : image initiale (taille connue)
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (28/02/2000); Include : image.h, proto2D.h, radar.h
*/

/* initialisation (8 bit)*/
int lee_addu1_init(lee_add_t *des, imau1 im0, imau1 *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(lee_add%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imau1(imres);
  return(0);  
}

/* initialisation (16 bit)*/
int lee_addu2_init(lee_add_t *des, imau2 im0, imau2 *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(lee_add%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imau2(imres);
  return(0);  
}

/* initialisation (float)*/
int lee_addfl_init(lee_add_t *des, imafl im0, imafl *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(lee_add%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imafl(imres);
  return(0);  
}

/* initialisation (double)*/
int lee_adddb_init(lee_add_t *des, imadb im0, imadb *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(lee_add%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imadb(imres);
  return(0);  
}

/* *************************  CALCUL  ***************************/
/** \ingroup lee_add
    XX = u1 | u2 | fl | db : calcul de l'operateur lee_add.\\
    @name lee_addXX_calc
    @param des : pointeur descripteur
    @param im0 : image initiale (amplitude)
    @param moyamp : image des moyennes locale de l'amplitude
    @param moycarre : image des moyenne locale de l'intensite (carre de l'amplitude)
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (21/11/99); Include : image.h, proto2D.h, radar.h
*/

/* calcul (8 bit) */
int lee_addu1_calc(lee_add_t *des, imau1 im0, imau1 moyamp, imafl moycarre, imau1 *imres){
  int i, j, nbneg = 0;
  double  resdb;
  
  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++){
	  resdb = filt_lee_add((double)im0.p[j][i], (double)moyamp.p[j][i],
			    (double)moycarre.p[j][i], des->vbr, &nbneg) + 0.5;
	  if( resdb < 0 ){
	      printf("\n ATT lee_addu1_calc : resultat negatif!\n");
	      imres->p[j][i] = 0;
	  }
	  else if (resdb >= 256.){
	      printf("\n ATT lee_addu1_calc : resultat superieur a 256\n");
	      imres->p[j][i] = 255;
	  }
	  else
	      imres->p[j][i] = (pixu1)resdb;
      }
  printf("\n ATT lee_addu1_calc : %d points tels que var_image < var_bruit\n", nbneg);   
  return(0);
}

/* calcul (16 bit) */
int lee_addu2_calc(lee_add_t *des, imau2 im0, imau2 moyamp, imafl moycarre, imau2 *imres){
  int i, j, nbneg = 0;
  double  resdb;
  
  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++){
	  resdb = filt_lee_add((double)im0.p[j][i], (double)moyamp.p[j][i],
			    (double)moycarre.p[j][i], des->vbr, &nbneg) + 0.5;
	  if( resdb < 0 ){
	      printf("\n ATT lee_addu2_calc : resultat negatif!\n");
	      imres->p[j][i] = 0;
	  }
	  else if (resdb >= USHRT_MAX+1){
	      printf("\n ATT lee_addu2_calc : resultat superieur a %d\n", USHRT_MAX);
	      imres->p[j][i] = USHRT_MAX;
	  }
	  else
	      imres->p[j][i] = (pixu2)resdb;
      }
  printf("\n ATT lee_addu2_calc : %d points tels que var_image < var_bruit \n", nbneg);   
  return(0);
}

/* calcul (float) */
int lee_addfl_calc(lee_add_t *des, imafl im0, imafl moyamp, imafl moycarre, imafl *imres){
  int i, j, nbneg = 0;
  
  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++)
	  imres->p[j][i] = (pixfl)filt_lee_add((double)im0.p[j][i], (double)moyamp.p[j][i],
			    (double)moycarre.p[j][i], des->vbr, &nbneg);
  printf("\n ATT lee_addfl_calc : %d points tels que var_image < var_bruit\n", nbneg);   
  return(0);
}

/* calcul (double) */
int lee_adddb_calc(lee_add_t *des, imadb im0, imadb moyamp, imadb moycarre, imadb *imres){
  int i, j, nbneg = 0;
  
  for(j=0; j<im0.nr; j++)
      for(i=0; i<im0.nc; i++)
	  imres->p[j][i] = filt_lee_add((double)im0.p[j][i], (double)moyamp.p[j][i],
			    (double)moycarre.p[j][i], des->vbr, &nbneg);
  printf("\n ATT lee_adddb_calc : %d points tels que var_image < var_bruit\n", nbneg);   
  return(0);
}

/* fonction : ******************************* */
double filt_lee_add(double amp0, double moyamp, double moycarre, double vbr, int* nbneg){
  double  vim;  /* variance mesure (image bruite */
  double  vte;  /* variance de la texture (image non-bruitee) */ 

  vim = moycarre - moyamp*moyamp;   /* sigma2 = moycarre - mu^2 */
  vte = vim - vbr;
  if(vte < 0){
      /*  printf("\n ATT lee_addu2_calc : V_image=%f < V_bruit=%f\n", vim, vbr); */
      *nbneg = *nbneg + 1;
      return(moyamp);
  }
  else
      return(moyamp + (amp0 - moyamp) * vte / vim);
} 

