/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup filtrage 
    \defgroup logdct operateur logdct (logarithme et DCT sur axe temporel)

Premier operateur du filtre multidate frequentiel de Daniela Coltuc : \\
conversion de  N images RSO (espace initial) en N canaux de DCT (espace transforme).
On prend le log des valeurs initiales pour rendre le bruit additif,  
puis on calcule la DCT  en chaque point de l'image 2D.\\
\\
Descripteur : \\
   int N : nombre d'images de la serie  =  taille de la DCT \\
   float tablog[65536] : valeurs du log tabulees  \\
   float *v0 : vecteur pour stoker les points a transformer \\
   float **matdct : matrice de calcul de DCT N points, 1 ligne par coef \\
\\
Type entree : serie images u2 \\
Type sortie : serie images fl \\
@name llogdct
@author E. Trouve
@version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h 
@see lidctexp.c, mlogdct_u2.c mfiltmdate_dct.c
*/

#include "image.h"
#include "proto2D.h"
#include "radar.h"
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <limits.h> /* pour  USHRT_MAX ... */

/* #include "radar_comp.h" */



/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup logdct
    lecture des parametres de l'operateur logdct. \\
    N : nombre d'images de la serie = taille de la DCT\\
    
    @name logdct_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h
*/
param * logdct_lect(logdct_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s nombre d'images", debq); 
  lec_param(question, ptp);
  des->N = atoi(ptp->rep);
  ptp = ptp->next; 

  return(ptp);
}

/* **********************  INITIALISATION  *****************************/
/** \ingroup logdct
    XX =  u2  : initialisation de l'operateur logdct.\\
    allocation serie images resultat, tableau vecteur et resultat,  \\
    @name logdct_XX_init
    @param des : pointeur descripteur
    @param pt_im0 : pointeur serie images initiales (taille connue)
    @param pt_imres : pointeur serie images resultat
    @author E. Trouve
    @version 1.0 (09/09/2000); Include : image.h, proto2D.h, radar.h
*/

/* initialisation (8 bit)*/

/* initialisation (16 bit)*/
int logdct_u2_init(logdct_t *des, imau2 *pt_im0, imafl *pt_imres){
  int n, k, N;
  float ck;
  double tmpdb;

  N = des->N;
  /* allocation images resultat */
  for(n=0; n<N; n++){
      pt_imres[n].nc = pt_im0[0].nc;
      pt_imres[n].nr = pt_im0[0].nr;
      sprintf(pt_imres[n].nom, "%s_logdct_%d", pt_im0[0].nom, n);
      alloc_imafl(pt_imres+n);
  }

  /* allocation vecteur de donnees */
  if(  (des->v0 = (float *)malloc(N*sizeof(float))) == NULL ){
      printf ("\n ERREUR logdct_u2_init : allocation vecteur impossible\n");  
      exit(1); 
  }  

  /* allocation matrice coef de calcul de DCT */ 
  if(  (des->matdct = (float **)malloc(N*sizeof(float *))) == NULL ){
      printf ("\n ERREUR logdct_u2_init : allocation matrice impossible\n");  
      exit(1); 
  }  
  if( (des->matdct[0] = (float *)malloc(N*N*sizeof(float))) == NULL){
      printf ("\n ERREUR logdct_u2_init : allocation image impossible\n");  
      exit(1); 
  }
  for(k=1; k<N; k++)
        des->matdct[k] =  des->matdct[0] + k*N;

  /* calcul matrice coef de calcul de DCT */ 
  ck = (float)sqrt(1./(double)N);  /* k=0 */
  for(n=0; n<N; n++){
      des->matdct[0][n] = ck;
  }
  ck = (float)sqrt(2./(double)N);
  for(k=1; k<N; k++)
      for(n=0; n<N; n++)
	  des->matdct[k][n] = ck*(float)cos(M_PI*(double)((2*n+1)*k)/(double)(2*N));
  
  /* tabulation du log*/
  des->tablog[0] = 0. ;
  for(n=1; n<65536; n++)
      des->tablog[n] = (float)log((double)n);

  return(0);  
}

/* initialisation (float)*/

/* initialisation (double)*/


/* *************************  CALCUL  ***************************/
/** \ingroup logdct
    XX = u2 : calcul de l'operateur logdct.\\
    @name logdct_XX_calc
    @param des : pointeur descripteur
    @param pt_im0 : pointeur serie images initiales 
    @param pt_imres : pointeur serie images resultat
    @author E. Trouve
    @version 1.0 (09/09/00); Include : image.h, proto2D.h, radar.h
*/

/* calcul (8 bit) */

/* calcul (16 bit) */
int logdct_u2_calc(logdct_t *des, imau2 *pt_im0, imafl *pt_imres){
  int i, j, dimX, dimY, n, k, N;
  float som;
  
  N = des->N;
  dimX = pt_im0->nc; 
  dimY = pt_im0->nr; 
  for(j=0; j<dimY; j++)
      for(i=0; i<dimX; i++){
	  for( n=0; n<N; n++)
	      des->v0[n] = des->tablog[pt_im0[n].p[j][i]];
	  for( k=0; k<N; k++){
	      som = des->matdct[k][0] * des->v0[0];
	      for( n=1; n<N; n++)
		  som += des->matdct[k][n] * des->v0[n];
	      pt_imres[k].p[j][i] = som;
	  }
      }
  return(0);
}

/* calcul (float) */

/* calcul (double) */

/* fonction : ******************************* */


