/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \ingroup filtrage
    \defgroup moy2D operateur moyenne 2D glissante.
Moyenne glissante ligne puis colonne avec moyenne exacte 
des points disponibles sur les bords.\\
\\
Descripteur : \\
   int dimX, dimY   :  dimension de la fenetre de moyennage \\
   pixu4 *tmpcolu4  :  colonne de travail pour u1, u2 \\
   double *tmpcoldb;  colonne de travail pour fl, db \\
\\
Type entree : imau1 | u2 | fl | db\\
Type sortie : imau1 | u2 | fl | db (respectivement)
@name lmoy2D
@author E. Trouve
@version 1.0 (25/11/99); Include : image.h, proto2D.h 
@see  mmoy2D
*/

#include "image.h"
#include "proto2D.h"
#include <stdlib.h>
#include <fcntl.h>


/* **********************  LECTURE DE PARAMETRES   *************************/
/** \ingroup moy2D
    lecture des parametres de l'operateur moy2D. \\
    dimX : nombre de colonnes de la fenetre de moyennage (impaire) \\
    dimY : nombre de lignes de la fenetre de moyennage (impaire)   \\
    @name moy2D_lect
    @param   des  pointeur descripteur	  
    @param   ptp  pointeur sur parametre courant deja alloue 
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue 
    @author  E. Trouve
    @version 1.0 (25/11/99); Include : image.h, proto2D.h
*/
param * moy2D_lect(moy2D_t *des, param *ptp, char *debq){
  char question[500];

  sprintf(question, "%s nombre de colonnes moyennees (impaire)", debq); 
  lec_param(question, ptp);
  des->dimX = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, "%s nombre de lignes moyennees (impaire)", debq); 
  lec_param(question, ptp);
  des->dimY = atoi(ptp->rep);
  ptp = ptp->next;

  if( ((des->dimX*des->dimY)%2 == 0) || des->dimX<=0 || des->dimY<=0){
      printf("\n>> ERREUR moy2D_lect : dimensions incorrectes\n");
      exit(1);
  }

  return(ptp);
}


/* **********************  INITIALISATION  *****************************/
/** \ingroup moy2D
    XX = u1 | u2 | fl | db : initialisation de l'operateur moy2D.\\
    allocation image resultat et colonne de travail\\
    @name moy2DXX_init
    @param des : pointeur descripteur
    @param im0 : image initiale (taille connue)
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (21/11/99); Include : image.h, proto2D.h
*/
int moy2Du1_init(moy2D_t *des, imau1 im0, imau1 *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(moy%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imau1(imres);
  /* colonne de travail */
  if( (des->tmpcolu4=(pixu4 *)malloc(im0.nr*sizeof(pixu4))) == NULL ){
	  printf("\n>> ERREUR moy2Du1_init : allocation impossible\n");
	  exit(1);
  }
  return(0);  
}

/* initialisation (16 bit)*/
int moy2Du2_init(moy2D_t *des, imau2 im0, imau2 *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(moy%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imau2(imres);
  /* colonne de travail */
  if( (des->tmpcolu4=(pixu4 *)malloc(im0.nr*sizeof(pixu4))) == NULL ){
	  printf("\n>> ERREUR moy2Du2_init : allocation impossible\n");
	  exit(1);
  }
  return(0);  
}

/* initialisation (float) */
int moy2Dfl_init(moy2D_t *des, imafl im0, imafl *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(moy%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imafl(imres);
  /* colonne de travail */
  if( (des->tmpcoldb=(pixdb *)malloc(im0.nr*sizeof(pixdb))) == NULL ){
	  printf("\n>> ERREUR moy2Dfl_init : allocation impossible\n");
	  exit(1);
  }
  return(0);  
}

/* initialisation (double) */
int moy2Ddb_init(moy2D_t *des, imadb im0, imadb *imres){
  /* image resultat */ 
  imres->nc = im0.nc;
  imres->nr = im0.nr;
  sprintf(imres->nom, "%s(moy%dx%d)", im0.nom, des->dimX, des->dimY);
  alloc_imadb(imres);
  /* colonne de travail */
  if( (des->tmpcoldb=(pixdb *)malloc(im0.nr*sizeof(pixdb))) == NULL ){
	  printf("\n>> ERREUR moy2Ddb_init : allocation impossible\n");
	  exit(1);
  }
  return(0);  
}


/* *************************  CALCUL  ***************************/
/** \ingroup moy2D
    XX = u1 | u2 | fl | db : calcul de l'operateur moy2D.\\
    Calcul glissant des moyennes ligne puis colonne sur toute l'image
    avec moyenne exacte des points disponibles sur les bords. 
    @name moy2DXX_calc
    @param des : pointeur descripteur
    @param im0 : image initiale 
    @param imres : pointeur image resultat
    @author E. Trouve
    @version 1.0 (21/11/99); Include : image.h, proto2D.h
*/
int moy2Du1_calc(moy2D_t *des, imau1 im0, imau1 *imres){
  pixu1 *ptu1, *ptres;
  int i, j, nc, nr, nbp, dlfX, dlfY;
  pixu4 tmpu4, somu4, *ptu4;
  double tmpdb;
  
  nc = im0.nc; 
  nr = im0.nr; 
  dlfX = des->dimX/2; /* demi largeur de fenetre en X */
  dlfY = des->dimY/2; /* demi largeur de fenetre en Y */

  /* moyenne ligne */
  for(j=0; j<nr; j++){
      nbp = 0;
      somu4 = 0;
      ptu1 = im0.p[j];
      ptres = imres->p[j];
      for(i=0; i<dlfX; i++){
	  somu4 += (pixu4)ptu1[i];
	  nbp++;
      }
      for(i=0; i<dlfX; i++){
	  somu4 += (pixu4)ptu1[i+dlfX];
	  nbp++;
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu1)tmpdb; 
      } 
      nbp++;
      for(  ; i<nc-dlfX; i++){
	  somu4 += (pixu4)ptu1[i+dlfX];
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu1)tmpdb; 
	  somu4 -= (pixu4)ptu1[i-dlfX];
      }
      nbp--;
      for(  ; i<nc; i++){
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu1)tmpdb; 
	  somu4 -= (pixu4)ptu1[i-dlfX];
	  nbp--;  
      }
  }

  /* moyenne colonne */     
  ptu4 = des->tmpcolu4;
  for(i=0; i<nc; i++){
      /* recopie */
      for(j=0; j<nr; j++)
	  ptu4[j] = (pixu4)imres->p[j][i];
      nbp = 0;
      somu4 = 0;
      for(j=0; j<dlfY; j++){
	  somu4 += ptu4[j];
	  nbp++;
      }
      for(j=0; j<dlfY; j++){
	  somu4 += ptu4[j+dlfY];
	  nbp++;
	  tmpdb = (double)somu4/nbp + 0.5;
	  imres->p[j][i] = (pixu1)tmpdb;  
      } 
      nbp++;
      for(  ; j<nr-dlfY; j++){
	  somu4 += ptu4[j+dlfY];
	  tmpdb = (double)somu4/nbp + 0.5; 	  
	  imres->p[j][i] = (pixu1)tmpdb;  
	  somu4 -= ptu4[j-dlfY];
      }
      nbp--;
      for(  ; j<nr; j++){
	  tmpdb = (double)somu4/nbp + 0.5; 	  
	  imres->p[j][i] = (pixu1)tmpdb;  
	  somu4 -= ptu4[j-dlfY];
	  nbp--;  
      }
  }
  return(0);
}

/* calcul (16 bit) */
int moy2Du2_calc(moy2D_t *des, imau2 im0, imau2 *imres){
  pixu2 *ptu2, *ptres;
  int i, j, nc, nr, nbp, dlfX, dlfY;
  pixu4 tmpu4, somu4, *ptu4;
  double tmpdb;
  
  nc = im0.nc; 
  nr = im0.nr; 
  dlfX = des->dimX/2; /* demi largeur de fenetre en X */
  dlfY = des->dimY/2; /* demi largeur de fenetre en Y */
 
  /* moyenne ligne */
  for(j=0; j<nr; j++){
      nbp = 0;
      somu4 = 0;
      ptu2 = im0.p[j];
      ptres = imres->p[j];
      for(i=0; i<dlfX; i++){
	  somu4 += (pixu4)ptu2[i];
	  nbp++;
      }
      for(i=0; i<dlfX; i++){
	  somu4 += (pixu4)ptu2[i+dlfX];
	  nbp++;
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu2)tmpdb; 
      } 
      nbp++;
      for(  ; i<nc-dlfX; i++){
	  somu4 += (pixu4)ptu2[i+dlfX];
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu2)tmpdb; 
	  somu4 -= (pixu4)ptu2[i-dlfX];
      }
      nbp--;
      for(  ; i<nc; i++){
	  tmpdb = (double)somu4/nbp + 0.5;
	  ptres[i] = (pixu2)tmpdb; 
	  somu4 -= (pixu4)ptu2[i-dlfX];
	  nbp--;  
      }
  }

  /* moyenne colonne */     
  ptu4 = des->tmpcolu4;
  for(i=0; i<nc; i++){
      /* recopie */
      for(j=0; j<nr; j++)
	  ptu4[j] = (pixu4)imres->p[j][i];
      nbp = 0;
      somu4 = 0;
      for(j=0; j<dlfY; j++){
	  somu4 += ptu4[j];
	  nbp++;
      }
      for(j=0; j<dlfY; j++){
	  somu4 += ptu4[j+dlfY];
	  nbp++;
	  tmpdb = (double)somu4/nbp + 0.5;
	  imres->p[j][i] = (pixu2)tmpdb;  
      } 
      nbp++;
      for(  ; j<nr-dlfY; j++){
	  somu4 += ptu4[j+dlfY];
	  tmpdb = (double)somu4/nbp + 0.5; 	  
	  imres->p[j][i] = (pixu2)tmpdb;  
	  somu4 -= ptu4[j-dlfY];
      }
      nbp--;
      for(  ; j<nr; j++){
	  tmpdb = (double)somu4/nbp + 0.5; 	  
	  imres->p[j][i] = (pixu2)tmpdb;  
	  somu4 -= ptu4[j-dlfY];
	  nbp--;  
      }
  }
  return(0);
}

/* calcul (float) */
int moy2Dfl_calc(moy2D_t *des, imafl im0, imafl *imres){
  pixfl *ptfl, *ptres;
  int i, j, nc, nr, nbp, dlfX, dlfY;
  pixdb  somdb, *ptdb;
  
  nc = im0.nc; 
  nr = im0.nr; 
  dlfX = des->dimX/2; /* demi largeur de fenetre en X */
  dlfY = des->dimY/2; /* demi largeur de fenetre en Y */
 
  /* moyenne ligne */
  for(j=0; j<nr; j++){
      nbp = 0;
      somdb = 0.;
      ptfl = im0.p[j];
      ptres = imres->p[j];
      for(i=0; i<dlfX; i++){
	  somdb += (pixdb)ptfl[i];
	  nbp++;
      }
      for(i=0; i<dlfX; i++){
	  somdb += (pixdb)ptfl[i+dlfX];
	  nbp++;
	  ptres[i] = (pixfl)somdb/nbp; 
      } 
      nbp++;
      for(  ; i<nc-dlfX; i++){
	  somdb += (pixdb)ptfl[i+dlfX];
	  ptres[i] = (pixfl)somdb/nbp; 
	  somdb -= (pixdb)ptfl[i-dlfX];
      }
      nbp--;
      for(  ; i<nc; i++){
	  ptres[i] = (pixfl)somdb/nbp; 
	  somdb -= (pixdb)ptfl[i-dlfX];
	  nbp--;  
      }
  }

  /* moyenne colonne */     
  ptdb = des->tmpcoldb;
  for(i=0; i<nc; i++){
      /* recopie */
      for(j=0; j<nr; j++)
	  ptdb[j] = (pixdb)imres->p[j][i];
      nbp = 0;
      somdb = 0.;
      for(j=0; j<dlfY; j++){
	  somdb += ptdb[j];
	  nbp++;
      }
      for(j=0; j<dlfY; j++){
	  somdb += ptdb[j+dlfY];
	  nbp++;
	  imres->p[j][i] = (pixfl)somdb/nbp;  
      } 
      nbp++;
      for(  ; j<nr-dlfY; j++){
	  somdb += ptdb[j+dlfY];
	  imres->p[j][i] = (pixfl)somdb/nbp;  
	  somdb -= ptdb[j-dlfY];
      }
      nbp--;
      for(  ; j<nr; j++){
	  imres->p[j][i] = (pixfl)somdb/nbp;  
	  somdb -= ptdb[j-dlfY];
	  nbp--;  
      }
  }
  return(0);
}

/* calcul (double) */
int moy2Ddb_calc(moy2D_t *des, imadb im0, imadb *imres){
  pixdb *ptdb, *ptres;
  int i, j, nc, nr, nbp, dlfX, dlfY;
  pixdb  somdb;
  
  nc = im0.nc; 
  nr = im0.nr; 
  dlfX = des->dimX/2; /* demi largeur de fenetre en X */
  dlfY = des->dimY/2; /* demi largeur de fenetre en Y */
 
  /* moyenne ligne */
  for(j=0; j<nr; j++){
      nbp = 0;
      somdb = 0.;
      ptdb = im0.p[j];
      ptres = imres->p[j];
      for(i=0; i<dlfX; i++){
	  somdb += ptdb[i];
	  nbp++;
      }
      for(i=0; i<dlfX; i++){
	  somdb += ptdb[i+dlfX];
	  nbp++;
	  ptres[i] = somdb/nbp; 
      } 
      nbp++;
      for(  ; i<nc-dlfX; i++){
	  somdb += ptdb[i+dlfX];
	  ptres[i] = somdb/nbp; 
	  somdb -= ptdb[i-dlfX];
      }
      nbp--;
      for(  ; i<nc; i++){
	  ptres[i] = (pixdb)somdb/nbp; 
	  somdb -= ptdb[i-dlfX];
	  nbp--;  
      }
  }

  /* moyenne colonne */     
  ptdb = des->tmpcoldb;
  for(i=0; i<nc; i++){
      /* recopie */
      for(j=0; j<nr; j++)
	  ptdb[j] = imres->p[j][i];
      nbp = 0;
      somdb = 0.;
      for(j=0; j<dlfY; j++){
	  somdb += ptdb[j];
	  nbp++;
      }
      for(j=0; j<dlfY; j++){
	  somdb += ptdb[j+dlfY];
	  nbp++;
	  imres->p[j][i] = somdb/nbp;  
      } 
      nbp++;
      for(  ; j<nr-dlfY; j++){
	  somdb += ptdb[j+dlfY];
	  imres->p[j][i] = somdb/nbp;  
	  somdb -= ptdb[j-dlfY];
      }
      nbp--;
      for(  ; j<nr; j++){
	  imres->p[j][i] = somdb/nbp;  
	  somdb -= ptdb[j-dlfY];
	  nbp--;  
      }
  }
  return(0);
}

