/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/

/* **************************** lread_ima.c ******************************/
/*                                                                       */
/* ***********************************************************************/

/** \ingroup manip2d
    \defgroup read_ima Lecture de fichier image
 operateur read_ima : lecture d'une image en totalite.
Formats reconnus : .ras (u1 et rvb) | .ima+.dim[+.3lt] | .imw+.dim |.iml+.dim | .imf+.dim | .cxs+.dim | .cxf+.dim

Gestion des Look Up Table (colormap) pour les formats 8 bit :
- pour les .ima : test de la presence de fichier .3lt; si oui, allocation des pointeurs 
lutr, lutv, lutb et chargement (recopie) de cette LUT;
- pour les .ras|RAS : test de la presence de LUT;
===> si LUT de type "niveau de gris" (R=V=B), les vrais niveaux de 
gris des pixels sont restaures au travers de cette LUT qui est ensuite ignoree; 
===> si LUT de type "fausses couleurs" (au moins un niveau ou R V B sont differents), 
allocation des pointeurs lutr, lutv, lutb et chargement (recopie) de cette LUT, completee 
si necessaire jusqu'a n=255 (256 valeurs) avec R=V=B=n;
- si aucune LUT n'est chargee, les pointeurs lutr, lutv, lutb sont mis a NULL.

Descripteur : 
   -char nom[200] : nom du fichier, sans l'extension 
   -char ext[10] : extension 

Type sortie : imau1 | u2 | u4 | fl | cx4 | cx8 | rvb

MODIF 19/12/2007 : ajout lecture des fichiers sauvegardes au format INTEL par test des extensions .IMW, .IML, .IMF, .CXS, .CXF et supression du swap sous LINUX (dans le FLAG INVERSE_DATA)
 
MODIF 20/04/2008 : ajout du format Unsigned int (u4) en lecture : .iml / .IML


@author E. Trouve
@version 1.1 (19/12/07); Include : image.h, proto2D.h
@see mbidon_XX, write_ima, mcpima
*/


#include "image.h"
#include "proto2D.h"
#include "permut.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "rasterfile.h"
#include <string.h>


#define PERMS   0644


/* *************************  LECTURE  *******************************/
/** \ingroup  read_ima
    lecture des parametres de l'operateur read_ima. \\
    nom : nom du fichier, sans suffix \\
    ext : extension (suffix) : ras | ima | imw IMW | imf IMF | cxs CXS | cxf CXF\\
         
    @param des  pointeur descripteur     
    @param ptp  pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author E. Trouve
    @version 1.0 (25/11/99); Include : image.h, proto2D.h
*/
param *read_ima_lect(read_ima_t *des, param *ptp, char *debq){
  char nom[300], question[500];
  int l;

  sprintf(question, "%s nom complet du fichier image a lire", debq); 
  lec_param(question, ptp);
  strcpy(nom, ptp->rep);
  ptp = ptp->next;

  l = strlen(nom);
  do
      l--;
  while(nom[l] != '.'  && l>0);
  if( l>0 ){
      strncpy(des->nom, nom, l);
      des->nom[l] = '\0';
      strcpy(des->ext, nom+l+1);
  }
  else{
      printf("\n>> ERREUR read_ima_lect : nom image incomplet (extension?)\n");
      exit(1);
  }

  return(ptp);
}



/* *************************  INITIALISATION  ***************************/
/** \ingroup read_ima
    XX = u1 | rvb | u2 | fl | cx4 | cx8 : initialisation de l'operateur read_ima.\\
    allocation et lecture de l'image en totalite + LUT eventuelle (cf. description de l'operateur).
    @param des  pointeur descripteur
    @param ima  pointeur image resultat de la lecture
    @author E. Trouve
    @version 1.0 (12/12/99); Include : image.h, proto2D.h
*/
int read_imau1_init(read_ima_t *des, imau1 *ima){ 
  FILE *fpdim, *fp3lt;
  int fima;
  int headras[8], tmpi, erreur, fausse_couleur;
  char nom_ima[200], nom_dim[200], nom_ras[200], nom_3lt[200];  
  register int tmp_int, i, j, nb_color;
  pixu1 tmppix;
  pixu1 cmap[3][256];
  pixu1 *ptpix;

  ima->lutr = NULL;
  ima->lutv = NULL;
  ima->lutb = NULL;

  /*  sprintf(ima->nom, "%s.%s", des->nom, des->ext);  mis a la fin */
  /* format .ima+.dim+[.3lt] */
  
  if(strcmp(des->ext, "ima")== 0){  
      sprintf(nom_ima, "%s.ima", des->nom); 
      sprintf(nom_dim, "%s.dim", des->nom); 
      sprintf(nom_3lt, "%s.3lt", des->nom); 
      /* dimension */
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &ima->nc, &ima->nr );
	  fclose(fpdim);
      }
      else{
	  printf("\n>> ERREUR read_imau1_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }
      /* donnees */
      if( (fima=open(nom_ima, O_RDONLY, 0)) != -1 ){
	  alloc_imau1(ima);  
	  for(j = 0; j < ima->nr; j++)
	      if( read(fima, ima->p[j], ima->nc) != ima->nc ){
		  printf("\n>> ERREUR read_imau1_init: lecture fichier %s\n",
			 nom_ima);
		  exit(1); 
	      }
	  close(fima);  
      }
      else{
	  printf("\n>> ERREUR read_imau1_init : fichier %s introuvable\n", 
		 nom_ima);
	  exit(1);
      }  
      printf("\n>> read_imau1_init : image lue %s, dimX=%d, dimY=%d\n", 
	     nom_ima, ima->nc, ima->nr);
      /* look up table */
      if( fp3lt=fopen(nom_3lt,"r") ){ /* presence de colormap */
	  if( (ima->lutr = (pixu1*)malloc(256)) == NULL ||
	      (ima->lutv = (pixu1*)malloc(256)) == NULL ||
	      (ima->lutb = (pixu1*)malloc(256)) == NULL ){
	      printf ("\nERREUR  read_imau1_init: allocation impossible\n");  
	      exit(1);      
	  }
	  erreur = 0;
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  ima->lutr[j] = (pixu1)tmpi;
	      else
		  erreur = 1;
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  ima->lutv[j] = (pixu1)tmpi;
	      else
		  erreur = 1;
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  ima->lutb[j] = (pixu1)tmpi;
	      else
		  erreur = 1;
	  fclose(fp3lt);
	  if( erreur ){	 
	      printf ("\nATT read_imau1_init: lecture %s incomplete, LUT ignoree\n", nom_3lt);  
	      free(ima->lutr); ima->lutr = NULL;
	      free(ima->lutv); ima->lutv = NULL;
	      free(ima->lutb); ima->lutb = NULL;
	  }
	  else
	      printf ("\nread_imau1_init: LUT %s chargee\n",nom_3lt);  
      }
  }


  /* format raster 8 bit*/
  else if(!strcmp(des->ext, "ras")){ 
    sprintf(nom_ras, "%s.%s", des->nom, des->ext); 
      if( (fima=open(nom_ras, O_RDONLY, 0)) == -1 ){
    	  printf("\n>> ERREUR read_imau1_init : fichier %s introuvable\n", nom_ras);
	  exit(1);
      }
      if( read(fima, (char*)headras, 32) != 32 ){
	  printf("\n>> ERREUR read_imau1_init : lecture fichier %s \n", nom_ras);  
	  exit(1); 
      }

#ifdef INVERSE_DATA
	printf("inversion entete raster (comme des float)\n");
	permuteFloatTab((float *)&headras, 8);
#endif

      
/*       printf("\nFichier RASTER : \n -> headraster[0] : %x\n", headras[0]);    */
      if( headras[0] != 0x59a66a95 ){
	  printf("\n>> ERREUR read_imau1_init: magic number incorrect %x!=0x59a66a95\n", 
                  headras[0]);
          exit(1);
      }
/*       for(j = 1; j < 8; j++) */
/*         printf(" -> headraster[%d] : %d\n", j, headras[j]);       */
      ima->nc = headras[1];
      ima->nr = headras[2];
      alloc_imau1(ima);  
      printf("\n>> read_imau1_init : image %s, dimX=%d, dimY=%d\n",
	     nom_ras, ima->nc, ima->nr);
      if(headras[3]!=8){
	  printf("\n>> ERREUR read_imau1_init: nb bit par pixel %d != 8", headras[3]);
          exit(1);
      }
      nb_color = 0;
      if(headras[6]){  /* presence de colormap */
         nb_color = headras[7]/3;
	 printf("\nATT read_imau1_init : detection d'une LUT de %d couleurs dans %s\n",
		nb_color, nom_ras);
         for(i=0; i<3; i++){ 
            if( read(fima, cmap[i], nb_color) != nb_color ){
	       printf("\n>> ERREUR read_imau1_init: lecture LUT fichier %s \n",
		      nom_ras);  
	       exit(1); 
            }
	    for(j=nb_color; j<256; j++) 
		cmap[i][j] = j;
	 }
	 fausse_couleur = 0;
         for(j=0; j<nb_color; j++)
	     if(cmap[0][j] != cmap[1][j] || cmap[1][j] != cmap[2][j]){
		  fausse_couleur = 1;
		  break;
	     }
	 if( fausse_couleur ){
	     if((ima->lutr = (pixu1*)malloc(256)) == NULL ||
		(ima->lutv = (pixu1*)malloc(256)) == NULL ||
		(ima->lutb = (pixu1*)malloc(256)) == NULL ){
		 printf ("\nERREUR  read_imau1_init: allocation impossible\n");  
		 exit(1);      
	     }
	     for(j = 0; j < 256; j++){
		 ima->lutr[j] = cmap[0][j];
		 ima->lutv[j] = cmap[1][j];
		 ima->lutb[j] = cmap[2][j];
	     }
	 }
	 else
	     printf("\n>> ATT read_imau1_init: LUT de niveau de gris utilisee comme valeur des pixels\n");
      }
      /*lecture image*/
      for(j = 0; j < ima->nr; j++){
         if( read(fima, ima->p[j], ima->nc) != ima->nc ){
	     printf("\n>> ERREUR read_imau1_init : lecture fichier %s \n", nom_ras); 
	     exit(1); 
	 }
	 if(ima->nc % 2)  /* nombre de colonnes impaire => 0 a eliminer */
	     read(fima, &tmppix, 1);
      }
      close(fima);
      /* mapping des pixel par le vrai niveau de gris si colormap de type R==V==B */
      if (headras[6] && (!fausse_couleur)){
	ptpix = ima->p[0];
        tmp_int = ima->nc * ima->nr;
        for(j=0; j< tmp_int; j++, ptpix++)
	    *ptpix = cmap[0][*ptpix];	    
      }     
  }
  else{
      printf("\n>> ERREUR read_imau1_init : extension %s inconnue (.ima/.ras)\n", 
              des->ext);
      exit(1);
  }
  sprintf(ima->nom, "%s.%s", des->nom, des->ext);

  return(0);  
}



/* image 16 bit */
int read_imau2_init(read_ima_t *des, imau2 *im){   /* .imw  et IMW     */
  FILE *fpdim;
  int fima;
  char nom_ima[200], nom_dim[200];
  register int  j, nb_oct;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  /*  strcpy(im->nom, nom_ima); mis a la fin */
  sprintf(nom_dim, "%s.dim", des->nom);   
  if(fpdim=fopen(nom_dim,"r")){
      fscanf(fpdim,"%d %d", &im->nc, &im->nr);
      fclose(fpdim);
  }
  else{
      printf("\n>> ERREUR read_imau2_init : fichier %s introuvable\n", nom_dim);
      exit(1);
  } 
  if((fima=open(nom_ima, O_RDONLY, 0)) != -1) {
      alloc_imau2(im);
      nb_oct = 2*im->nc;
      for(j = 0; j < im->nr; j++)
         if( read(fima, (char *)im->p[j], nb_oct) != nb_oct ){
	     printf("\n>> ERREUR read_imau2_init :lecture fichier %s \n", nom_ima);
	     exit(1);
	 }
      close(fima);
#ifdef INVERSE_DATA
      if(!strcmp(des->ext, "imw")){
	printf("Swap des octets	des donnees lors de la lecture\n");
	for( j=0; j<im->nr; j++)
	permuteShortTab( (short *)im->p[j], im->nc);
      }
      else if(!strcmp(des->ext, "IMW")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IMW (INTEL)\n");
      }
      else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
      } 
#endif

  
  }
  else{
      printf("\n>> ERREUR read_imau2_init : Fichier %s introuvable\n", nom_ima);
      exit(1);
  }
  printf("\n>> read_imau2_init : image %s dimX=%d, dimY=%d\n", 
	 nom_ima, im->nc, im->nr);
  strcpy(im->nom, nom_ima);
  return(0);
}

/***************************************************/
/* image 32 bit */
int read_imau4_init(read_ima_t *des, imau4 *im){   /* .iml  et IML     */
  FILE *fpdim;
  int fima;
  char nom_ima[200], nom_dim[200];
  register int  j, nb_oct;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  /*  strcpy(im->nom, nom_ima); mis a la fin */
  sprintf(nom_dim, "%s.dim", des->nom);   
  if(fpdim=fopen(nom_dim,"r")){
      fscanf(fpdim,"%d %d", &im->nc, &im->nr);
      fclose(fpdim);
  }
  else{
      printf("\n>> ERREUR read_imau4_init : fichier %s introuvable\n", nom_dim);
      exit(1);
  } 
  if((fima=open(nom_ima, O_RDONLY, 0)) != -1) {
      alloc_imau4(im);
      nb_oct = 4*im->nc;
      for(j = 0; j < im->nr; j++)
         if( read(fima, (char *)im->p[j], nb_oct) != nb_oct ){
	     printf("\n>> ERREUR read_imau4_init :lecture fichier %s \n", nom_ima);
	     exit(1);
	 }
      close(fima);
#ifdef INVERSE_DATA
      if(!strcmp(des->ext, "iml")){
	printf("Swap des octets	des donnees lors de la lecture\n");
	for( j=0; j<im->nr; j++)
	permuteIntTab( (unsigned int *)im->p[j], im->nc);
      }
      else if(!strcmp(des->ext, "IML")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IML (INTEL)\n");
      }
      else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
      } 
#endif

  
  }
  else{
      printf("\n>> ERREUR read_imau4_init : Fichier %s introuvable\n", nom_ima);
      exit(1);
  }
  printf("\n>> read_imau4_init : image %s dimX=%d, dimY=%d\n", 
	 nom_ima, im->nc, im->nr);
  strcpy(im->nom, nom_ima);
  return(0);
}

/***************************************************/
/* image float */
int read_imafl_init(read_ima_t *des, imafl *im){   /* .imf et IMF    */
  FILE *fpdim;
  int fima;
  char nom_ima[200], nom_dim[200];
  register int  j, nb_oct;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  /* strcpy(im->nom, nom_ima);  plus loin */ 
  sprintf(nom_dim, "%s.dim", des->nom);
  if(fpdim=fopen(nom_dim,"r")){
      fscanf(fpdim,"%d %d", &im->nc, &im->nr);
      fclose(fpdim);
  }
  else{
      printf("\n>> ERREUR read_imafl_init : fichier %s introuvable\n", nom_dim);
      exit(1);
  } 
  if((fima=open(nom_ima, O_RDONLY, 0)) != -1) {
      alloc_imafl(im);
      nb_oct = 4*im->nc;
      for(j = 0; j < im->nr; j++)
         if( read(fima, (char *)im->p[j], nb_oct) != nb_oct ){
	     printf("\n>> ERREUR read_imafl_init :lecture fichier %s \n", nom_ima);
	     exit(1);
	 }
      close(fima);  
  
#ifdef INVERSE_DATA

	if(!strcmp(des->ext, "imf")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
		permuteFloatTab( im->p[j], im->nc);
	}
	else if(!strcmp(des->ext, "IMF")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IMF (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif
  
  }
  else{
      printf("\n>> ERREUR read_imafl_init : Fichier %s introuvable\n", nom_ima);
      exit(1);
  }
  printf("\n>> read_imafl_init : image %s dimX=%d, dimY=%d\n", 
	 nom_ima, im->nc, im->nr);
  strcpy(im->nom, nom_ima);
  return(0);
}

/* image complexe short */
int read_imacx4_init(read_ima_t *des, imacx4 *im){ /* .cxs  ou CXS     */
  FILE *fpdim;
  int fima;
  char nom_ima[200], nom_dim[200];
  register int  j, nb_oct;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  /* strcpy(im->nom, nom_ima);  plus loin */
  sprintf(nom_dim, "%s.dim", des->nom);
  if(fpdim=fopen(nom_dim,"r")){
      fscanf(fpdim,"%d %d", &im->nc, &im->nr);
      fclose(fpdim);
  }
  else{
      printf("\n>> ERREUR read_imacx4_init : fichier %s introuvable\n", nom_dim);
      exit(1);
  } 
  if((fima=open(nom_ima, O_RDONLY, 0)) != -1) {
      alloc_imacx4(im);
      nb_oct = 4*im->nc;
      for(j = 0; j < im->nr; j++)
         if( read(fima, (char *)im->p[j], nb_oct) != nb_oct ){
	     printf("\n>> ERREUR read_imacx4_init :lecture fichier %s \n", nom_ima);
	     exit(1);
	 }
      close(fima); 

#ifdef INVERSE_DATA
	if(!strcmp(des->ext, "cxs")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
	    permuteShortTab( (short *)&im->p[j][0].re, 2*im->nc);
	}
	else if(!strcmp(des->ext, "CXS")){
	  printf("Pas de swap des octets lors de la lecture car fichier .CXS (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif

 
  }
  else{
      printf("\n>> ERREUR read_imacx4_init : Fichier %s introuvable\n", nom_ima);
      exit(1);
  }
  printf("\n>> read_imacx4_init : image %s dimX=%d, dimY=%d\n", 
	 nom_ima, im->nc, im->nr);
  strcpy(im->nom, nom_ima);
  return(0);
}

/* image complexe float */
int read_imacx8_init(read_ima_t *des, imacx8 *im){ /* .cxf et CXF    */
  FILE *fpdim;
  int fima;
  char nom_ima[200], nom_dim[200];
  register int  j, nb_oct;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  /*   strcpy(im->nom, nom_ima);  plus loin  */
  sprintf(nom_dim, "%s.dim", des->nom);
  if(fpdim=fopen(nom_dim,"r")){
      fscanf(fpdim,"%d %d", &im->nc, &im->nr);
      fclose(fpdim);
  }
  else{
      printf("\n>> ERREUR read_imacx8_init : fichier %s introuvable\n", nom_dim);
      exit(1);
  } 
  if((fima=open(nom_ima, O_RDONLY, 0)) != -1) {
      alloc_imacx8(im);
      nb_oct = 8*im->nc;
      for(j = 0; j < im->nr; j++)
         if( read(fima, (char *)im->p[j], nb_oct) != nb_oct ){
	     printf("\n>> ERREUR read_imacx8_init :lecture fichier %s \n", nom_ima);
	     exit(1);
	 }
      close(fima);
#ifdef INVERSE_DATA

	if(!strcmp(des->ext, "cxf")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
	    permuteFloatTab( &im->p[j][0].re, 2*im->nc);
	}
	else if(!strcmp(des->ext, "CXF")){
	  printf("Pas de swap des octets lors de la lecture car fichier .CXF (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif 
  
  }
  else{
      printf("\n>> ERREUR read_imacx8_init : Fichier %s introuvable\n", nom_ima);
      exit(1);
  }
  printf("\n>> read_imacx8_init : image %s dimX=%d, dimY=%d\n", 
	 nom_ima, im->nc, im->nr);
  strcpy(im->nom, nom_ima);
  return(0);
}

/******************************************/
/* image couleur */
int read_imarvb_init(read_ima_t *des, imarvb *im){ /* .ras       */
  FILE *fichier;
  char nom_ima[200];
  struct rasterfile entete;
  register int l, c;
  unsigned char nb_col_impaire, tampon;

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  if((fichier=fopen(nom_ima,"r")) == NULL) {
      printf("ERREUR: impossible d'ouvrir le fichier %s\n", nom_ima);
      exit(1);
  }

  /*Lecture entete */
  fread(&entete, 1, sizeof(struct rasterfile), fichier);
#ifdef INVERSE_DATA
    printf("inversion entete raster image rvb\n");
    permuteFloatTab((float *)&entete, 8);
#endif

  if(entete.ras_depth != 24){
      printf("ERREUR: image pas au format 24 bits");
      exit(1);
  }
  if(entete.ras_maplength != 0){
      printf("cette image contient une palette, non suporte actuellement\n");
      exit(1);
  }
  im->nc = entete.ras_width;
  im->nr = entete.ras_height;
  nb_col_impaire = im->nc%2;
  
  alloc_imarvb(im);
  /*
  printf("ETAPE LECTURE\n");  
  printf("entete.ras_width=%d\n",entete.ras_width);
  printf("entete.ras_height=%d\n",entete.ras_height);
  printf("entete.ras_depth=%d\n",entete.ras_depth);
  printf("entete.ras_length=%d\n",entete.ras_length);
  printf("entete.ras_type=%d\n",entete.ras_type);
  printf("entete.ras_maptype=%d\n",entete.ras_maptype);
  printf("entete.ras_maplength=%d\n",entete.ras_maplength);
  */
  /* format RVB ou BVR */
  if(entete.ras_type == RT_FORMAT_RGB)
      {
	  for(l=0; l<im->nr; l++) {
	      for(c=0; c<im->nc; c++) {
		  fread( &im->r[l][c], 1, sizeof(pixu1), fichier);
		  fread( &im->v[l][c], 1, sizeof(pixu1), fichier);
		  fread( &im->b[l][c], 1, sizeof(pixu1), fichier);
	      }
	      if(nb_col_impaire) {
		  fread( &tampon, 1, sizeof(pixu1), fichier);
		  fread( &tampon, 1, sizeof(pixu1), fichier);
		  fread( &tampon, 1, sizeof(pixu1), fichier);
	      }
	  }
      }
  else
      {
	  if(entete.ras_type == RT_STANDARD)
	      {
		  for(l=0; l<im->nr; l++) {
		      for(c=0; c<im->nc; c++) {
			  fread( &im->b[l][c], 1, sizeof(pixu1), fichier);
			  fread( &im->v[l][c], 1, sizeof(pixu1), fichier);
			  fread( &im->r[l][c], 1, sizeof(pixu1), fichier);
                        }
		      if(nb_col_impaire) {
			  fread( &tampon, 1, sizeof(pixu1), fichier);
		      }
		  }
	      }
      }
  fclose(fichier);
  strcpy(im->nom, nom_ima);
  return(0);
}
/* *************************  CALCUL (aucun) ***********************/








/* ********************* Op�ateur d'ajout d'une colormap  ***********************/
/** \ingroup manip2d
    \defgroup add_colormap Add a colormap.
\ingroup test
Les luts contenues dans un fichier au format ascii sont charg�s dans le descripteur. Si la lut est incompl�e, l'op�ation est annul�, un message pr�ient l'utilisateur et le programme peut continuer sans probl�e. Dans ce cas les luts du descripteur sont vid�s et ramen�s �un pointeur NULL. Donc au moment de l'op�ation de chargement sur une image (add_colormap_calc), si une lut �ait pr�ente avant, celle-ci est donc perdue. Il en est de meme si le nom de fichier conteant la lut est errone. Dans ce cas le programme se poursuit normalement en ayant remis les pointeurs de lut a NULL.

Descripteur :
   - char nom[200] : nom du fichier complet de la colormap
   - pixu1 *lutr : colormap pour le rouge
   - pixu1 *lutv : colormap pour le vert
   - pixu1 *lutb : colormap pour le bleu

@author L. VALET
@version 1.0 (21/12/99); Include : proto2D.h
@see lread_ima
*/





/* *************************  LECTURE  *******************************/
/** \ingroup add_colormap
    Lecture des parametres de l'op�ateur add_colormap.
    Cette fonction demande �l'utilisateur le nom du fichier contenant la colormap.\\
    @param des  pointeur descripteur     
    @param ptp  pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author L. VALET
    @version 1.0 (21/12/99); Include : proto2D.h
*/
param * add_colormap_lect(add_colormap_t *des, param * ptp, char * debq){
  char question[500];

  sprintf(question, "%s Nom du fichier contenant la colormap", debq);
  lec_param(question, ptp);
  des->nom = ptp->rep;
  ptp = ptp->next;

  return(ptp);
}

/** \ingroup add_colormap
    Initialisation de l'op�ateur add_colormap.\\
    lecture de la colormap et initialisation des champs lutr, lutv, lub du descripteur.
    @param des  pointeur descripteur
    @author L. VALET
    @version 1.0 (21/12/99); Include : proto2D.h
*/
int add_colormap_init(add_colormap_t *des)
{
    FILE *map;
    int tmp, erreur=0, j;

    /*ouverture fichier colormap*/
    if( (map=fopen(des->nom, "r"))!=NULL ){

	/*allocation des luts*/
	if( (des->lutr = (pixu1*)malloc(256)) == NULL ||
	    (des->lutv = (pixu1*)malloc(256)) == NULL ||
	    (des->lutb = (pixu1*)malloc(256)) == NULL ){
	  printf ("add_colormap_init: ERREUR allocation impossible\n");  
	  exit(1); 
	}

	/*lecture de la colormap*/
	for(j = 0; j < 256; j++)
	  if( fscanf(map, "%d", &tmp) == 1 )
	    des->lutr[j] = (pixu1)tmp;
	  else
	    erreur = 1;
	for(j = 0; j < 256; j++)
	  if( fscanf(map, "%d", &tmp) == 1 )
	    des->lutv[j] = (pixu1)tmp;
	  else
	    erreur = 1;
	for(j = 0; j < 256; j++)
	  if( fscanf(map, "%d", &tmp) == 1 )
	    des->lutb[j] = (pixu1)tmp;
	  else
	    erreur = 1;

	if( erreur ){  
	  printf ("add_colormap_init: ATTENTION colormap %s incomplete, LUT ignoree\n", des->nom);  
	  free(des->lutr); des->lutr = NULL;
	  free(des->lutv); des->lutv = NULL;
	  free(des->lutb); des->lutb = NULL;
	}    
	else
	  printf ("add_colormap_init: LUT %s chargee\n", des->nom);

	fclose(map);
    }
    else{
      /* si fichier non ouvert, annule modif colormap */
      printf("add_colormap_init: ATTENTION, impossible d'ouvrir le fichier %s\n", des->nom);
      des->lutr = NULL;
      des->lutv = NULL;
      des->lutb = NULL;
    }

  
}

/** \ingroup add_colormap
    Chargement de la colormap sur une image.
    Si l'image contenait deja une colormap, l'espace m�oire pour celle-ci est lib�� Les champs lutr, lutv, lutb de cette image recoivent ensuite les valeurs des luts contenues dans le descripteur.
    @param des  pointeur descripteur
    @param ima  pointeur sur l'image de type u1
    @author L. VALET
    @version 1.0 (21/12/99); Include : proto2D.h
*/
int add_colormap_calc(add_colormap_t *des, imau1 *ima){

    if( ima->lutr!=NULL && ima->lutv!=NULL && ima->lutb!=NULL ){
	free(ima->lutr);
	free(ima->lutv);
	free(ima->lutb);
    }    	
    ima->lutr = des->lutr;
    ima->lutv = des->lutv;
    ima->lutb = des->lutb;
}



/** \ingroup utility 
    Allocation d'une colormap pour une image u1.\\
    Si l'image contenait deja une colormap, l'espace m�oire pour celle-ci est lib�� Une nouvelle lut complete de 3*256 est allouee. On peut ensuite remplir cette colormap par des valeur ici d'un calcul par exemple.
    @param im  pointeur sur une image de type u1
    @author B. Ionescu
    @version 1.0 (11/01/01); Include : proto2D.h
*/
int alloc_lut_u1(imau1 *im)
{
   if( im->lutr!=NULL && im->lutv!=NULL && im->lutb!=NULL ){
        free(im->lutr);
        free(im->lutv);
        free(im->lutb);
    }    
  
   /*allocation des luts*/
        if( (im->lutr = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL ||
            (im->lutv = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL ||
            (im->lutb = (pixu1*)malloc(256 * sizeof(pixu1))) == NULL )
          {
          printf ("alloc_lut_u1: lutr, lutv, lutb: ERREUR allocation impossible\n");  
          exit(1); 
          }
}



/** \ingroup utility
    Suppression d'une colormap pour une image u1.\\
    La colormap, si elle existe, de l'image de type u1 passe en parametre est supprimee.
    @param im  pointeur sur une image de type u1
    @author B. Ionescu
    @version 1.0 (11/01/01); Include : proto2D.h
*/
int free_lut_u1(imau1 *im)
{
    if(im->lutb!=NULL){
	free(im->lutb); im->lutb = NULL;
    }
    if(im->lutv){
	free(im->lutv); im->lutv = NULL;
    }
    if(im->lutr){
	free(im->lutr); im->lutr = NULL;
    }
  
}



