/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ***************************** lsave_ima.c ******************************/
/*                                                                        */
/* ************************************************************************/

/** \ingroup manip2d
    \defgroup save_ima operateur save_ima : ecriture d'une image bloc par bloc.

Formats de sortie : .ras (8 bit) | .ima | .imw | .iml | .imf | .cxs | .cxf + .dim\\
Visualisation par "xima" des formats .ima | .imw | .imf | .cxs | .cxf + .dim 
et par xv pour le format .ras\\
Pour utiliser "xv" sur les images 8 bits (toto.ima+.dim), passer par le script 
"xvima" sans l'extension : >xvima toto\\
Dans le cas des image de type u1, si une palette existe, 
un fichier d'extension .3lt (3 luts) est cr dans le cas d'une sauvegarde au format ".ima, .dim".
Pour la sauvegarde en rasterfile, la palette est enregistre dans l'entete.\\ 
Optionel (cf. parametre legende) : creation d'un fichier .oigima avec les coordonnee de la zone traitee\\
\\
Descripteur :
\begin{itemize}
   \item FILE *pfd : pointeur fichier dimension ou donnees pour .ras 
   \item int  dfi : numero logique fichier image 
   \item char nom[200] : nom du fichier, sans extension 
   \item char ext[10] : extension  
   \item int  nb_lig_ecr : nombre de lignes ecrites   (calcule par save_cal)
   \item int  nb_col_ecr : nombre de colonnes ecrites (calcule par save_cal)
   \item char  legende[200] : commentaire pour .origima (pas de .origima si vide)
\end{itemize}   
\\
Type entree : imau1 | u2 | u4 | fl | cx4 | cx8 

MODIF 20/04/2008 : ajout du format Unsigned int (u4) en ecriture : .iml 


@name lsave_ima
@author E. Trouve
@version 1.0 (15/03/2000); Include : image.h proto2D.h
@see mselect_XX, lselect_ima, addcolormap
*/

#include "image.h"
#include "proto2D.h"
#include "rasterfile.h"
#include "permut.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#define PERMS   0644

/* ********************   LECTURE DE PARAMETRES   ***************************/
/** \ingroup save_ima
    lecture des parametres de l'operateur save_ima. \\
    nom : nom complet du fichier a ecrire (extension  ras | ima | imw | iml | imf | cxs | cxf )\\ 
    legende : legende pour fichier .origima (non cree si vide)
    @name save_ima_lect
    @param des  pointeur descripteur     
    @param ptp  pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : proto2D.h
*/
param *save_ima_lect(save_ima_t *des, param *ptp, char *debq){
  char nom[300], question[500];
  int l;

  sprintf(question, "%s nom complet du fichier", debq); 
  lec_param(question, ptp);
  strcpy(nom, ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s legende pour fichier .origima (non cree si vide)", debq); 
  lec_param(question, ptp);
  strcpy(des->legende, ptp->rep);
  ptp = ptp->next;

  l = strlen(nom);
  do
      l--;
  while(nom[l] != '.'  && l>0);
  if( l>0 ){
      strncpy(des->nom, nom, l);
      des->nom[l] = '\0';
      strcpy(des->ext, nom+l+1);
  }
  else{
      printf("\n>> ERREUR save_ima_lect : nom image incomplet (extension?)\n");
      exit(1);
  }

  return(ptp);
}

/* ********* INITIALISATION  (8 bit) *******/
/** \ingroup save_ima
    initialisation de l'operateur save_ima pour donnees 8 bits (u1).\\
    ouverture fichiers donnees, \\
     - format .ima : ouverture .dim; ouverture, ecriture, fermeture .3lt si une palette est prsente \\
     - format .ras : ecriture entete (dont LUT si presente)
    @name save_imau1_init
    @param des  pointeur descripteur
    @param im   pointeur structure imau1
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : image.h, proto2D.h
*/


/************************************/
/* type u1 */
/************************************/
int save_imau1_init(save_ima_t *des, imau1 im){ 
  FILE *fp;
  char nom_ima[200], nom_dim[200], nom_lut[200];
  int j;
  struct rasterfile entete;
  unsigned char colormap[768], col_impair;

  if( strcmp(des->ext, "ima")== 0 ){
      sprintf(nom_ima, "%s.ima", des->nom);
      sprintf(nom_dim, "%s.dim", des->nom);
      /* ouverture du .ima */
      if( (des->dfi=creat(nom_ima, PERMS)) == -1 ){
	  printf("\n>> ERREUR save_imau1_init: creation fichier %s impossible\n",nom_ima);
	  exit(1);
      }
      /* ouverture du .dim */
      if( (des->pfd=fopen(nom_dim,"w")) == NULL ){
	  printf("\n>> ERREUR save_imau1_init: creation fichier %s impossible\n",nom_dim);
	  exit(1);
      }
      /* ecriture du .3lt */
      if( im.lutr!=NULL && im.lutv!=NULL && im.lutb!=NULL) {
	  sprintf(nom_lut, "%s.%s", des->nom, "3lt"); 
	  printf("\n>> save_imau1_init: sauvegarde palette dans %s\n", nom_lut);
	  if( (fp=fopen(nom_lut,"w")) == NULL ){ 
	      printf("\n>> ERREUR save_imau1_init: creation fichier %s impossible\n",nom_lut); 
	      exit(1); 
	  } 
	  fprintf(fp,"%3d ",im.lutr[0]);
	  for(j = 1; j < 256; j++) {
	      if((j % 20) == 0 ) 
		  fprintf(fp,"\n");
	      fprintf(fp,"%3d ", im.lutr[j]);
	  } 
	  fprintf(fp,"\n"); 

	  fprintf(fp,"%3d ",im.lutv[0]);
	  for(j = 1; j < 256; j++) {
	      if((j % 20) == 0 ) 
		  fprintf(fp,"\n");
	      fprintf(fp,"%3d ", im.lutv[j]);
	  } 
	  fprintf(fp,"\n"); 

	  fprintf(fp,"%3d ",im.lutb[0]);
	  for(j = 1; j < 256; j++) {
	      if((j % 20) == 0 ) 
		  fprintf(fp,"\n");
	      fprintf(fp,"%3d ", im.lutb[j]); 
	  }
	  fprintf(fp,"\n"); 
	  fclose(fp);
      }
  }
  else if(strcmp(des->ext, "ras")== 0){
      sprintf(nom_ima, "%s.ras", des->nom);
      /* ouverture du .ras */
      if( (des->pfd=fopen(nom_ima,"w")) == NULL ){
	  printf("\n>> ERREUR save_imau1_init: creation fichier %s impossible\n",nom_dim);
	  exit(1);
      }
      /*creation de l'entete*/
      col_impair=im.nc%2;
      entete.ras_magic=RAS_MAGIC;
      entete.ras_width=im.nc;
      entete.ras_height=im.nr;
      entete.ras_depth=8;	
      entete.ras_length=im.nr*(im.nc+col_impair);
      entete.ras_type=1;
      entete.ras_maptype=1;
      entete.ras_maplength=768;

      /* ecriture de l'entete*/
      fwrite( (char *)&entete, sizeof(struct rasterfile), 1, des->pfd);
      /*ecriture de la palette*/	
      if( im.lutr!=NULL && im.lutv!=NULL && im.lutb!=NULL) {
	  fwrite( im.lutr, sizeof(unsigned char), 256, des->pfd); 
	  fwrite( im.lutv, sizeof(unsigned char), 256, des->pfd); 
	  fwrite( im.lutb, sizeof(unsigned char), 256, des->pfd); 
      } 
      else {
	  for (j=0;j<768;j++) colormap[j]=j; 	
	  fwrite( colormap, sizeof(unsigned char), 768, des->pfd); 
      }
  } 
  else{
      printf("\n>> ERREUR save_imau1_init: extension fichier %s inconnues\n", des->ext);
      exit(1);  
  }
  des->nb_lig_ecr = 0;
  return(0);  
}

/* ********* INITIALISATION (autre que 8 bit) *******/
/** \ingroup save_ima
    initialisation de l'operateur save_ima pour donnees autres ques 8 bit (u2, fl, cx4, cx8).\\
    ouverture fichiers donnees + .dim 
    @name save_ima_init
    @param des  pointeur descripteur
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : proto2D.h
*/
/************************************/
/* type autre que u1 */
/************************************/
int save_ima_init(save_ima_t *des){ 
  char nom_ima[200], nom_dim[200];

  sprintf(nom_ima, "%s.%s", des->nom, des->ext);
  sprintf(nom_dim, "%s.dim", des->nom);

  if( strcmp(des->ext, "imw")== 0 ||
      strcmp(des->ext, "iml")== 0 ||
      strcmp(des->ext, "imf")== 0 ||
      strcmp(des->ext, "cxs")== 0 ||
      strcmp(des->ext, "cxf")== 0 ){
      /* ouverture du .imw|... */
      if( (des->dfi=creat(nom_ima, PERMS)) == -1 ){
	  printf("\n>> ERREUR save_ima_init: creation fichier %s impossible\n",nom_ima);
	  exit(1);
      }
      /* ouverture du .dim */
      if( (des->pfd=fopen(nom_dim,"w")) == NULL ){
	  printf("\n>> ERREUR save_ima_init: creation fichier %s impossible\n",nom_dim);
	  exit(1);
      }
  }
  else{
      printf("\n>> ERREUR save_ima_init: extension fichier %s inconnue\n", des->ext);
      exit(1);  
  }
  des->nb_lig_ecr = 0;
  return(0);  
}


/* ********************************  CALCUL  **********************************/
/** \ingroup save_ima
    XX = u1 | u2 | u4 | fl | cx4 | cx8 : partie "calcul" de l'operateur save_ima .\\
    ecriture d'un bloc d'image
    @name save_imaXX_calc
    @param des  pointeur descripteur
    @param im  bloc image a ecrire
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : image.h, proto2D.h
*/

/************************************/
/* image 8 bit */
/************************************/
int save_imau1_calc(save_ima_t *des, imau1 im){   /* .ima | .dim  */
  int j, nb_oct, col_impair;
  pixu1 tmp;
      
  nb_oct = sizeof(pixu1)*im.nc;
  if(strcmp(des->ext, "ima") == 0){
      for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, im.p[j], nb_oct) != nb_oct ){ 
	  printf("\n>> ERREUR save_imau1_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  }
  else if(strcmp(des->ext, "ras") == 0 ){
      tmp = 0;
      col_impair = im.nc%2;
      for(j = 0; j < im.nr; j++)				
	  fwrite(im.p[j], sizeof(pixu1), nb_oct, des->pfd); 
      if(col_impair)
	  fwrite(&tmp, sizeof(pixu1), 1, des->pfd);
      }  
  else{
      printf("\n>> ERREUR save_imau1_calc : extension fichier %s inconnue\n", des->ext);
      exit(1);  
  }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;
  return(0);
}

/************************************/
/* image 16 bit */
/************************************/
int save_imau2_calc(save_ima_t *des, imau2 im){   /* .imw       */
  int j, nb_oct;
      
  nb_oct = sizeof(pixu2)*im.nc;

#ifdef INVERSE_DATA
  printf("Swap des octets	des donnees pour ecriture sur fichier\n");
  for( j=0; j<im.nr; j++)
    permuteShortTab( (short *)im.p[j], im.nc);
#endif


  for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, (char *)im.p[j], nb_oct) != nb_oct){ 
	  printf("\n>> ERREUR save_imau2_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;

#ifdef INVERSE_DATA
  //printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteShortTab( (short *)im.p[j], im.nc);
#endif

  return(0);
}

/************************************/
/* image 32 bit */
/************************************/
int save_imau4_calc(save_ima_t *des, imau4 im){   /* .iml       */
  int j, nb_oct;
      
  nb_oct = sizeof(pixu4)*im.nc;

#ifdef INVERSE_DATA
  printf("Swap des octets des donnees pour ecriture sur fichier\n");
  for( j=0; j<im.nr; j++)
    permuteIntTab( (unsigned int *)im.p[j], im.nc);
#endif


  for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, (char *)im.p[j], nb_oct) != nb_oct){ 
	  printf("\n>> ERREUR save_imau4_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;

#ifdef INVERSE_DATA
  //printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteIntTab( (unsigned int *)im.p[j], im.nc);
#endif

  return(0);
}

/************************************/
/* image float */
/************************************/
int save_imafl_calc(save_ima_t *des, imafl im){   /* .imf       */
  int j, nb_oct;
      
  nb_oct = sizeof(pixfl)*im.nc;
#ifdef INVERSE_DATA
  printf("Swap des octets	des donnees pour ecriture sur fichier\n");
  for( j=0; j<im.nr; j++)
    permuteFloatTab( im.p[j], im.nc);
#endif

  for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, (char *)im.p[j], nb_oct) != nb_oct){ 
	  printf("\n>> ERREUR save_imafl_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;

#ifdef INVERSE_DATA
	//printf("re Swap des octets des donnees float pour traitement ultrieur\n");
	for( j=0; j<im.nr; j++)
		permuteFloatTab( im.p[j], im.nc);
#endif

  return(0);
}

/************************************/
/* image complex short */
/************************************/
int save_imacx4_calc(save_ima_t *des, imacx4 im){ /* .cxs       */
  int j, nb_oct;
      
  nb_oct = sizeof(pixcx4)*im.nc;
#ifdef INVERSE_DATA
	printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteShortTab( (short *)&im.p[j][0].re, 2*im.nc);
#endif
  for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, (char *)im.p[j], nb_oct) != nb_oct){ 
	  printf("\n>> ERREUR save_imacx4_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;

#ifdef INVERSE_DATA
  //printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteShortTab( (short *)&im.p[j][0].re, 2*im.nc);
#endif
  return(0);
}

/************************************/
/* image complex float */
/************************************/
int save_imacx8_calc(save_ima_t *des, imacx8 im){ /* .cxf       */
  int j, nb_oct;
      
  nb_oct = sizeof(pixcx8)*im.nc;

#ifdef INVERSE_DATA
	printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteFloatTab( (float *)im.p[j], 2*im.nc);
#endif

  for(j = 0; j < im.nr; j++)				
      if( write(des->dfi, (char *)im.p[j], nb_oct) != nb_oct){ 
	  printf("\n>> ERREUR save_imacx8_calc : ecriture du fichier %s.%s\n",
		 des->nom, des->ext); 
	  exit(1); 
      }
  if(des->nb_lig_ecr == 0)
      des->nb_col_ecr = im.nc;
  des->nb_lig_ecr += im.nr;

#ifdef INVERSE_DATA
    //printf("Swap des octets	des donnees pour ecriture sur fichier\n");
	for( j=0; j<im.nr; j++)
		permuteFloatTab( (float *)im.p[j], 2*im.nc);
#endif
  return(0);
}


/* *********************** FERMETURE : ecriture .dim, .origima  *******************/
/** \ingroup save_ima
    fermeture de l'operateur save_ima .\\
    Fermeture fichier donnees (avec reecriture de l'entete si raster),\\
    Ecriture et fermeture .dim (sauf cas raster)\\
    Overture, ecriture et fermeture .origima (sauf si legende vide).  
    @name save_ima_ferm
    @param sel  pointeur descripteur operateur select_ima
    @param des  pointeur descripteur operateur save_ima
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : proto2D.h
*/
int save_ima_ferm(select_ima_t sel, save_ima_t des){
  FILE *fp;
  char nom[200];
  int tmp, col_impair;

  if( strcmp(des.ext, "ima")== 0 ||
      strcmp(des.ext, "imw")== 0 ||
      strcmp(des.ext, "iml")== 0 ||
      strcmp(des.ext, "imf")== 0 ||
      strcmp(des.ext, "cxs")== 0 ||
      strcmp(des.ext, "cxf")== 0 ){
      /* ferme donnees */ 
      close(des.dfi);
      printf(">> save_ima_ferm : ecriture %s.%s terminee\n", des.nom, des.ext);
      /* ecrit et ferme .dim */
      fprintf(des.pfd, "%d %d", des.nb_col_ecr, des.nb_lig_ecr);
      fclose(des.pfd);
  }
  /* cas RASTER */   
  else if( strcmp(des.ext, "ras") == 0){
      col_impair = des.nb_col_ecr%2;
      tmp = des.nb_lig_ecr*(des.nb_col_ecr+col_impair);
      fseek(des.pfd, 2*sizeof(int), SEEK_SET);
      fwrite(&des.nb_lig_ecr, sizeof(int), 1, des.pfd);
      printf("nb ligne ecrite:%d",des.nb_lig_ecr);
      fseek(des.pfd, 4*sizeof(int), SEEK_SET);
      fwrite(&tmp, sizeof(int), 1, des.pfd);
      }
  else{
      printf("\n>> ERREUR save_ima_ferm: extension fichier %s inconnue\n", des.ext);
      exit(1);  
  }

  /* .origima */
  if( des.legende[0] != '\0' ){
      sprintf(nom, "%s.origima", des.nom);
      if( (fp=fopen(nom,"w")) == NULL ){
	  printf("\n>> ERREUR save_ima_ferm: creation fichier %s impossible\n",nom);
	  exit(1);
      }  
      fprintf(fp,"Fichier origine : %s.%s\n", sel.nom, sel.ext);
      /*      fprintf(fp,"\tdecimation ligne 1  colonne 1\n");  */
      fprintf(fp,"\tpoint haut gauche : np = %d, nl = %d\n", sel.col0, sel.lig0);
      fprintf(fp,"\tpoint bas droite  : np = %d, nl = %d\n", 
	      sel.col0+sel.nb_col-1, sel.lig0+des.nb_lig_ecr-1);
      fprintf(fp,"Legende : %s\n", des.legende);
      fclose(fp);
  }
  return(0);
}


