/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ***************************** lselect_ima.c ****************************/
/*                                                                        */
/* ************************************************************************/

/** \ingroup manip2d
    \defgroup select_ima operateur select_ima : lecture d'une image bloc par bloc

Formats reconnus : .ras (8 bit) | .ima+.dim[+.3lt] | .imw+.dim | .iml+.dim | .imf+.dim | .cxs+.dim | .cxf+.dim\\
\\
Gestion des Look Up Table (colormap) pour les formats 8 bit :\\
1) pour les .ima : test de la presence de fichier .3lt; si oui, allocation des pointeurs 
lutr, lutv, lutb et chargement (recopie) de cette LUT;\\
2) pour les .ras : test de la presence de LUT; si oui, allocation des pointeurs 
lutr, lutv, lutb et chargement (recopie) de cette LUT, completee 
si necessaire jusqu'a n=255 (256 valeurs) avec R=V=B=n;\\
test si LUT de type "NG (niveau de gris, R=V=B)" ou bien de type "FC (fausse couleur)".\\
A la lecture des blocs si LUT de type NG, les vrais niveaux de gris des pixels sont restaures.\\
3) si aucune LUT n'est chargee, les pointeurs lutr, lutv, lutb sont mis a NULL.\\
\\
Descripteur : \\
   char nom[200] : nom du fichier, sans l'extension \\
   char ext[10] : extension \\
   int dimX, dimY :  dimension initiales de l'image \\
   int col0, lig0 : coin superieur gauche premier bloc \\
   int nb_col, nb_lig : nombre de colonne, nombre de ligne d'un bloc \\
   int nb_bloc : nombre de bloc a lire \\
   int dfi : nomero logique fichier image \\
   int next_lig : indice premiere ligne bloc suivant \\
   int nb_bloc_lu : nombre de bloc lu \\
   int header : nombre d'octets de l'entete \\
   int lut_ng : presence de LUT de type "niveau de gris" \\ 

Type sortie : imau1 | u2 | u4 | fl | cx4 | cx8 


MODIF 20/04/2008 :\\ 
- ajout du format Unsigned int (u4) en lecture : .iml / .IML\\
- ajout lecture des fichiers sauvegardes au format INTEL par test des extensions .IMW, .IML, .IMF, .CXS, .CXF et supression du swap sous LINUX (dans le FLAG INVERSE_DATA)\\ 

@name lselect_ima
@author E. Trouve
@version 1.0 (15/03/2000); Include : image.h, proto2D.h
@see mselect_XX, lsave_ima
*/

#include "image.h"
#include "proto2D.h"
#include "permut.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "rasterfile.h"
#include <string.h>

#define PERMS   0644


/* **********************  LECTURE DE PARAMETRES   ***************************/
/** \ingroup select_ima
    lecture des parametres de l'operateur select_ima. \\
    nom : nom du fichier avec son extension : ras | ima | imw | imf | cxs | cxf\\
    col0 : premiere colonne a extraire\\
    nb_col : nombre de colonnes des blocs\\
    lig0 : premiere ligne a extraire\\
    nb_lig : nombre de lignes des blocs\\
    nb_bloc : nombre de blocs a lire\\
    @name select_ima_lect
    @param des  pointeur descripteur     
    @param ptp  pointeur sur parametre courant deja alloue (a lire),     
    @param debq pointeur de chaine de caracteres placee en debut de question
    @return  renvoie nouveau pointeur de parametre deja alloue (pour la suite) 
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : image.h, proto2D.h
*/
param *select_ima_lect(select_ima_t *des, param *ptp, char *debq){
  FILE *fpdim;
  char nom[200], nom_dim[200], question[500];
  int l, c2, l2, nb;

  sprintf(question, "%s nom complet du fichier", debq); 
  lec_param(question, ptp);
  strcpy(nom, ptp->rep);
  ptp = ptp->next;

  l = strlen(nom);
  do
      l--;
  while(nom[l] != '.'  && l>0);
  if( l>0 ){
      strncpy(des->nom, nom, l);
      des->nom[l] = '\0';
      strcpy(des->ext, nom+l+1);
  }
  else{
      printf("\n>> ERREUR select_ima_lect : nom image incomplet (extension?)\n");
      exit(1);
  }

  /* dimensions des blocs de lignes */
  sprintf(question, "%s premiere colonne a extraire", debq); 
  lec_param(question, ptp);
  des->col0 = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s nombre de colonnes des blocs", debq); 
  lec_param(question, ptp);
  des->nb_col = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s premiere ligne a extraire", debq); 
  lec_param(question, ptp);
  des->lig0 = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s nombre de lignes des blocs", debq); 
  lec_param(question, ptp);
  des->nb_lig = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, "%s nombre de blocs a lire", debq); 
  lec_param(question, ptp);
  des->nb_bloc = atoi(ptp->rep);
  ptp = ptp->next;

  return(ptp);
}


/* **********************  INITIALISATION  *****************************/
/** \ingroup select_ima
XX = u1 | u2 | u4 | fl | cx4 | cx8 : initialisation de l'operateur select_ima.\\
    ouverture fichier, allocation du bloc, lecture dimension, lecture LUT eventuelle (cf. description de l'operateur).
    @name select_imaXX_init
    @param des  pointeur descripteur
    @param im  pointeur image resultat de la lecture
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : image.h, proto2D.h
*/

/************************************/
/* image 8 bit : .ima+.dim et .ras */
/************************************/
int select_imau1_init(select_ima_t *des, imau1 *im){ 
  FILE *fpdim, *fp3lt;
  char nom_ima[200], nom_dim[200], nom_3lt[200], nom_ras[200];  
  int c2, l2, tmpi, erreur, j;
  
  int headras[8], fausse_couleur, fima;
  register int nb_color, i;
  pixu1 cmap[3][256];

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imau1(im);  

  /* cas .ima + .dim + [.3lt] */
  if(strcmp(des->ext, "ima") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.ima", des->nom); 
      sprintf(nom_dim, "%s.dim", des->nom); 
      sprintf(nom_3lt, "%s.3lt", des->nom); 

      /* dimension */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imau1_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imau1_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imau1_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imau1_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  

      /* lecture de .3lt */
      if( fp3lt=fopen(nom_3lt,"r") ){ /* presence de colormap */
	  if( (im->lutr = (pixu1*)malloc(256)) == NULL ||
	      (im->lutv = (pixu1*)malloc(256)) == NULL ||
	      (im->lutb = (pixu1*)malloc(256)) == NULL ){
	      printf ("\nERREUR select_imau1_init: allocation impossible\n");  
	      exit(1);      
	  }
	  erreur = 0;
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  im->lutr[j] = (pixu1)tmpi;
	      else{
		  erreur = 1;
		  break;
	      }
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  im->lutv[j] = (pixu1)tmpi;
	      else{
		  erreur = 1;
		  break;
	      }
	  for(j = 0; j < 256; j++)
	      if( fscanf(fp3lt, "%d", &tmpi) == 1 )
		  im->lutb[j] = (pixu1)tmpi;
	      else{
		  erreur = 1;
		  break;
	      }
	  fclose(fp3lt);
	  if( erreur ){	 
	      printf ("\nATT select_imau1_init: lecture %s incomplete, LUT ignoree\n", nom_3lt);  
	      free(im->lutr); im->lutr = NULL;
	      free(im->lutv); im->lutv = NULL;
	      free(im->lutb); im->lutb = NULL;
	  }
	  else
	      printf ("\nselect_imau1_init: LUT %s chargee\n",nom_3lt);  
      }
      else{
	  im->lutr = NULL;
	  im->lutv = NULL;
	  im->lutb = NULL;
      }
  }
  /* cas RASTER */
  else if(strcmp(des->ext, "ras")==0){
      /*ouverture du fichier*/
      sprintf(nom_ras, "%s.ras", des->nom); 
      if( (des->dfi = open(nom_ras, O_RDONLY, 0)) == -1 ){
    	  printf("\n>> ERREUR select_imau1_init : fichier %s introuvable\n", nom_ras);
	  exit(1);
      }
      /*lecture entete*/
      if( read(des->dfi, (char*)headras, 32) != 32 ){
	  printf("\n>> ERREUR select_imau1_init : lecture fichier %s \n", nom_ras);  
	  exit(1); 
      } 

      /*conservation de la dimension de l'entete*/
      des->header = 32;

      /*verification type raster*/
      if( headras[0] != 0x59a66a95 ){
	  printf("\n>> ERREUR read_imau1_init: magic number incorrect %x!=0x59a66a95\n", 
                  headras[0]);
          exit(1);
      }

      /*verif coherence des tailles*/
      des->dimX = headras[1];
      des->dimY = headras[2];
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imau1_init : dimensions incoherentes\n");
	  exit(1); 
      }
      /*test presence colormap*/
      nb_color = 0;
      des->lut_ng = 0;
      if(headras[6]){
	  des->lut_ng = 1;
	  des->header = des->header + headras[7];
	  nb_color = headras[7]/3;
	  printf("\nATT read_imau1_init : detection d'une LUT de %d niveaux dans %s\n",
		 nb_color, nom_ras);
	  /*lecture de la lut*/
	  for(i=0; i<3; i++){ 
	      if( read(des->dfi, cmap[i], nb_color) != nb_color ){
		  printf("\n>> ERREUR read_imau1_init: lecture LUT fichier %s \n", nom_ras);  
		  exit(1); 
	      }
	      for(j=nb_color; j<256; j++) 
		  cmap[i][j] = j;
	  }
	  /*test si cas fausses couleurs*/
	  for(j=0; j<nb_color; j++)
	      if(cmap[0][j] != cmap[1][j] || cmap[1][j] != cmap[2][j]){
		  des->lut_ng = 0;
		  break;
	      }	 
	  if((im->lutr = (pixu1*)malloc(256)) == NULL ||
	     (im->lutv = (pixu1*)malloc(256)) == NULL ||
	     (im->lutb = (pixu1*)malloc(256)) == NULL ){
	      printf ("\nERREUR  read_imau1_init: allocation impossible\n");  
	      exit(1);      
	  }
	  for(j = 0; j < 256; j++){
	      im->lutr[j] = cmap[0][j];
	      im->lutv[j] = cmap[1][j];
	      im->lutb[j] = cmap[2][j];
	  }
	  printf("\n>> ATT read_imau1_init: LUT chargee\n");
      }
      else{
	  im->lutr = NULL;
	  im->lutv = NULL;
	  im->lutb = NULL;
      }
      /*fin du cas rasterfile*/
  }
  else{
      printf("\n>> ERREUR select_imau1_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;
  return(0);  
}

/************************************************/
/* image 16 bit */
/************************************************/
int select_imau2_init(select_ima_t *des, imau2 *im){   /* .imw       */
  FILE *fpdim;
  char nom_ima[200], nom_dim[200];  
  int c2, l2, tmpi, erreur, j;

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imau2(im);  

  /* cas .imw|.IMW + .dim  */
  if(strcmp(des->ext, "imw") == 0 || strcmp(des->ext, "IMW") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.%s", des->nom, des->ext); 
      sprintf(nom_dim, "%s.dim", des->nom); 

      /* dimesion */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imau2_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imau2_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imau2_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imau2_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  
  }
  else{
      printf("\n>> ERREUR select_imau2_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;

  return(0);  
}

/************************************************/
/* image 32 bit */
/************************************************/
int select_imau4_init(select_ima_t *des, imau4 *im){   /* .iml       */
  FILE *fpdim;
  char nom_ima[200], nom_dim[200];  
  int c2, l2, tmpi, erreur, j;

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imau4(im);  

  /* cas .iml|IML + .dim  */
  if(strcmp(des->ext, "iml") == 0 || strcmp(des->ext, "IML") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.%s", des->nom, des->ext); 
      sprintf(nom_dim, "%s.dim", des->nom); 

      /* dimesion */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imau4_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imau4_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imau4_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imau4_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  
  }
  else{
      printf("\n>> ERREUR select_imau4_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;

  return(0);  
}

/************************************************/
/* image float */
/************************************************/
int select_imafl_init(select_ima_t *des, imafl *im){   /* .imf       */
  FILE *fpdim;
  char nom_ima[200], nom_dim[200];  
  int c2, l2, tmpi, erreur, j;

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imafl(im);  

  /* cas .imf|IMF + .dim  */
  if(strcmp(des->ext, "imf") == 0 || strcmp(des->ext, "IMF") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.%s", des->nom, des->ext);  
      sprintf(nom_dim, "%s.dim", des->nom); 

      /* dimesion */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imafl_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imafl_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imafl_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imafl_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  
  }
  else{
      printf("\n>> ERREUR select_imafl_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;

  return(0);  
}

/************************************************/
/* image complexe short */
/************************************************/
int select_imacx4_init(select_ima_t *des, imacx4 *im){ /* .cxs       */
  FILE *fpdim;
  char nom_ima[200], nom_dim[200];  
  int c2, l2, tmpi, erreur, j;

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imacx4(im);  

  /* cas .cxs|CXS + .dim  */
  if(strcmp(des->ext, "cxs") == 0 || strcmp(des->ext, "CXS") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.%s", des->nom, des->ext);  
      sprintf(nom_dim, "%s.dim", des->nom); 

      /* dimesion */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imacx4_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imacx4_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imacx4_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imacx4_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  
  }
  else{
      printf("\n>> ERREUR select_imacx4_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;

  return(0);  
}

/************************************************/
/* image complexe float */
/************************************************/
int select_imacx8_init(select_ima_t *des, imacx8 *im){ /* .cxf       */
  FILE *fpdim;
  char nom_ima[200], nom_dim[200];  
  int c2, l2, tmpi, erreur, j;

  /* allocation du bloc */
  sprintf(im->nom, "%s_bloc.%s", des->nom, des->ext);
  im->nc = des->nb_col;
  im->nr = des->nb_lig;
  alloc_imacx8(im);  

  /* cas .cxf|CXF + .dim  */
  if(strcmp(des->ext, "cxf") == 0 || strcmp(des->ext, "CXF") == 0){  
      des->header = 0;
      sprintf(nom_ima, "%s.%s", des->nom, des->ext); 
      sprintf(nom_dim, "%s.dim", des->nom); 

      /* dimesion */ 
      if( fpdim=fopen(nom_dim,"r") ){
	  fscanf(fpdim,"%d %d", &des->dimX, &des->dimY );
	  fclose(fpdim);
	  printf("\n>> select_imacx8_init : image %s, dimX=%d, dimY=%d\n", 
		 nom_ima, des->dimX, des->dimY);
      }
      else{
	  printf("\n>> ERREUR select_imacx8_init : fichier %s introuvable\n",
		 nom_dim);
	  exit(1);
      }      

      /* verif coherence des tailles de bloc */
      c2 = des->col0 + des->nb_col;
      l2 = des->lig0 + des->nb_lig*des->nb_bloc;
      if( des->col0 < 0  || des->col0 >= des->dimX ||
	  c2 < des->col0 || c2 >  des->dimX        || 
	  des->lig0 < 0  || des->lig0 >= des->dimY ||
	  l2 < des->lig0 || l2 > des->dimY         ){ 
	  printf("\n>> ERREUR select_imacx8_init : dimensions incoherentes\n");
	  exit(1); 
      }

      /* ouverture donnees  */
      if( (des->dfi=open(nom_ima, O_RDONLY, 0)) == -1 ){
	  printf("\n>> ERREUR select_imacx8_init : fichier %s introuvable\n", nom_ima);
	  exit(1);
      }  
  }
  else{
      printf("\n>> ERREUR select_imacx8_init : extension %s inconnue\n", des->ext);
      exit(1);
  }

  des->next_lig = des->lig0;  
  des->nb_bloc_lu = 0;

  return(0);  
}


/* ************************  CALCUL  ****************************************/
/** \ingroup select_ima
XX = u1 | u2 | u4 | fl | cx4 | cx8 : partie "calcul" de l'operateur select_ima.\\
    lecture d'un nouveau bloc de l'image; Conversion via la LUT si raster 8 bit en niveau de gris 
    (cf. description de l'operateur); Fermeture du fichier en fin de lecture. 
    @name select_imaXX_calc
    @param des  pointeur descripteur
    @param im  pointeur image resultat de la lecture
    @return 0 si lecture 0K, 1 si nb bloc deja lu >= nb bloc a lire, 2 si probleme (fin de fichier...)  
    @author E. Trouve
    @version 1.0 (15/03/2000); Include : image.h, proto2D.h
*/

/************************************************/
/* .ima   ou .ras    */
/************************************************/
int select_imau1_calc(select_ima_t *des, imau1 *im){

  int i, j, nb_oct, sp, col_impair;

  sp = sizeof(pixu1);
  col_impair = des->dimX%2;
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imau1_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  /* cas .ima + .dim  */
  if(strcmp(des->ext, "ima") == 0){
      for(j=0; j<im->nr; j++){
	  lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
	  if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	      printf("\n>> ATT select_imau1_calc: lecture %s incomplete\n", im->nom);
	      close(des->dfi);
	      return(2); 
	  }
      }
  }
  /* cas RASTER */
  else if(strcmp(des->ext, "ras")==0){
      for(j=0; j<im->nr; j++){
	  lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*(des->dimX+col_impair) + des->col0)*sp), 0);
	  if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	      printf("\n>> ATT select_imau1_calc: lecture %s incomplete\n", im->nom);
	      close(des->dfi);
	      return(2); 
	  }
	  /*conversion des donnees par la lut si necessaire*/
	  if(des->lut_ng && im->lutr!=NULL)
	      for(i=0; i<im->nc; i++)
		  im->p[j][i] = im->lutr[im->p[j][i]];
		  
      }
  }
  else{
      printf("\n>> ERREUR select_imau1_calc : extension %s inconnue\n", des->ext);
      exit(1);
  }
  printf("\n>> select_imau1_calc: lecture %s : bloc %d, lignes [%d %d] OK\n",
	 im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);

  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}


/************************************************/
/* .imw  | IMW     */
/************************************************/
int select_imau2_calc(select_ima_t *des, imau2 *im){

  int j, nb_oct, sp;

  sp = sizeof(pixu2);
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imau2_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  printf("\n>> select_imau2_calc: lecture %s : bloc %d, lignes [%d %d]\n",
      im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);
  for(j=0; j<im->nr; j++){
      lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
      if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	    printf("\n>> ATT select_imau2_calc: lecture %s incomplete\n", im->nom);
	    close(des->dfi);
	    return(2); 
      }
  }

#ifdef INVERSE_DATA
      if(!strcmp(des->ext, "imw")){
	printf("Swap des octets	des donnees lors de la lecture\n");
	for( j=0; j<im->nr; j++)
	permuteShortTab( (short *)im->p[j], im->nc);
      }
      else if(!strcmp(des->ext, "IMW")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IMW (INTEL)\n");
      }
      else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
      } 
#endif

  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}

/************************************************/
/* .iml | IML      */
/************************************************/
int select_imau4_calc(select_ima_t *des, imau4 *im){

  int j, nb_oct, sp;

  sp = sizeof(pixu4);
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imau4_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  printf("\n>> select_imau4_calc: lecture %s : bloc %d, lignes [%d %d]\n",
      im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);
  for(j=0; j<im->nr; j++){
      lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
      if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	    printf("\n>> ATT select_imau4_calc: lecture %s incomplete\n", im->nom);
	    close(des->dfi);
	    return(2); 
      }
  }

#ifdef INVERSE_DATA
      if(!strcmp(des->ext, "iml")){
	printf("Swap des octets	des donnees lors de la lecture\n");
	for( j=0; j<im->nr; j++)
	permuteIntTab( (unsigned int *)im->p[j], im->nc);
      }
      else if(!strcmp(des->ext, "IML")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IML (INTEL)\n");
      }
      else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
      } 
#endif

  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}

/************************************************/
/* .imf | IMF       */
/************************************************/
int select_imafl_calc(select_ima_t *des, imafl *im){

  int j, nb_oct, sp;

  sp = sizeof(pixfl);
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imafl_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  printf("\n>> select_imafl_calc: lecture %s : bloc %d, lignes [%d %d]\n",
	  im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);
  for(j=0; j<im->nr; j++){
      lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
      if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	    printf("\n>> ATT select_imafl_calc: lecture %s incomplete\n", im->nom);
	    close(des->dfi);
	    return(2); 
      }
  }

#ifdef INVERSE_DATA

	if(!strcmp(des->ext, "imf")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
		permuteFloatTab( im->p[j], im->nc);
	}
	else if(!strcmp(des->ext, "IMF")){
	  printf("Pas de swap des octets lors de la lecture car fichier .IMF (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif
  
  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}

/************************************************/
/* .cxs | CXS      */
/************************************************/
int select_imacx4_calc(select_ima_t *des, imacx4 *im){

  int j, nb_oct, sp;

  sp = sizeof(pixcx4);
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imacx4_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  printf("\n>> select_imacx4_calc: lecture %s : bloc %d, lignes [%d %d]\n",
	  im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);
  for(j=0; j<im->nr; j++){
      lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
      if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	    printf("\n>> ATT select_imacx4_calc: lecture %s incomplete\n", im->nom);
	    close(des->dfi);
	    return(2); 
      }
  }

#ifdef INVERSE_DATA
	if(!strcmp(des->ext, "cxs")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
	    permuteShortTab( (short *)&im->p[j][0].re, 2*im->nc);
	}
	else if(!strcmp(des->ext, "CXS")){
	  printf("Pas de swap des octets lors de la lecture car fichier .CXS (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif

  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}

/************************************************/
/* .cxf  | CXF     */
/************************************************/
int select_imacx8_calc(select_ima_t *des, imacx8 *im){

  int j, nb_oct, sp;

  sp = sizeof(pixcx8);
  nb_oct = im->nc*sp;
  if(des->nb_bloc_lu >= des->nb_bloc){
      close(des->dfi);
      printf("\n>> select_imacx8_calc: lecture %s terminee\n", im->nom);
      return(1); 
  }
  printf("\n>> select_imacx8_calc: lecture %s : bloc %d, lignes [%d %d]\n",
	  im->nom, des->nb_bloc_lu, des->next_lig, des->next_lig+des->nb_lig-1);
  for(j=0; j<im->nr; j++){
      lseek(des->dfi,(long)(des->header + ((des->next_lig+j)*des->dimX + des->col0)*sp), 0);  
      if( read(des->dfi, (char *)im->p[j], nb_oct) != nb_oct ){
	    printf("\n>> ATT select_imacx8_calc: lecture %s incomplete\n", im->nom);
	    close(des->dfi);
	    return(2); 
      }
  }

#ifdef INVERSE_DATA

	if(!strcmp(des->ext, "cxf")){
	  printf("Swap des octets des donnees lors de la lecture\n");
	  for( j=0; j<im->nr; j++)
	    permuteFloatTab( &im->p[j][0].re, 2*im->nc);
	}
	else if(!strcmp(des->ext, "CXF")){
	  printf("Pas de swap des octets lors de la lecture car fichier .CXF (INTEL)\n");
	}
	else{
	  printf("\n extension %s inconnue\n", des->ext);
	  exit(0);
	} 
#endif 

  des->next_lig += des->nb_lig;
  des->nb_bloc_lu++;
  return(0);
}

	  
