/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/
/* ************************** mfiltmdate_dct.c ****************************/
/*                                                                        */
/*   FILTRE MULTIDATE DE DANIELA COLCUC - VERSION GOURMANDE EN MEMOIRE    */
/*   LOG, DCT SUR AXE TEMPOREL, FILTRE SPATIAL DES CANAUX, DCT-1, EXP     */
/*                                                                        */
/*  Derniere modif : E. TROUVE, 13/09/2000                                */
/* ************************************************************************/
#include <math.h>
#include <limits.h>
#include "image.h"
#include "proto2D.h"
#include "radar.h"
/* #include "radar_comp.h" */

/** executable : filtre multidate frequentiel de Daniela Coltuc.\\ 
Filtrage homomorphique (log) bas sur la DCT de l'axe temporel d'une serie 
d'images RSO et le filtrage spatial des canaux frequentiels.\\
Version gourmande en espace memoire : les N images u2 (2 octets) sont chargees en memoire 
ainsi que N+2 images float (4 octets) : les canaux de DCT et 2 images de travail 
(moy, moycarre). Pour traiter des images entieres ou de grandes series, 
mieux vaut utiliser les 3 executables suivant :\\
- mlogdct_u2 : transforme la serie RSO en serie de canaux de DCT avec lecture/ecriture bloc par bloc\\
- mlee_add : filtrage spatial de Lee a appliquer sur chaque canal frequentiel \\
- midctexp : transfo inverse (DCT-1, exp) et compensation de biais avec lecture/ecriture bloc par bloc\\
\\
Convention sur la numerotation des images :\\
1) les numeros se suivent!\\
2) ils sont donnes a la fin du nom de fichier, avant le .imw/.imf...\\
3) si les numero se reduisent a 1 chiffre, c'est le chiffre seul (ex: 0, 1, 2...)\\
4) si des numeros a 2 chiffres (>9) sont utilises, tous les numeros sont ecrits 
avec 2 digits (ex: 01, 02,...,10, 11...).\\
5) si les numeros depassent 99 (on en est loin!), etendre la fonction nom_image_suivante.\\
\\
Entree : suite d'images u2 (amplitude) \\
Sortie : suite d'images u2 filtrees\\
@name mfiltmdate_dct
@author E. Trouve
@version 1.0 (13/09/2000)
@see lread, lwrite, llogdct, lidctexp, llee_add
*/
int main(int argc, char *argv[]){

  /* operateurs  */
  read_ima_t *pt_rea;
  write_ima_t *pt_wri;
  logdct_t ld;
  idctexp_t idc;
  lee_add_t lee_add;
  moy2D_t  moy;
  moycarre2D_t  moycar;

  /* main  */
  param par0, *ptp;      /* tete et pointeur pour la chaine de parametres */
  int n, N, n0, j, lectOK;
  
  /* images */
  imau2 *pt_im0;       /* images initiales */ 
  imafl *pt_imdct;      /* images des coef de DCT */
  imafl moyamp, moycarre; /* pour accelerer le calcul du filtage spatial */ 
 

/* LECTURE PARAMETRES */
  param_debut(argc, argv, &par0); 
  ptp = &par0; 

  ptp = idctexp_lect(&idc, ptp, "Filtrage multidate dans l'espace transforme (DCT) : ");
  N = idc.N;
  ld.N = N;     /* operateur logdct, lecture inutile, N deja connu */

  /* allocation des  tableaux de descripteurs */
  if( (pt_rea = (read_ima_t *)malloc(N*sizeof(read_ima_t))) == NULL ){
      printf("\n>>mfiltmdate_dct: ERREUR allocation tableau descripteurs impossible\n");
      exit(1);
  }
  if( (pt_wri = (write_ima_t *)malloc(N*sizeof(write_ima_t))) == NULL ){
      printf("\n>>mfiltmdate_dct: ERREUR allocation tableau descripteurs impossible\n");
      exit(1);
  }

  ptp = read_ima_lect(pt_rea, ptp, "premiere image initiale (u2) :");
  ptp = write_ima_lect(pt_wri, ptp, "premiere image resultat (u2) :");

  ptp = lee_add_lect(&lee_add, ptp, "Filtrage spatial des canaux frequentiels (Lee_Add) :");
  moy.dimX = lee_add.dimX;     moy.dimY = lee_add.dimY;    /* operateur moy2D */
  moycar.dimX = lee_add.dimX;  moycar.dimY = lee_add.dimY; /* operateur moycarre2D */

  param_fin(argc, argv, &par0);

  for(n=1; n<N; n++){
      pt_rea[n] = pt_rea[0];
      nom_image_suivante( pt_rea[n-1].nom, pt_rea[n].nom);
      pt_wri[n] = pt_wri[0];
      nom_image_suivante( pt_wri[n-1].nom, pt_wri[n].nom);
  }


/* INITIALISATION  */
  /* allocation des tableaux de bloc images  */
  if( (pt_im0 = (imau2 *)malloc(N*sizeof(imau2))) == NULL ){
      printf ("\n>>mfiltmdate_dct: ERREUR allocation tableau image impossible\n");  
      exit(1);
  }  
  if( (pt_imdct = (imafl *)malloc(N*sizeof(imafl))) == NULL ){
      printf ("\n>>mfiltmdate_dct: ERREUR allocation tableau image impossible\n");  
      exit(1);
  }  

  /* init lecture/ecriture */
  for(n=0; n<N; n++){
      read_imau2_init(pt_rea+n , pt_im0+n);
      write_ima_init(pt_wri+n);
  }

  /* transfo, filtre spatial, transfo inverse */
  logdct_u2_init(&ld, pt_im0, pt_imdct);
  moy2Dfl_init(&moy, pt_imdct[0], &moyamp);
  moycarre2Dfl_init(&moycar, pt_imdct[0], &moycarre);
  /* lee_addfl_init inutil : le resutat est calcule dans l'image initiale  */
  idctexp_u2_init(&idc, pt_imdct, pt_im0); /* resultat dans images inititiales */

/* CALCUL */
  logdct_u2_calc(&ld, pt_im0, pt_imdct);
  for(n=1; n<N; n++){ /* filtrage spatial des canaux de DCT sauf le canal 0 */
      moy2Dfl_calc(&moy, pt_imdct[n], &moyamp);
      moycarre2Dfl_calc(&moycar, pt_imdct[n], &moycarre);
      lee_addfl_calc(&lee_add, pt_imdct[n], moyamp, moycarre, pt_imdct+n);  
  }
  idctexp_u2_calc(&idc, pt_imdct, pt_im0);
      
/* FREMETURE */
  for(n=0; n<N; n++)
      write_imau2_ferm(pt_wri+n, pt_im0[n]);

}



