/*
*    Copyright (c) 2007. The BATI team. All right reserved.
*
*    This file is part of BATI library.
*
*    BATI library is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    BATI library  is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with BATI library.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include "image.h"
#include "proto2D.h"
#include "volume.h"
#include "vect_ima.h"
#include <string.h>



/** \defgroup vector tracing of vectors in an image
operator vector : tracing of vectors in an image or directional filtering of the image along given vectors.

<b> directional filtering:</b> \n
To protect the edges from blurring while smoothing a directional filter is suited. As direction the vectors given in the input file are used.

Filter type:\n
 As direction the eigenvectors calculated by the PCA are suited. Since the third eigenvector is pointing in the most homogeneous direction, the filter was applied in that direction to preserve the edges. 
For reason of simplicity the mean along the direction is chosen. A 1D mean filter is calculated for each voxel along the direction of the given vector.

\f$ Y(x,y,z)=\sum(X(x,y,z)) \qquad \forall (x,y,z) \in A \f$

where A denotes the area of all voxels belonging to the calculated straight line segment of a certain length.

<b> line tracing: </b> \n
This operator visualizes vectors in an image of the format ima3ds2.
The user of the operator has to specify the radius of the window for tracing the vectors. The operator will visualize the vectors for all points with a distance of 2*window radius + 1 in each dimension. Only the orientation is printed with a chosen intensity in the image not the direction. 

Example in 2D:\n
size=2, actual pixel is the center, vector=[0.7,0.7]\n
\verbatim
- - - - X
- - - X -
- - X - -
- X - - -
X - - - -
\endverbatim

<b> Description of the algorithm used to trace the vector for the filtering and the printing:</b> \n
To follow a given direction by a vector a fast algorithm is appropriate. For this purpose the Bresenham algorithm is suited. This algorithm has its origin in computer graphics like CAD. In many of these kind of programs a digital straight line segment connecting two points p1 and p2 of the image domain must be traced on a screen. \n

To use this algorithm in our case the start and end coordinates of the line to draw have been specified using the actual site, the window size and the vector. 
We consider the straight line segment is a finite set of grid points of the square raster grid, individuated by a straight line segment of the Euclidean plane by making use of the grid intersection digitization criterion; the grid point closest to each intersection between the grid and the Euclidean segment is an element of the digital straight line segment. Normally this point set has two unique end points, it can be ordered as 8-path from one end point to the other, and the digital straight line segment can be identified with this 8-path.

The computation of a digital straight line segment uses consecutive 'small' steps from pixel to pixel, starting at p1=(x1,y1) and ending at p2=(x2,y2). These elementary steps can be restricted to be parallel to he coordinate axes (4-path), or can also include diagonal steps (8-path). The latter case is mostly used for representing a digital curve, and it is also used in the straight line algorithm given below. Besides the given points p1 and p2 the algorithm generates additional

	max{|x1-x2|,|y1-y2|}-1

'new' points on the screen for a digital straight line from  p1 to p2.\n
To apply this algorithm in 3D, the y has to be replaced by the z to get the third coordinate of the line points. Thus the 3D case uses the two 2D projection of the line in the XY- and the ZY-plane.

<b> Descriptor:</b> \n
    - char * nom_in:       filename of the input image of format ima3ds2
    - char * nom_out:      filename of the output image of format ima3ds2
    - char * nom_vect:     filename of the vector file (format described below)
    - int size:            size of the operator window 
    - int dir:             flag of modus (directional filtering: 0, or tracing: 1)
    - int intensity:       intensity of the line to draw (only used if flag=1)

The <b> vector file</b> has to have the following format:\n
 <tt> dimensions (struct dimensions)</tt>   \n
 <tt> vector1 (struct eigenvectors) </tt>  \n
 <tt> . </tt> \n
 <tt> vectori (struct eigenvectors) </tt>  \n
 <tt> . </tt>  \n
 <tt> . </tt>  \n

The vectors are saved sequentially in the counting order x, y, z.
The type definitions of these structures are available in the header file filelib.h.
@name vect_ima
@author C. MAIER
@version 1.0 (03/04/00); Include : vect_ima.h

*/

int getev(allEigenvectors * result,char *filename)
{
  FILE * fattr;
  dimensions * dim;  
  if ((fattr = fopen(filename,"r"))==NULL){
   printf("file %s not found\n",filename);
   exit(3);
  }   
  result->dim=(dimensions *) malloc(sizeof(dimensions));
  fread(result->dim ,sizeof(dimensions),1,fattr);
  result->data=(eigenvector *)malloc(result->dim->dimx * result->dim->dimy* result->dim->dimz * sizeof(eigenvector));
  fread(result->data ,sizeof(eigenvector),result->dim->dimx * result->dim->dimy* result->dim->dimz,fattr);
  close(fattr);
  return(0);  
}

short round_perso(float in)
{
  float min, max;
  max=floor(in);
  min=ceil(in);
  if (in > ((min + max)/2))
    return (short) min;
  else
    return (short) max;
  
}
short getDiff(float * vect,short size,short * diffX,short * diffY,short *diffZ)
     /* returns the differnet cooridinates between the endpoint and the strarting point of the vector in the parameters the return value is the value with the largest difference posible values: 0,1,2 */

{
  float fakt;
  short val;
 
  if (vect[0]==vect[1]&&vect[1]==vect[2]&&vect[0]==(float) 0)
    return 99;
  if ((fabs(vect[0])>=fabs(vect[1]))&&(fabs(vect[0])>=fabs(vect[2]))){
    fakt=size/fabs(vect[0]);
    val=0;   
  }
  else if ((fabs(vect[1])>=fabs(vect[0]))&&(fabs(vect[1])>=fabs(vect[2]))){
    fakt=size/fabs(vect[1]);
    val =1;
  }
  else if ((fabs(vect[2])>=fabs(vect[0]))&&(fabs(vect[2])>=fabs(vect[1]))) {
    fakt=size/fabs(vect[2]);
    val=2;
  }
  vect[0]*=fakt;
  vect[1]*=fakt;
  vect[2]*=fakt;
  
  *diffX=round_perso(vect[0]);
  *diffY=round_perso(vect[1]);
  *diffZ=round_perso(vect[2]); 
  return val;
}


short operatoru1print(ima3Du1* image,short x, short y, short z,short diffX,short diffY,short diffZ,short flag)
     /* flag: 0: xyz (the correct order)
	      1: yxz
	      2: zxy */
     /* traces the vector in the Original image */
	
{
  short switch1=2*fabs(diffY)-fabs(diffX),switch2=2*fabs(diffZ)-fabs(diffX);
  short inx1=2*fabs(diffY), inx2=2*fabs(diffZ), inxy=2*(fabs(diffY)-fabs(diffX)), inxz=2*(fabs(diffZ)-fabs(diffX));
  short  min=255 ,max=-255, stopX=x+diffX, stopY=y+diffY, stopZ=z+diffZ,data,pox=x,poy=y,poz=z,count=0;
  int sum=0;
 
  /* wenn der Ausganspunkt in der Mitte liegt wird erst in eine Rtg gegangen und dann in die entgegengesetzte, deshalb gibt es neben den eigentlichen x,y,z Zaehlern noch entgegengesetze */
  if (flag==0){     
       image->data[z][y][x]=0; image->data[poz][poy][pox]=0;
       }
  else if (flag==1){       
       image->data[z][x][y]=0; image->data[poz][pox][poy]=0; 
       }
  else {      
       image->data[x][z][y]=0; image->data[pox][poz][poy]=0; 
       } 
  while  (x!=stopX)  {         
     sum+=data;
     if (diffX<0){
       x--;
       pox++;
     }
     else{
       x++;
       pox--;
     }
     if (switch1>0){
       if (diffY<0){
	 poy++;
	 y--;
       }
       else{
	 y++;
	 poy--;
       }
       switch1+=inxy;  
     }
     else
       switch1+=inx1;
     if (switch2>0){
       if (diffZ<0){
        z--;
	poz++;
       }
       else{
	 z++;
	 poz--;
       }
       switch2+=inxz;  
     }
     else
       switch2+=inx2;
     if (flag==0){     
       image->data[z][y][x]=0; image->data[poz][poy][pox]=0;
       }
     else if (flag==1){       
         image->data[z][x][y]=0; image->data[poz][pox][poy]=0; 
	 }
     else	 {      
	 image->data[x][z][y]=0; image->data[pox][poz][poy]=0; 
	 }
         }
  return sum/fabs(diffX+1);
}

short operators2print(ima3Ds2 * image,short x, short y, short z,short diffX,short diffY,short diffZ,short flag,short greylevel)
     /* flag: 0: xyz (the correct order)
	      1: yxz
	      2: zxy */
     /* traces the vector in the Original image */
	
{
  short switch1=2*fabs(diffY)-fabs(diffX),switch2=2*fabs(diffZ)-fabs(diffX);
  short inx1=2*fabs(diffY), inx2=2*fabs(diffZ), inxy=2*(fabs(diffY)-fabs(diffX)), inxz=2*(fabs(diffZ)-fabs(diffX));
  short  min=255 ,max=-255, stopX=x+diffX, stopY=y+diffY, stopZ=z+diffZ,data,count=0,pox=x,poy=y,poz=z;
  int sum=0;
  /* takes the input point as middle of the window and traces a line through this piont in every direction */
  /* wenn der Ausganspunkt in der Mitte liegt wird erst in eine Rtg gegangen und dann in die entgegengesetzte, deshalb gibt es neben den eigentlichen x,y,z Zaehlern noch entgegengesetze */
   if (flag==0){     
       image->data[z][y][x]=greylevel;
       }
   else if (flag==1){       
         image->data[z][x][y]=greylevel;
	 }
   else	 {      
	 image->data[x][z][y]=greylevel;
	 }
  while  (x!=stopX)  {         
    
     sum+=data;
     if (diffX<0){
       x--;
       pox++;
     }
     else{
       x++;
       pox--;
     }
     if (switch1>0){
       if (diffY<0){
	 poy++;
	 y--;
       }
       else{
	 y++;
	 poy--;
       }
       switch1+=inxy;  
     }
     else
       switch1+=inx1;
     if (switch2>0){
       if (diffZ<0){
        z--;
	poz++;
       }
       else{
	 z++;
	 poz--;
       }
       switch2+=inxz;  
     }
     else
       switch2+=inx2;
     if (flag==0){     
       image->data[z][y][x]=greylevel;image->data[poz][poy][pox]=greylevel;
       }
     else if (flag==1){       
         image->data[z][x][y]=greylevel;  image->data[poz][pox][poy]=greylevel; 
	 }
     else	 {      
	 image->data[x][z][y]=greylevel;  image->data[pox][poz][poy]=greylevel;
	 }
         } 
  if (flag==0){     
       image->data[z+diffZ][y+diffY][x+diffX]=greylevel;
       }
   else if (flag==1){       
         image->data[z+diffZ][x+diffX][y+diffY]=greylevel;
	 }
   else	 {      
	 image->data[x+diffX][z+diffZ][y+diffY]=greylevel;
	 }
  return 0;
}
short means2(ima3Ds2 * image,short x, short y, short z,short diffX,short diffY,short diffZ,short flag)
     /* flag: 0: xyz (the correct order)
	      1: yxz
	      2: zxy */
     /*calculates the mean of all points on the line of the vector */
	
{
  short switch1=2*fabs(diffY)-fabs(diffX),switch2=2*fabs(diffZ)-fabs(diffX);
  short inx1=2*fabs(diffY), inx2=2*fabs(diffZ), inxy=2*(fabs(diffY)-fabs(diffX)), inxz=2*(fabs(diffZ)-fabs(diffX));
  short  min=255 ,max=-255, stopX=x+diffX, stopY=y+diffY, stopZ=z+diffZ,data,count=0,pox=x,poy=y,poz=z;
  int sum=0;
 
  /* wenn der Ausganspunkt in der Mitte liegt wird erst in eine Rtg gegangen und dann in die entgegengesetzte, deshalb gibt es neben den eigentlichen x,y,z Zaehlern noch entgegengesetze */
  if (flag==0){     
         sum=image->data[z][y][x];
	       }
  else if (flag==1){       
         sum=image->data[z][x][y]; 
  }
  else	 {      
	 sum=image->data[x][z][y]; 
  }
  while  (x!=stopX)  {                 
     if (diffX<0){
       x--;
       pox++;
     }
     else{
       x++;
       pox--;
     }
     if (switch1>0){
       if (diffY<0){
	 poy++;
	 y--;
       }
       else{
	 y++;
	 poy--;
       }
       switch1+=inxy;  
     }
     else
       switch1+=inx1;
     if (switch2>0){
       if (diffZ<0){
        z--;
	poz++;
       }
       else{
	 z++;
	 poz--;
       }
       switch2+=inxz;  
     }
     else
       switch2+=inx2;
     if (flag==0){     
         sum+=(image->data[z][y][x]+image->data[poz][poy][pox]);
     }
     else if (flag==1){       
         sum+=(image->data[z][x][y]+image->data[poz][pox][poy]); 
     }	
     else	 {      
	 sum+=(image->data[x][z][y]+image->data[pox][poz][poy]); 
     }
 }  
  /*if (((sum/(2*fabs(diffX)+1))>127.0)||((sum/(2*fabs(diffX)+1))<-128.0)){
    printf("sumfl %f, sum %d\n", (sum/(2*fabs(diffX)+1)),(short) (sum/(2*fabs(diffX))+1));
    printf("x: %d ,y: %d,z: %d flag: %d\n",x-diffX,y-diffY,z-diffZ,flag);
  }*/
 
  return (short) sum/(2*fabs(diffX)+1);
}
/* *************************  INITIALISATION  ***************************/
/* \ingroup vector
    Initialisation. 

    @param des  descriptor pointer
  
    @author C. MAIER
    @version 1.0 (03/04/00); Include : vect_ima.h, filelib.h , volume.h

   
  
*/
int vector_init(vector_t *des){
  return(0);
}


/** \ingroup vector
    reading of parameters for the operator vector. 

    - nom_in: String representing the input image file name
    - nom_out: String representing the output image file name
    - nom_vect: String representing the vector file name
    - size: radius of the analyse window
    - intensity: Greylevel of the traced line 
   
    @param des  descriptor pointer
    @param ptp  pointer of the current parameters allready allocated,     
    @param debq pointer to a string placed to the begin of the question
    @return  returns a new pointer for the parameters
    @author C. MAIER
    @version 1.0 (03/04/00); Include : vect_ima.h, filelib.h , volume.h
*/
param *vector_lect(vector_t *des, param *ptp, char *debq){

  char question[500];
 
  sprintf(question, "%s nom complet du fichier image a lire",debq); 
  lec_param(question, ptp); 
  des->nom_in = ptp->rep;
  ptp = ptp->next;

  sprintf(question, "%s nom complet du fichier avec les vecteurs a lire", debq); 
  lec_param(question, ptp);  
  des->nom_vect = ptp->rep;
  ptp = ptp->next;
 
  sprintf(question, "%s nom complet du fichier image a ecrire", debq); 
  lec_param(question, ptp);
  des->nom_out = ptp->rep;
  ptp = ptp->next;
 
  sprintf(question, " rayon de la fenetre d'analyse"); 
  lec_param(question, ptp);
  des->size = atoi(ptp->rep);
  ptp = ptp->next; 

  sprintf(question, " modus (filtrage directionelle: 0, ou tracer les vectuers: 1)"); 
  lec_param(question, ptp); 
  des->dir = atoi(ptp->rep);
  ptp = ptp->next;

  sprintf(question, " intensity de la ligne (uniquement quand modus=1)"); 
  lec_param(question, ptp); 
  des->intensity = atoi(ptp->rep);
  ptp = ptp->next;
  
  return(ptp);
}


/** \ingroup vector
    Printing the vectors in the image / directional filtering of the image .\\

    @param des  pointeur descripteur
   
    @author C. MAIER
    @version 1.0 (03/04/00); Include : vect_ima.h, filelib.h , volume.h
*/

int vector_calc(vector_t *des)
 { 
  ima3Ds2 ima; 
  ima3Ds2 ima2; 
  ima3Ds2 *image=&ima;
  ima3Ds2 *imaout=&ima2;
  allEigenvectors  * result; 
  short dummy;
  short z,y,x,size=des->size;
  short diffX, diffY,diffZ,flag;
  long j;
  float vect[3];
  if ((des->dir!=0)&&(des->dir!=1)){
    printf("illegal Modus %d\n",des->dir);
    exit(3);
  }
  result=(allEigenvectors *)malloc( sizeof( allEigenvectors)); 
  lect_ima3Ds2(des->nom_in,image,ALL); 
  imaout->dimx=image->dimx;
  imaout->dimy=image->dimy;
  imaout->dimz=image->dimz; 
  allouer_ima3Ds2(imaout);
  getev(result,des->nom_vect); 
  if ((result->dim->dimz!=image->dimz)||(result->dim->dimy!=image->dimy)||(result->dim->dimx!=image->dimx)){
    printf("The X,Y or Z diminsions of the input data blocs are not cooresponding to each other!\n");
    exit(3);
  }
  z=-1;
  for (j=0;j<(result->dim->dimx*result->dim->dimy*result->dim->dimz);j++){
    if ((j%result->dim->dimx)==0){ 
      if ((j%(result->dim->dimx*result->dim->dimy))==0){
	z++;
	y=-1;   
      }
      y++;
      x=-1;     
    }
    x++;
    if ((x>size-1)&&(x<result->dim->dimx-size)&&(y>size-1)&&(y<result->dim->dimy-size)&&(z>size-1)&&(z<result->dim->dimz-size)) {  
     if (des->dir==1){
      /*tracing a line along the given vector */         
      if (((z%(2*size+1))==0)&&((y%(2*size+1))==0)&&((x%(2*size+1))==0)) {  
	 vect[0]=result->data[j].element1;
         vect[1]=result->data[j].element2;
         vect[2]=result->data[j].element3;
	 flag=getDiff(vect, size, &diffX, &diffY, &diffZ);    
       	 if (flag==0)
	    dummy=operators2print(imaout,x,y,z,diffX,diffY,diffZ,flag,des->intensity);
	 else if (flag==1)	   
	    dummy=operators2print(imaout,y,x,z,diffY,diffX,diffZ,flag,des->intensity); 
	 else if (flag==2)
	    dummy=operators2print(imaout,z,x,y,diffZ,diffX,diffY,flag,des->intensity);
      }
      else
	   imaout->data[z][y][x]=image->data[z][y][x]; 	 	
      }
     else{
        vect[0]=result->data[j].element1;
        vect[1]=result->data[j].element2;
        vect[2]=result->data[j].element3;
	flag=getDiff(vect, size, &diffX, &diffY, &diffZ);    
       	if (flag==0)
	    imaout->data[z][y][x]=means2(image,x,y,z,diffX,diffY,diffZ,flag);
	else if (flag==1)	   
	    imaout->data[z][y][x]=means2(image,y,x,z,diffY,diffX,diffZ,flag); 
	else if (flag==2)
	    imaout->data[z][y][x]=means2(image,z,x,y,diffZ,diffX,diffY,flag);
	else 
	    imaout->data[z][y][x]=image->data[z][y][x];
      } 
     }
    else
	   imaout->data[z][y][x]=image->data[z][y][x]; 	 
    }	     
                      
  ecr_ima3Ds2(des->nom_out,imaout); 
  liberer_ima3Ds2(imaout); 
  liberer_ima3Ds2(image);   
 }


